// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
#line 1 "../V3Const.cpp"
// -*- mode: C++; c-file-style: "cc-mode" -*-
//*************************************************************************
// DESCRIPTION: Verilator: Constant folding
//
// Code available from: https://verilator.org
//
//*************************************************************************
//
// Copyright 2003-2021 by Wilson Snyder. This program is free software; you
// can redistribute it and/or modify it under the terms of either the GNU
// Lesser General Public License Version 3 or the Perl Artistic License
// Version 2.0.
// SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0
//
//*************************************************************************
// CONST TRANSFORMATIONS:
//   Call on one node for PARAM values, or netlist for overall constant folding:
//      Bottom up traversal:
//          Attempt to convert operands to constants
//          If operands are constant, replace this node with constant.
//*************************************************************************

#include "config_build.h"
#include "verilatedos.h"

#include "V3Global.h"
#include "V3String.h"
#include "V3Const.h"
#include "V3Ast.h"
#include "V3Width.h"
#include "V3Simulate.h"
#include "V3Stats.h"

#include <algorithm>

//######################################################################
// Utilities

class ConstVarMarkVisitor final : public AstNVisitor {
    // NODE STATE
    // AstVar::user4p           -> bool, Var marked, 0=not set yet
private:
    // VISITORS
    virtual void visit(AstVarRef* nodep) override {
        if (nodep->varp()) nodep->varp()->user4(1);
    }
    virtual void visit(AstNode* nodep) override { iterateChildren(nodep); }

public:
    // CONSTRUCTORS
    explicit ConstVarMarkVisitor(AstNode* nodep) {
        AstNode::user4ClearTree();  // Check marked InUse before we're called
        iterate(nodep);
    }
    virtual ~ConstVarMarkVisitor() override = default;
};

class ConstVarFindVisitor final : public AstNVisitor {
    // NODE STATE
    // AstVar::user4p           -> bool, input from ConstVarMarkVisitor
    // MEMBERS
    bool m_found = false;

private:
    // VISITORS
    virtual void visit(AstVarRef* nodep) override {
        if (nodep->varp() && nodep->varp()->user4()) m_found = true;
    }
    virtual void visit(AstNode* nodep) override { iterateChildren(nodep); }

public:
    // CONSTRUCTORS
    explicit ConstVarFindVisitor(AstNode* nodep) { iterateAndNextNull(nodep); }
    virtual ~ConstVarFindVisitor() override = default;
    // METHODS
    bool found() const { return m_found; }
};

// This visitor can be used in the post-expanded Ast from V3Expand, where the Ast satisfies:
// - Constants are 64 bit at most (because words are accessed via AstWordSel)
// - Variables are scoped.
class ConstBitOpTreeVisitor final : public AstNVisitor {
    // TYPES

    struct LeafInfo {  // Leaf node (either AstConst or AstVarRef)
        bool m_polarity = true;
        int m_lsb = 0;
        int m_wordIdx = -1;  // -1 means AstWordSel is not used.
        AstVarRef* m_refp = nullptr;
        AstConst* m_constp = nullptr;
    };
    // Collect information for each Variable to transform as below
    class VarInfo final {
        // MEMBERS
        int m_constResult = -1;  // -1: result is not constant, 0 or 1: result of this tree
        ConstBitOpTreeVisitor* m_parentp;  // ConstBitOpTreeVisitor that holds this VarInfo
        AstVarRef* m_refp;  // Points the variable that this VarInfo covers
        V3Number m_bitPolarity;  // Coefficient of each bit
        static int widthOfRef(AstVarRef* refp) {
            if (AstWordSel* selp = VN_CAST(refp->backp(), WordSel)) return selp->width();
            if (AstCCast* castp = VN_CAST(refp->backp(), CCast)) return castp->width();
            return refp->width();
        }

    public:
        // METHODS
        bool hasConstantResult() const { return m_constResult >= 0; }
        bool sameVarAs(const AstNodeVarRef* otherp) const { return m_refp->sameGateTree(otherp); }
        void setPolarity(bool compBit, int bit) {
            UASSERT_OBJ(!hasConstantResult(), m_refp, "Already has result of " << m_constResult);
            if (m_bitPolarity.bitIsX(bit)) {  // The bit is not yet set
                m_bitPolarity.setBit(bit, compBit);
            } else {  // Priviously set the bit
                const bool sameFlag = m_bitPolarity.bitIs1(bit) == compBit;
                if (m_parentp->isXorTree()) {
                    // ^{x[0], ~x[0], x[2], x[3]} === ~^{x[2], x[3]}
                    UASSERT_OBJ(sameFlag, m_refp, "Only true is set in Xor tree");
                    m_bitPolarity.setBit(bit, 'x');
                } else {  // And, Or
                    // Can ignore this nodep as the bit is already registered
                    if (sameFlag) return;  // a & a == a, b | b == b
                    // Otherwise result is constant
                    m_constResult = m_parentp->isAndTree() ? 0 : 1;
                    m_bitPolarity.setAllBitsX();  // The variable is not referred anymore
                }
            }
        }
        AstNode* getResult() const {
            FileLine* fl = m_refp->fileline();
            AstNode* srcp = VN_CAST(m_refp->backp(), WordSel);
            if (!srcp) srcp = m_refp;
            const int width = widthOfRef(m_refp);

            if (hasConstantResult())
                return new AstConst{fl,
                                    V3Number{srcp, width, static_cast<vluint32_t>(m_constResult)}};

            AstConst* maskValuep = new AstConst{fl, V3Number{srcp, width, 0}};
            maskValuep->num().opBitsNonX(m_bitPolarity);  // 'x' -> 0, 0->1, 1->1
            // Let AstConst be in lhs as it is the common convention
            AstAnd* maskedp = new AstAnd{fl, maskValuep, srcp->cloneTree(false)};
            AstNode* resultp;
            if (m_parentp->isXorTree()) {
                resultp = new AstRedXor{fl, maskedp};
                resultp->dtypep()->widthForce(width, 1);
            } else {
                AstConst* compValuep = maskValuep->cloneTree(false);
                compValuep->num().opBitsOne(m_bitPolarity);  // 'x'->0, 0->0, 1->1
                if (m_parentp->isAndTree()) {
                    resultp = new AstEq{fl, compValuep, maskedp};
                } else {  // Or
                    compValuep->num().opXor(V3Number{compValuep->num()}, maskValuep->num());
                    resultp = new AstNeq{fl, compValuep, maskedp};
                }
            }
            return resultp;
        }

        // CONSTRUCTORS
        VarInfo(ConstBitOpTreeVisitor* parent, AstVarRef* refp)
            : m_parentp(parent)
            , m_refp(refp)
            , m_bitPolarity(refp, widthOfRef(refp)) {
            m_bitPolarity.setAllBitsX();
        }
    };

    // MEMBERS
    bool m_failed = false;
    bool m_polarity = true;  // Flip when AstNot comes
    int m_ops = 0;  // Number of operations such as And, Or, Xor, Sel...
    int m_lsb = 0;  // Current LSB
    LeafInfo* m_leafp = nullptr;  // AstConst or AstVarRef that currently looking for
    AstNode* m_rootp;  // Root of this AST subtree
    AstNode* m_curOpp = nullptr;  // The node that should be added to m_frozenNodes

    AstUser4InUse m_inuser4;
    std::vector<AstNode*> m_frozenNodes;  // Nodes that cannot be optimized
    std::vector<VarInfo*> m_varInfos;  // VarInfo for each variable, [0] is nullptr

    // NODE STATE
    // AstVarRef::user4u      -> Base index of m_varInfos that points VarInfo
    // AstVarScope::user4u    -> Same as AstVarRef::user4

    // METHODS
    VL_DEBUG_FUNC;  // Declare debug()

    bool isAndTree() const { return VN_IS(m_rootp, And); }
    bool isOrTree() const { return VN_IS(m_rootp, Or); }
    bool isXorTree() const { return VN_IS(m_rootp, Xor) || VN_IS(m_rootp, RedXor); }

#define CONST_BITOP_RETURN_IF(cond, nodep) \
    if (setFailed(cond, #cond, nodep, __LINE__)) return

#define CONST_BITOP_SET_FAILED(reason, nodep) setFailed(true, reason, nodep, __LINE__)

    bool setFailed(bool fail, const char* reason, AstNode* nodep, int line) {
        if (fail) {
            UINFO(9, "cannot optimize " << m_rootp << " reason:" << reason << " called from line:"
                                        << line << " when checking:" << nodep << std::endl);
            // if (debug() >= 9) m_rootp->dumpTree(std::cout << "Root node:\n");
        }
        m_failed |= fail;
        return m_failed;
    }
    void incrOps(const AstNode* nodep, int line) {
        ++m_ops;
        UINFO(9, "Increment to " << m_ops << " " << nodep << " called from line " << line << "\n");
    }
    VarInfo& getVarInfo(const LeafInfo& ref) {
        UASSERT_OBJ(ref.m_refp, m_rootp, "null varref in And/Or/Xor optimization");
        AstNode* nodep = ref.m_refp->varScopep();
        if (!nodep) nodep = ref.m_refp->varp();  // Not scoped
        int baseIdx = nodep->user4();
        if (baseIdx == 0) {  // Not set yet
            baseIdx = m_varInfos.size();
            const int numWords
                = ref.m_refp->dtypep()->isWide() ? ref.m_refp->dtypep()->widthWords() : 1;
            m_varInfos.resize(m_varInfos.size() + numWords, nullptr);
            nodep->user4(baseIdx);
        }
        const size_t idx = baseIdx + std::max(0, ref.m_wordIdx);
        VarInfo* varInfop = m_varInfos[idx];
        if (!varInfop) {
            varInfop = new VarInfo{this, ref.m_refp};
            m_varInfos[idx] = varInfop;
        } else {
            if (!varInfop->sameVarAs(ref.m_refp))
                CONST_BITOP_SET_FAILED("different var (scope?)", ref.m_refp);
        }
        return *varInfop;
    }

    // Traverse down to see AstConst or AstVarRef
    LeafInfo findLeaf(AstNode* nodep, bool expectConst) {
        LeafInfo info;
        {
            VL_RESTORER(m_leafp);
            m_leafp = &info;
            iterate(nodep);
        }

        bool ok = !m_failed;
        if (expectConst)
            ok &= !info.m_refp && info.m_constp;
        else
            ok &= info.m_refp && !info.m_constp;
        return ok ? info : LeafInfo{};
    }
    AstNode* combineTree(AstNode* lhsp, AstNode* rhsp) {
        if (!lhsp) return rhsp;
        if (isAndTree())
            return new AstAnd(m_rootp->fileline(), lhsp, rhsp);
        else if (isOrTree())
            return new AstOr(m_rootp->fileline(), lhsp, rhsp);
        else {
            UASSERT_OBJ(isXorTree(), m_rootp, "must be either Xor or RedXor");
            return new AstXor(m_rootp->fileline(), lhsp, rhsp);
        }
    }

    // VISITORS
    virtual void visit(AstNode* nodep) override {
        CONST_BITOP_SET_FAILED("Hit unexpected op", nodep);
    }
    virtual void visit(AstCCast* nodep) override { iterateChildren(nodep); }
    virtual void visit(AstShiftR* nodep) override {
        CONST_BITOP_RETURN_IF(!m_leafp, nodep);
        AstConst* constp = VN_CAST(nodep->rhsp(), Const);
        CONST_BITOP_RETURN_IF(!constp, nodep->rhsp());
        m_lsb += constp->toUInt();
        incrOps(nodep, __LINE__);
        iterate(nodep->lhsp());
        m_lsb -= constp->toUInt();
    }
    virtual void visit(AstNot* nodep) override {
        CONST_BITOP_RETURN_IF(nodep->widthMin() != 1, nodep);
        AstNode* lhsp = nodep->lhsp();
        CONST_BITOP_RETURN_IF(VN_IS(lhsp, And) || VN_IS(lhsp, Or) || VN_IS(lhsp, Const), lhsp);
        incrOps(nodep, __LINE__);
        m_polarity = !m_polarity;
        iterateChildren(nodep);
        // Don't restore m_polarity for Xor as it counts parity of the entire tree
        if (!isXorTree()) m_polarity = !m_polarity;
    }
    virtual void visit(AstWordSel* nodep) override {
        CONST_BITOP_RETURN_IF(!m_leafp, nodep);
        AstConst* constp = VN_CAST(nodep->bitp(), Const);
        CONST_BITOP_RETURN_IF(!constp, nodep->rhsp());
        UASSERT_OBJ(m_leafp->m_wordIdx == -1, nodep, "Unexpected nested WordSel");
        m_leafp->m_wordIdx = constp->toSInt();
        iterate(nodep->fromp());
    }
    virtual void visit(AstVarRef* nodep) override {
        CONST_BITOP_RETURN_IF(!m_leafp, nodep);
        UASSERT_OBJ(!m_leafp->m_refp, nodep, m_leafp->m_refp << " is already set");
        m_leafp->m_refp = nodep;
        m_leafp->m_polarity = m_polarity;
        m_leafp->m_lsb = m_lsb;
    }
    virtual void visit(AstConst* nodep) override {
        CONST_BITOP_RETURN_IF(!m_leafp, nodep);
        UASSERT_OBJ(!m_leafp->m_constp, nodep, m_leafp->m_constp << " is already set");
        m_leafp->m_constp = nodep;
        m_leafp->m_lsb = m_lsb;
    }

    virtual void visit(AstRedXor* nodep) override {  // Expect '^(mask & v)'
        CONST_BITOP_RETURN_IF(!VN_IS(m_rootp, Xor), nodep);
        AstAnd* andp = VN_CAST(nodep->lhsp(), And);
        CONST_BITOP_RETURN_IF(!andp, nodep->lhsp());

        auto mask = findLeaf(andp->lhsp(), true);
        CONST_BITOP_RETURN_IF(!mask.m_constp || mask.m_lsb != 0, andp->lhsp());

        LeafInfo leaf = findLeaf(andp->rhsp(), false);
        CONST_BITOP_RETURN_IF(!leaf.m_refp, andp->rhsp());

        incrOps(nodep, __LINE__);
        incrOps(andp, __LINE__);
        const V3Number& maskNum = mask.m_constp->num();
        VarInfo& varInfo = getVarInfo(leaf);
        for (int i = 0; i < maskNum.width(); ++i) {
            // Set true, m_treePolarity takes care of the entire parity
            if (maskNum.bitIs1(i)) varInfo.setPolarity(true, i + leaf.m_lsb);
        }
    }

    virtual void visit(AstNodeBiop* nodep) override {
        auto isConst = [](AstNode* nodep, vluint64_t v) -> bool {
            AstConst* constp = VN_CAST(nodep, Const);
            return constp && constp->toUQuad() == v;
        };
        if (nodep->type() == m_rootp->type()) {  // And, Or, Xor
            CONST_BITOP_RETURN_IF(!m_polarity && isXorTree(), nodep);
            incrOps(nodep, __LINE__);
            VL_RESTORER(m_curOpp);
            VL_RESTORER(m_leafp);

            for (int i = 0; i < 2; ++i) {
                LeafInfo leafInfo;
                m_leafp = &leafInfo;
                m_curOpp = i == 0 ? nodep->lhsp() : nodep->rhsp();
                const size_t origFrozens = m_frozenNodes.size();
                const int origOps = m_ops;
                const bool origFailed = m_failed;
                iterate(m_curOpp);
                if (leafInfo.m_constp || m_failed) {
                    // Rvert changes in leaf
                    if (m_frozenNodes.size() > origFrozens) m_frozenNodes.resize(origFrozens);
                    m_frozenNodes.push_back(m_curOpp);
                    m_ops = origOps;
                    m_failed = origFailed;
                } else if (leafInfo.m_refp) {
                    VarInfo& varInfo = getVarInfo(leafInfo);
                    if (!varInfo.hasConstantResult()) {
                        varInfo.setPolarity(isXorTree() || leafInfo.m_polarity, leafInfo.m_lsb);
                    }
                }
            }
            return;
        } else if (VN_IS(m_rootp, Xor) && VN_IS(nodep, Eq) && isConst(nodep->lhsp(), 0)
                   && VN_IS(nodep->rhsp(), And)) {  // 0 == (1 & RedXor)
            AstAnd* andp = static_cast<AstAnd*>(nodep->rhsp());  // already checked above
            CONST_BITOP_RETURN_IF(!isConst(andp->lhsp(), 1), andp->lhsp());
            AstRedXor* redXorp = VN_CAST(andp->rhsp(), RedXor);
            CONST_BITOP_RETURN_IF(!redXorp, andp->rhsp());
            incrOps(nodep, __LINE__);
            incrOps(andp, __LINE__);
            m_polarity = !m_polarity;
            iterate(redXorp);
            return;
        } else if (VN_IS(m_rootp, Xor) && VN_IS(nodep, And) && isConst(nodep->lhsp(), 1)
                   && (VN_IS(nodep->rhsp(), Xor)
                       || VN_IS(nodep->rhsp(), RedXor))) {  // 1 & (v[3] ^ v[2])
            incrOps(nodep, __LINE__);
            iterate(nodep->rhsp());
            return;
        } else if ((isAndTree() && VN_IS(nodep, Eq)) || (isOrTree() && VN_IS(nodep, Neq))) {
            CONST_BITOP_RETURN_IF(!m_polarity, nodep);
            const bool maskFlip = isOrTree();
            LeafInfo comp = findLeaf(nodep->lhsp(), true);
            CONST_BITOP_RETURN_IF(!comp.m_constp || comp.m_lsb != 0, nodep->lhsp());

            AstAnd* andp = VN_CAST(nodep->rhsp(), And);  // comp == (mask & v)
            CONST_BITOP_RETURN_IF(!andp, nodep->rhsp());

            LeafInfo mask = findLeaf(andp->lhsp(), true);
            CONST_BITOP_RETURN_IF(!mask.m_constp || mask.m_lsb != 0, andp->lhsp());

            LeafInfo ref = findLeaf(andp->rhsp(), false);
            CONST_BITOP_RETURN_IF(!ref.m_refp, andp->rhsp());

            VarInfo& varInfo = getVarInfo(ref);

            const V3Number maskNum = mask.m_constp->num();
            const V3Number compNum = comp.m_constp->num();
            for (int i = 0; i < maskNum.width() && !varInfo.hasConstantResult(); ++i) {
                const int bit = i + ref.m_lsb;
                if (maskNum.bitIs0(i)) continue;
                varInfo.setPolarity(compNum.bitIs1(i) ^ maskFlip, bit);
            }
            incrOps(nodep, __LINE__);
            incrOps(andp, __LINE__);
            return;
        }
        CONST_BITOP_SET_FAILED("Mixture of different ops cannot be optimized", nodep);
    }

    // CONSTRUCTORS
    ConstBitOpTreeVisitor(AstNode* nodep, int ops)
        : m_ops(ops)
        , m_rootp(nodep) {
        // Fill nullptr at [0] because AstVarScope::user4 is 0 by default
        m_varInfos.push_back(nullptr);
        CONST_BITOP_RETURN_IF(!isAndTree() && !isOrTree() && !isXorTree(), nodep);
        AstNode::user4ClearTree();
        if (AstNodeBiop* biopp = VN_CAST(nodep, NodeBiop)) {
            iterate(biopp);
        } else {
            incrOps(nodep, __LINE__);
            iterateChildren(nodep);
        }
        UASSERT_OBJ(isXorTree() || m_polarity, nodep, "must be the original polarity");
    }
    virtual ~ConstBitOpTreeVisitor() {
        for (size_t i = 0; i < m_varInfos.size(); ++i) {
            VL_DO_DANGLING(delete m_varInfos[i], m_varInfos[i]);
        }
    }
#undef CONST_BITOP_RETURN_IF
#undef CONST_BITOP_SET_FAILED

public:
    // Transform as below.
    // v[0] & v[1] => 2'b11 == (2'b11 & v)
    // v[0] | v[1] => 2'b00 != (2'b11 & v)
    // v[0] ^ v[1] => ^{2'b11 & v}
    // (3'b011 == (3'b011 & v)) & v[2]  => 3'b111 == (3'b111 & v)
    // (3'b000 != (3'b011 & v)) | v[2]  => 3'b000 != (3'b111 & v)
    // Reduction ops are transformed in the same way.
    // &{v[0], v[1]} => 2'b11 == (2'b11 & v)
    static AstNode* simplify(AstNode* nodep, int ops, VDouble0& reduction) {
        ConstBitOpTreeVisitor visitor{nodep, ops};
        if (visitor.m_failed || visitor.m_varInfos.size() == 1) return nullptr;

        // Two ops for each varInfo. (And and Eq)
        const int vars = visitor.m_varInfos.size() - 1;
        int constTerms = 0;
        for (const VarInfo* v : visitor.m_varInfos) {
            if (v && v->hasConstantResult()) ++constTerms;
        }
        // Expected number of ops after this simplification
        // e.g. (comp0 == (mask0 & var0)) & (comp1 == (mask1 & var1)) & ....
        // e.g. redXor(mask1 & var0) ^ redXor(mask1 & var1)
        //  2 ops per variables, numVars - 1 ops among variables
        int expOps = 2 * (vars - constTerms) + vars - 1;
        expOps += 2 * visitor.m_frozenNodes.size();
        if (visitor.isXorTree()) {
            ++expOps;  // AstRedXor::cleanOut() == false, so need 1 & redXor
            if (!visitor.m_polarity) ++expOps;  // comparison with 0
        }
        if (visitor.m_ops <= expOps) return nullptr;  // Unless benefitial, return

        reduction += visitor.m_ops - expOps;

        AstNode* resultp = nullptr;
        // VarInfo in visitor.m_varInfos appears in deterministic order,
        // so the optimized AST is deterministic too.
        for (const VarInfo* varinfop : visitor.m_varInfos) {
            if (!varinfop) continue;
            AstNode* partialresultp = varinfop->getResult();
            resultp = visitor.combineTree(resultp, partialresultp);
        }
        AstNode* frozensp = nullptr;
        for (AstNode* frozenp : visitor.m_frozenNodes) {
            frozenp->unlinkFrBack();
            frozensp = visitor.combineTree(frozensp, frozenp);
        }
        if (frozensp) resultp = visitor.combineTree(resultp, frozensp);

        if (visitor.isXorTree()) {
            // VL_REDXOR_N functions don't guarantee to return only 0/1
            const int width = resultp->width();
            FileLine* fl = nodep->fileline();
            resultp = new AstAnd{fl, new AstConst{fl, V3Number{nodep, width, 1}}, resultp};
            if (!visitor.m_polarity) {
                resultp = new AstEq{fl, new AstConst{fl, V3Number{nodep, width, 0}}, resultp};
                resultp->dtypep()->widthForce(1, 1);
            }
        }
        if (resultp->width() != nodep->width()) {
            resultp = new AstCCast{resultp->fileline(), resultp, nodep};
        }
        return resultp;
    }
};

//######################################################################
// Const state, as a visitor of each AstNode

class ConstVisitor final : public AstNVisitor {
private:
    // NODE STATE
    // ** only when m_warn/m_doExpensive is set.  If state is needed other times,
    // ** must track down everywhere V3Const is called and make sure no overlaps.
    // AstVar::user4p           -> Used by ConstVarMarkVisitor/ConstVarFindVisitor
    // AstJumpLabel::user4      -> bool.  Set when AstJumpGo uses this label
    // AstEnum::user4           -> bool.  Recursing.

    // STATE
    bool m_params = false;  // If true, propagate parameterized and true numbers only
    bool m_required = false;  // If true, must become a constant
    bool m_wremove = true;  // Inside scope, no assignw removal
    bool m_warn = false;  // Output warnings
    bool m_doExpensive = false;  // Enable computationally expensive optimizations
    bool m_doCpp = false;  // Enable late-stage C++ optimizations
    bool m_doNConst = false;  // Enable non-constant-child simplifications
    bool m_doShort = true;  // Remove expressions that short circuit
    bool m_doV = false;  // Verilog, not C++ conversion
    bool m_doGenerate = false;  // Postpone width checking inside generate
    bool m_hasJumpDelay = false;  // JumpGo or Delay under this while
    AstNodeModule* m_modp = nullptr;  // Current module
    AstArraySel* m_selp = nullptr;  // Current select
    AstNode* m_scopep = nullptr;  // Current scope
    AstAttrOf* m_attrp = nullptr;  // Current attribute
    VDouble0 m_statBitOpReduction;  // Ops reduced in ConstBitOpTreeVisitor

    // METHODS
    VL_DEBUG_FUNC;  // Declare debug()

    bool operandConst(AstNode* nodep) { return VN_IS(nodep, Const); }
    bool operandAsvConst(const AstNode* nodep) {
        // BIASV(CONST, BIASV(CONST,...)) -> BIASV( BIASV_CONSTED(a,b), ...)
        const AstNodeBiComAsv* bnodep = VN_CAST_CONST(nodep, NodeBiComAsv);
        if (!bnodep) return false;
        if (!VN_IS(bnodep->lhsp(), Const)) return false;
        const AstNodeBiComAsv* rnodep = VN_CAST_CONST(bnodep->rhsp(), NodeBiComAsv);
        if (!rnodep) return false;
        if (rnodep->type() != bnodep->type()) return false;
        if (rnodep->width() != bnodep->width()) return false;
        if (rnodep->lhsp()->width() != bnodep->lhsp()->width()) return false;
        if (!VN_IS(rnodep->lhsp(), Const)) return false;
        return true;
    }
    bool operandAsvSame(const AstNode* nodep) {
        // BIASV(SAMEa, BIASV(SAMEb,...)) -> BIASV( BIASV(SAMEa,SAMEb), ...)
        const AstNodeBiComAsv* bnodep = VN_CAST_CONST(nodep, NodeBiComAsv);
        if (!bnodep) return false;
        const AstNodeBiComAsv* rnodep = VN_CAST_CONST(bnodep->rhsp(), NodeBiComAsv);
        if (!rnodep) return false;
        if (rnodep->type() != bnodep->type()) return false;
        if (rnodep->width() != bnodep->width()) return false;
        return operandsSame(bnodep->lhsp(), rnodep->lhsp());
    }
    bool operandAsvLUp(const AstNode* nodep) {
        // BIASV(BIASV(CONSTll,lr),r) -> BIASV(CONSTll,BIASV(lr,r)) ?
        //
        // Example of how this is useful:
        // BIASV(BIASV(CONSTa,b...),BIASV(CONSTc,d...))  // hits operandAsvUp
        // BIASV(CONSTa,BIASV(b...,BIASV(CONSTc,d...)))  // hits operandAsvUp
        // BIASV(CONSTa,BIASV(CONSTc,BIASV(c...,d...)))  // hits operandAsvConst
        // BIASV(BIASV(CONSTa,CONSTc),BIASV(c...,d...))) // hits normal constant propagation
        // BIASV(CONST_a_c,BIASV(c...,d...)))
        //
        // Idea for the future: All BiComAsvs could be lists, sorted by if they're constant
        const AstNodeBiComAsv* bnodep = VN_CAST_CONST(nodep, NodeBiComAsv);
        if (!bnodep) return false;
        const AstNodeBiComAsv* lnodep = VN_CAST_CONST(bnodep->lhsp(), NodeBiComAsv);
        if (!lnodep) return false;
        if (lnodep->type() != bnodep->type()) return false;
        if (lnodep->width() != bnodep->width()) return false;
        return VN_IS(lnodep->lhsp(), Const);
    }
    bool operandAsvRUp(const AstNode* nodep) {
        // BIASV(l,BIASV(CONSTrl,rr)) -> BIASV(CONSTrl,BIASV(l,rr)) ?
        const AstNodeBiComAsv* bnodep = VN_CAST_CONST(nodep, NodeBiComAsv);
        if (!bnodep) return false;
        const AstNodeBiComAsv* rnodep = VN_CAST_CONST(bnodep->rhsp(), NodeBiComAsv);
        if (!rnodep) return false;
        if (rnodep->type() != bnodep->type()) return false;
        if (rnodep->width() != bnodep->width()) return false;
        return VN_IS(rnodep->lhsp(), Const);
    }
    static bool operandSubAdd(const AstNode* nodep) {
        // SUB( ADD(CONSTx,y), CONSTz) -> ADD(SUB(CONSTx,CONSTz), y)
        const AstNodeBiop* np = VN_CAST_CONST(nodep, NodeBiop);
        const AstNodeBiop* lp = VN_CAST_CONST(np->lhsp(), NodeBiop);
        return (lp && VN_IS(lp->lhsp(), Const) && VN_IS(np->rhsp(), Const)
                && lp->width() == np->width());
    }
    bool matchRedundantClean(AstAnd* andp) {
        // Remove And with constant one inserted by V3Clean
        // 1 & (a == b)  -> (IData)(a == b)
        // When bool is casted to int, the value is either 0 or 1
        AstConst* constp = VN_CAST(andp->lhsp(), Const);
        UASSERT_OBJ(constp && constp->isOne(), andp->lhsp(), "TRREEOPC must meet this condition");
        AstNode* rhsp = andp->rhsp();
        AstCCast* ccastp = nullptr;
        auto isEqOrNeq
            = [](AstNode* nodep) -> bool { return VN_IS(nodep, Eq) || VN_IS(nodep, Neq); };
        if (isEqOrNeq(rhsp)) {
            ccastp = new AstCCast{andp->fileline(), rhsp->unlinkFrBack(), andp};
        } else if (AstCCast* tmpp = VN_CAST(rhsp, CCast)) {
            if (isEqOrNeq(tmpp->lhsp())) {
                if (tmpp->width() == andp->width()) {
                    tmpp->unlinkFrBack();
                    ccastp = tmpp;
                } else {
                    ccastp = new AstCCast{andp->fileline(), tmpp->lhsp()->unlinkFrBack(), andp};
                }
            }
        }
        if (ccastp) {
            andp->replaceWith(ccastp);
            VL_DO_DANGLING(andp->deleteTree(), andp);
        }
        return ccastp;
    }

    static bool operandAndOrSame(const AstNode* nodep) {
        // OR( AND(VAL,x), AND(VAL,y)) -> AND(VAL,OR(x,y))
        // OR( AND(x,VAL), AND(y,VAL)) -> AND(OR(x,y),VAL)
        const AstNodeBiop* np = VN_CAST_CONST(nodep, NodeBiop);
        const AstNodeBiop* lp = VN_CAST_CONST(np->lhsp(), NodeBiop);
        const AstNodeBiop* rp = VN_CAST_CONST(np->rhsp(), NodeBiop);
        return (lp && rp && lp->width() == rp->width() && lp->type() == rp->type()
                && (operandsSame(lp->lhsp(), rp->lhsp()) || operandsSame(lp->rhsp(), rp->rhsp())));
    }
    static bool matchOrAndNot(AstNodeBiop* nodep) {
        // AstOr{$a, AstAnd{AstNot{$b}, $c}} if $a.width1, $a==$b => AstOr{$a,$c}
        // Someday we'll sort the biops completely and this can be simplified
        // This often results from our simplified clock generation:
        // if (rst) ... else if (enable)... -> OR(rst,AND(!rst,enable))
        AstNode* ap;
        AstNodeBiop* andp;
        if (VN_IS(nodep->lhsp(), And)) {
            andp = VN_CAST(nodep->lhsp(), And);
            ap = nodep->rhsp();
        } else if (VN_IS(nodep->rhsp(), And)) {
            andp = VN_CAST(nodep->rhsp(), And);
            ap = nodep->lhsp();
        } else
            return false;
        AstNodeUniop* notp;
        AstNode* cp;
        if (VN_IS(andp->lhsp(), Not)) {
            notp = VN_CAST(andp->lhsp(), Not);
            cp = andp->rhsp();
        } else if (VN_IS(andp->rhsp(), Not)) {
            notp = VN_CAST(andp->rhsp(), Not);
            cp = andp->lhsp();
        } else
            return false;
        AstNode* bp = notp->lhsp();
        if (!operandsSame(ap, bp)) return false;
        // Do it
        cp->unlinkFrBack();
        VL_DO_DANGLING(andp->unlinkFrBack()->deleteTree(), andp);
        VL_DANGLING(notp);
        // Replace whichever branch is now dangling
        if (nodep->rhsp()) {
            nodep->lhsp(cp);
        } else {
            nodep->rhsp(cp);
        }
        return true;
    }
    bool matchAndCond(AstAnd* nodep) {
        // Push down a AND into conditional, when one side of conditional is constant
        // (otherwise we'd be trading one operation for two operations)
        // V3Clean often makes this pattern, as it postpones the AND until
        // as high as possible, which is usally the right choice, except for this.
        AstNodeCond* condp = VN_CAST(nodep->rhsp(), NodeCond);
        if (!condp) return false;
        if (!VN_IS(condp->expr1p(), Const) && !VN_IS(condp->expr2p(), Const)) return false;
        AstConst* maskp = VN_CAST(nodep->lhsp(), Const);
        if (!maskp) return false;
        UINFO(4, "AND(CONSTm, CONDcond(c, i, e))->CONDcond(c, AND(m,i), AND(m, e)) " << nodep
                                                                                     << endl);
        AstNodeCond* newp = static_cast<AstNodeCond*>(
            condp->cloneType(condp->condp()->unlinkFrBack(),
                             new AstAnd(nodep->fileline(), maskp->cloneTree(false),
                                        condp->expr1p()->unlinkFrBack()),
                             new AstAnd(nodep->fileline(), maskp->cloneTree(false),
                                        condp->expr2p()->unlinkFrBack())));
        newp->dtypeFrom(nodep);
        newp->expr1p()->dtypeFrom(nodep);  // As And might have been to change widths
        newp->expr2p()->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        return true;
    }
    bool matchBitOpTree(AstNode* nodep) {
        if (!v3Global.opt.oConstBitOpTree()) return false;

        AstNode* newp = nullptr;
        bool tried = false;
        if (AstAnd* andp = VN_CAST(nodep, And)) {  // 1 & BitOpTree
            if (AstConst* bitMaskp = VN_CAST(andp->lhsp(), Const)) {
                if (bitMaskp->num().toUQuad() != 1) return false;
                newp = ConstBitOpTreeVisitor::simplify(andp->rhsp(), 1, m_statBitOpReduction);
                tried = true;
            }
        }
        if (!tried) {
            // (comp == BitOpTree) & BitOpTree
            // (comp != BitOpTree) | BitOpTree
            newp = ConstBitOpTreeVisitor::simplify(nodep, 0, m_statBitOpReduction);
        }
        if (newp) {
            UINFO(4, "Transformed leaf of bit tree to " << newp << std::endl);
            if (debug() >= 9) {  // LCOV_EXCL_START
                static int c = 0;
                std::cout << "Call matchBitOpTree[" << c << "]\n";
                nodep->dumpTree(std::cout);
                std::cout << "\nResult:\n";
                newp->dumpTree(std::cout);
                ++c;
            }  // LCOV_EXCL_STOP
            nodep->replaceWith(newp);
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
        }
        return newp;
    }
    static bool operandShiftSame(const AstNode* nodep) {
        const AstNodeBiop* np = VN_CAST_CONST(nodep, NodeBiop);
        {
            const AstShiftL* lp = VN_CAST_CONST(np->lhsp(), ShiftL);
            const AstShiftL* rp = VN_CAST_CONST(np->rhsp(), ShiftL);
            if (lp && rp) {
                return (lp->width() == rp->width() && lp->lhsp()->width() == rp->lhsp()->width()
                        && operandsSame(lp->rhsp(), rp->rhsp()));
            }
        }
        {
            const AstShiftR* lp = VN_CAST_CONST(np->lhsp(), ShiftR);
            const AstShiftR* rp = VN_CAST_CONST(np->rhsp(), ShiftR);
            if (lp && rp) {
                return (lp->width() == rp->width() && lp->lhsp()->width() == rp->lhsp()->width()
                        && operandsSame(lp->rhsp(), rp->rhsp()));
            }
        }
        return false;
    }
    bool operandHugeShiftL(const AstNodeBiop* nodep) {
        return (VN_IS(nodep->rhsp(), Const)
                && !VN_CAST_CONST(nodep->rhsp(), Const)->num().isFourState()
                && (VN_CAST_CONST(nodep->rhsp(), Const)->toUInt()
                    >= static_cast<uint32_t>(nodep->width()))
                && isTPure(nodep->lhsp()));
    }
    bool operandHugeShiftR(const AstNodeBiop* nodep) {
        return (VN_IS(nodep->rhsp(), Const)
                && !VN_CAST_CONST(nodep->rhsp(), Const)->num().isFourState()
                && (VN_CAST_CONST(nodep->rhsp(), Const)->toUInt()
                    >= static_cast<uint32_t>(nodep->lhsp()->width()))
                && isTPure(nodep->lhsp()));
    }
    bool operandIsTwo(const AstNode* nodep) {
        return (VN_IS(nodep, Const) && !VN_CAST_CONST(nodep, Const)->num().isFourState()
                && nodep->width() <= VL_QUADSIZE && VN_CAST_CONST(nodep, Const)->toUQuad() == 2);
    }
    bool operandIsTwostate(const AstNode* nodep) {
        return (VN_IS(nodep, Const) && !VN_CAST_CONST(nodep, Const)->num().isFourState());
    }
    bool operandIsPowTwo(const AstNode* nodep) {
        if (!operandIsTwostate(nodep)) return false;
        return (1 == VN_CAST_CONST(nodep, Const)->num().countOnes());
    }
    bool operandShiftOp(const AstNodeBiop* nodep) {
        if (!VN_IS(nodep->rhsp(), Const)) return false;
        const AstNodeBiop* lhsp = VN_CAST(nodep->lhsp(), NodeBiop);
        if (!lhsp || !(VN_IS(lhsp, And) || VN_IS(lhsp, Or) || VN_IS(lhsp, Xor))) return false;
        if (nodep->width() != lhsp->width()) return false;
        if (nodep->width() != lhsp->lhsp()->width()) return false;
        if (nodep->width() != lhsp->rhsp()->width()) return false;
        return true;
    }
    bool operandShiftShift(const AstNodeBiop* nodep) {
        // We could add a AND though.
        const AstNodeBiop* lhsp = VN_CAST(nodep->lhsp(), NodeBiop);
        if (!lhsp || !(VN_IS(lhsp, ShiftL) || VN_IS(lhsp, ShiftR))) return false;
        // We can only get rid of a<<b>>c or a<<b<<c, with constant b & c
        // because bits may be masked in that process, or (b+c) may exceed the word width.
        if (!(VN_IS(nodep->rhsp(), Const) && VN_IS(lhsp->rhsp(), Const))) return false;
        if (VN_CAST_CONST(nodep->rhsp(), Const)->num().isFourState()
            || VN_CAST_CONST(lhsp->rhsp(), Const)->num().isFourState())
            return false;
        if (nodep->width() != lhsp->width()) return false;
        if (nodep->width() != lhsp->lhsp()->width()) return false;
        return true;
    }
    bool operandWordOOB(const AstWordSel* nodep) {
        // V3Expand may make a arraysel that exceeds the bounds of the array
        // It was an expression, then got constified.  In reality, the WordSel
        // must be wrapped in a Cond, that will be false.
        return (VN_IS(nodep->rhsp(), Const) && VN_IS(nodep->fromp(), NodeVarRef)
                && VN_CAST_CONST(nodep->fromp(), NodeVarRef)->access().isReadOnly()
                && (static_cast<int>(VN_CAST_CONST(nodep->rhsp(), Const)->toUInt())
                    >= VN_CAST(nodep->fromp(), NodeVarRef)->varp()->widthWords()));
    }
    bool operandSelFull(const AstSel* nodep) {
        return (VN_IS(nodep->lsbp(), Const) && VN_IS(nodep->widthp(), Const)
                && nodep->lsbConst() == 0
                && static_cast<int>(nodep->widthConst()) == nodep->fromp()->width());
    }
    bool operandSelExtend(AstSel* nodep) {
        // A pattern created by []'s after offsets have been removed
        // SEL(EXTEND(any,width,...),(width-1),0) -> ...
        // Since select's return unsigned, this is always an extend
        AstExtend* extendp = VN_CAST(nodep->fromp(), Extend);
        if (!(m_doV && extendp && VN_IS(nodep->lsbp(), Const) && VN_IS(nodep->widthp(), Const)
              && nodep->lsbConst() == 0
              && static_cast<int>(nodep->widthConst()) == extendp->lhsp()->width()))
            return false;
        VL_DO_DANGLING(replaceWChild(nodep, extendp->lhsp()), nodep);
        return true;
    }
    bool operandSelBiLower(AstSel* nodep) {
        // SEL(ADD(a,b),(width-1),0) -> ADD(SEL(a),SEL(b))
        // Add or any operation which doesn't care if we discard top bits
        AstNodeBiop* bip = VN_CAST(nodep->fromp(), NodeBiop);
        if (!(m_doV && bip && VN_IS(nodep->lsbp(), Const) && VN_IS(nodep->widthp(), Const)
              && nodep->lsbConst() == 0))
            return false;
        if (debug() >= 9) nodep->dumpTree(cout, "SEL(BI)-in:");
        AstNode* bilhsp = bip->lhsp()->unlinkFrBack();
        AstNode* birhsp = bip->rhsp()->unlinkFrBack();
        bip->lhsp(new AstSel(nodep->fileline(), bilhsp, 0, nodep->widthConst()));
        bip->rhsp(new AstSel(nodep->fileline(), birhsp, 0, nodep->widthConst()));
        if (debug() >= 9) bip->dumpTree(cout, "SEL(BI)-ou:");
        VL_DO_DANGLING(replaceWChild(nodep, bip), nodep);
        return true;
    }
    bool operandSelShiftLower(AstSel* nodep) {
        // AND({a}, SHIFTR({b}, {c})) is often shorthand in C for Verilog {b}[{c} :+ {a}]
        // becomes thought other optimizations
        // SEL(SHIFTR({a},{b}),{lsb},{width}) -> SEL({a},{lsb+b},{width})
        AstShiftR* shiftp = VN_CAST(nodep->fromp(), ShiftR);
        if (!(m_doV && shiftp && VN_IS(shiftp->rhsp(), Const) && VN_IS(nodep->lsbp(), Const)
              && VN_IS(nodep->widthp(), Const))) {
            return false;
        }
        AstNode* ap = shiftp->lhsp();
        AstConst* bp = VN_CAST(shiftp->rhsp(), Const);
        AstConst* lp = VN_CAST(nodep->lsbp(), Const);
        if (bp->isWide() || bp->num().isFourState() || bp->num().isNegative() || lp->isWide()
            || lp->num().isFourState() || lp->num().isNegative()) {
            return false;
        }
        int newLsb = lp->toSInt() + bp->toSInt();
        if (newLsb + nodep->widthConst() > ap->width()) return false;
        //
        UINFO(9, "SEL(SHIFTR(a,b),l,w) -> SEL(a,l+b,w)\n");
        if (debug() >= 9) nodep->dumpTree(cout, "SEL(SH)-in:");
        AstSel* newp
            = new AstSel(nodep->fileline(), ap->unlinkFrBack(), newLsb, nodep->widthConst());
        newp->dtypeFrom(nodep);
        if (debug() >= 9) newp->dumpTree(cout, "SEL(SH)-ou:");
        VL_DO_DANGLING(nodep->replaceWith(newp), nodep);
        return true;
    }

    bool operandBiExtendConstShrink(AstNodeBiop* nodep) {
        // Loop unrolling favors standalone compares
        // EQ(const{width32}, EXTEND(xx{width3})) -> EQ(const{3}, xx{3})
        // The constant must have zero bits (+ 1 if signed) or compare
        // would be incorrect. See also operandBiExtendConst
        AstExtend* extendp = VN_CAST(nodep->rhsp(), Extend);
        if (!extendp) return false;
        AstNode* smallerp = extendp->lhsp();
        int subsize = smallerp->width();
        AstConst* constp = VN_CAST(nodep->lhsp(), Const);
        if (!constp) return false;
        if (!constp->num().isBitsZero(constp->width() - 1, subsize)) return false;
        //
        if (debug() >= 9) nodep->dumpTree(cout, "BI(EXTEND)-in:");
        smallerp->unlinkFrBack();
        VL_DO_DANGLING(extendp->unlinkFrBack()->deleteTree(), extendp);  // aka nodep->lhsp.
        nodep->rhsp(smallerp);

        constp->unlinkFrBack();
        V3Number num(constp, subsize);
        num.opAssign(constp->num());
        nodep->lhsp(new AstConst(constp->fileline(), num));
        VL_DO_DANGLING(constp->deleteTree(), constp);
        if (debug() >= 9) nodep->dumpTree(cout, "BI(EXTEND)-ou:");
        return true;
    }
    bool operandBiExtendConstOver(const AstNodeBiop* nodep) {
        // EQ(const{width32}, EXTEND(xx{width3})) -> constant
        // When the constant has non-zero bits above the extend it's a constant.
        // Avoids compiler warning
        const AstExtend* extendp = VN_CAST_CONST(nodep->rhsp(), Extend);
        if (!extendp) return false;
        AstNode* smallerp = extendp->lhsp();
        int subsize = smallerp->width();
        const AstConst* constp = VN_CAST_CONST(nodep->lhsp(), Const);
        if (!constp) return false;
        if (constp->num().isBitsZero(constp->width() - 1, subsize)) return false;
        return true;
    }

    AstNode* afterComment(AstNode* nodep) {
        // Ignore comments, such as to determine if a AstIf is empty.
        // nodep may be null, if so return null.
        while (nodep && VN_IS(nodep, Comment)) { nodep = nodep->nextp(); }
        return nodep;
    }

    bool isTPure(AstNode* nodep) {
        // Pure checks - if this node and all nodes under it are free of
        // side effects can do this optimization
        // Eventually we'll recurse through tree when unknown, memoizing results so far,
        // but for now can disable en-mass until V3Purify takes effect.
        return m_doShort || VN_IS(nodep, VarRef) || VN_IS(nodep, Const);
    }

    // Extraction checks
    bool warnSelect(AstSel* nodep) {
        if (m_doGenerate) {
            // Never checked yet
            V3Width::widthParamsEdit(nodep);
            iterateChildren(nodep);  // May need "constifying"
        }
        // Find range of dtype we are selecting from
        // Similar code in V3Unknown::AstSel
        bool doit = true;
        if (m_warn && VN_IS(nodep->lsbp(), Const) && VN_IS(nodep->widthp(), Const) && doit) {
            int maxDeclBit = nodep->declRange().hiMaxSelect() * nodep->declElWidth()
                             + (nodep->declElWidth() - 1);
            if (VN_CAST(nodep->lsbp(), Const)->num().isFourState()
                || VN_CAST(nodep->widthp(), Const)->num().isFourState()) {
                nodep->v3error("Selection index is constantly unknown or tristated: "
                               "lsb="
                               << nodep->lsbp()->name() << " width=" << nodep->widthp()->name());
                // Replacing nodep will make a mess above, so we replace the offender
                replaceZero(nodep->lsbp());
            } else if (nodep->declRange().ranged()
                       && (nodep->msbConst() > maxDeclBit || nodep->lsbConst() > maxDeclBit)) {
                // See also warning in V3Width
                // Must adjust by element width as declRange() is in number of elements
                nodep->v3warn(SELRANGE,
                              "Selection index out of range: "
                                  << (nodep->msbConst() / nodep->declElWidth()) << ":"
                                  << (nodep->lsbConst() / nodep->declElWidth()) << " outside "
                                  << nodep->declRange().hiMaxSelect() << ":0"
                                  << (nodep->declRange().lo() >= 0
                                          ? ""
                                          : (" (adjusted +" + cvtToStr(-nodep->declRange().lo())
                                             + " to account for negative lsb)")));
                UINFO(1, "    Related Raw index is " << nodep->msbConst() << ":"
                                                     << nodep->lsbConst() << endl);
                // Don't replace with zero, we'll do it later
            }
        }
        return false;  // Not a transform, so NOP
    }

    static bool operandsSame(AstNode* node1p, AstNode* node2p) {
        // For now we just detect constants & simple vars, though it could be more generic
        if (VN_IS(node1p, Const) && VN_IS(node2p, Const)) {
            return node1p->sameGateTree(node2p);
        } else if (VN_IS(node1p, VarRef) && VN_IS(node2p, VarRef)) {
            // Avoid comparing widthMin's, which results in lost optimization attempts
            // If cleanup sameGateTree to be smarter, this can be restored.
            // return node1p->sameGateTree(node2p);
            return node1p->same(node2p);
        } else {
            return false;
        }
    }
    bool ifSameAssign(const AstNodeIf* nodep) {
        const AstNodeAssign* ifp = VN_CAST_CONST(nodep->ifsp(), NodeAssign);
        const AstNodeAssign* elsep = VN_CAST_CONST(nodep->elsesp(), NodeAssign);
        if (!ifp || ifp->nextp()) return false;  // Must be SINGLE statement
        if (!elsep || elsep->nextp()) return false;
        if (ifp->type() != elsep->type()) return false;  // Can't mix an assigndly and an assign
        if (!ifp->lhsp()->sameGateTree(elsep->lhsp())) return false;
        if (!ifp->rhsp()->gateTree()) return false;
        if (!elsep->rhsp()->gateTree()) return false;
        return true;
    }
    bool operandIfIf(const AstNodeIf* nodep) {
        if (nodep->elsesp()) return false;
        const AstNodeIf* lowerIfp = VN_CAST_CONST(nodep->ifsp(), NodeIf);
        if (!lowerIfp || lowerIfp->nextp()) return false;
        if (nodep->type() != lowerIfp->type()) return false;
        if (afterComment(lowerIfp->elsesp())) return false;
        return true;
    }
    bool ifConcatMergeableBiop(const AstNode* nodep) {
        return (VN_IS(nodep, And) || VN_IS(nodep, Or) || VN_IS(nodep, Xor));
    }
    bool ifAdjacentSel(const AstSel* lhsp, const AstSel* rhsp) {
        if (!v3Global.opt.oAssemble()) return false;  // opt disabled
        if (!lhsp || !rhsp) return false;
        const AstNode* lfromp = lhsp->fromp();
        const AstNode* rfromp = rhsp->fromp();
        if (!lfromp || !rfromp || !lfromp->sameGateTree(rfromp)) return false;
        const AstConst* lstart = VN_CAST_CONST(lhsp->lsbp(), Const);
        const AstConst* rstart = VN_CAST_CONST(rhsp->lsbp(), Const);
        const AstConst* lwidth = VN_CAST_CONST(lhsp->widthp(), Const);
        const AstConst* rwidth = VN_CAST_CONST(rhsp->widthp(), Const);
        if (!lstart || !rstart || !lwidth || !rwidth) return false;  // too complicated
        int rend = (rstart->toSInt() + rwidth->toSInt());
        return (rend == lstart->toSInt());
    }
    bool ifMergeAdjacent(AstNode* lhsp, AstNode* rhsp) {
        // called by concatmergeable to determine if {lhsp, rhsp} make sense
        if (!v3Global.opt.oAssemble()) return false;  // opt disabled
        // two same varref
        if (operandsSame(lhsp, rhsp)) return true;
        AstSel* lselp = VN_CAST(lhsp, Sel);
        AstSel* rselp = VN_CAST(rhsp, Sel);
        // a[i:0] a
        if (lselp && !rselp && rhsp->sameGateTree(lselp->fromp()))
            rselp = new AstSel(rhsp->fileline(), rhsp->cloneTree(false), 0, rhsp->width());
        // a[i:j] {a[j-1:k], b}
        if (lselp && !rselp && VN_IS(rhsp, Concat))
            return ifMergeAdjacent(lhsp, VN_CAST(rhsp, Concat)->lhsp());
        // a a[msb:j]
        if (rselp && !lselp && lhsp->sameGateTree(rselp->fromp()))
            lselp = new AstSel(lhsp->fileline(), lhsp->cloneTree(false), 0, lhsp->width());
        // {b, a[j:k]} a[k-1:i]
        if (rselp && !lselp && VN_IS(lhsp, Concat))
            return ifMergeAdjacent(VN_CAST(lhsp, Concat)->rhsp(), rhsp);
        if (!lselp || !rselp) return false;

        // a[a:b] a[b-1:c] are adjacent
        AstNode* lfromp = lselp->fromp();
        AstNode* rfromp = rselp->fromp();
        if (!lfromp || !rfromp || !lfromp->sameGateTree(rfromp)) return false;
        AstConst* lstart = VN_CAST(lselp->lsbp(), Const);
        AstConst* rstart = VN_CAST(rselp->lsbp(), Const);
        AstConst* lwidth = VN_CAST(lselp->widthp(), Const);
        AstConst* rwidth = VN_CAST(rselp->widthp(), Const);
        if (!lstart || !rstart || !lwidth || !rwidth) return false;  // too complicated
        int rend = (rstart->toSInt() + rwidth->toSInt());
        // a[i:j] a[j-1:k]
        if (rend == lstart->toSInt()) return true;
        // a[i:0] a[msb:j]
        if (rend == rfromp->width() && lstart->toSInt() == 0) return true;
        return false;
    }
    bool concatMergeable(const AstNode* lhsp, const AstNode* rhsp) {
        // determine if {a OP b, c OP d} => {a, c} OP {b, d} is advantageous
        if (!v3Global.opt.oAssemble()) return false;  // opt disabled
        if (lhsp->type() != rhsp->type()) return false;
        if (!ifConcatMergeableBiop(lhsp)) return false;

        const AstNodeBiop* lp = VN_CAST_CONST(lhsp, NodeBiop);
        const AstNodeBiop* rp = VN_CAST_CONST(rhsp, NodeBiop);
        if (!lp || !rp) return false;
        // {a[]&b[], a[]&b[]}
        bool lad = ifMergeAdjacent(lp->lhsp(), rp->lhsp());
        bool rad = ifMergeAdjacent(lp->rhsp(), rp->rhsp());
        if (lad && rad) return true;
        // {a[] & b[]&c[], a[] & b[]&c[]}
        if (lad && concatMergeable(lp->rhsp(), rp->rhsp())) return true;
        // {a[]&b[] & c[], a[]&b[] & c[]}
        if (rad && concatMergeable(lp->lhsp(), rp->lhsp())) return true;
        // {(a[]&b[])&(c[]&d[]), (a[]&b[])&(c[]&d[])}
        if (concatMergeable(lp->lhsp(), rp->lhsp()) && concatMergeable(lp->rhsp(), rp->rhsp())) {
            return true;
        }
        return false;
    }
    bool operandsSameSize(AstNode* lhsp, AstNode* rhsp) { return lhsp->width() == rhsp->width(); }

    //----------------------------------------
    // Constant Replacement functions.
    // These all take a node, delete its tree, and replaces it with a constant

    void replaceNum(AstNode* oldp, const V3Number& num) {
        // Replace oldp node with a constant set to specified value
        UASSERT(oldp, "Null old");
        UASSERT_OBJ(!(VN_IS(oldp, Const) && !VN_CAST(oldp, Const)->num().isFourState()), oldp,
                    "Already constant??");
        AstNode* newp = new AstConst(oldp->fileline(), num);
        newp->dtypeFrom(oldp);
        if (debug() > 5) oldp->dumpTree(cout, "  const_old: ");
        if (debug() > 5) newp->dumpTree(cout, "       _new: ");
        oldp->replaceWith(newp);
        VL_DO_DANGLING(oldp->deleteTree(), oldp);
    }
    void replaceNum(AstNode* nodep, uint32_t val) {
        V3Number num(nodep, nodep->width(), val);
        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
    }
    void replaceNumSigned(AstNodeBiop* nodep, uint32_t val) {
        // We allow both sides to be constant, as one may have come from
        // parameter propagation, etc.
        if (m_warn && !(VN_IS(nodep->lhsp(), Const) && VN_IS(nodep->rhsp(), Const))) {
            nodep->v3warn(UNSIGNED, "Comparison is constant due to unsigned arithmetic");
        }
        VL_DO_DANGLING(replaceNum(nodep, val), nodep);
    }
    void replaceNumLimited(AstNodeBiop* nodep, uint32_t val) {
        // Avoids gcc warning about same
        if (m_warn) nodep->v3warn(CMPCONST, "Comparison is constant due to limited range");
        VL_DO_DANGLING(replaceNum(nodep, val), nodep);
    }
    void replaceZero(AstNode* nodep) { VL_DO_DANGLING(replaceNum(nodep, 0), nodep); }
    void replaceZeroChkPure(AstNode* nodep, AstNode* checkp) {
        // For example, "0 * n" -> 0 if n has no side effects
        // Else strength reduce it to 0 & n.
        // If ever change the operation note AstAnd rule specially ignores this created pattern
        if (isTPure(checkp)) {
            VL_DO_DANGLING(replaceNum(nodep, 0), nodep);
        } else {
            AstNode* newp = new AstAnd(nodep->fileline(), new AstConst(nodep->fileline(), 0),
                                       checkp->unlinkFrBack());
            newp->dtypeFrom(nodep);
            nodep->replaceWith(newp);
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
        }
    }
    void replaceAllOnes(AstNode* nodep) {
        V3Number ones(nodep, nodep->width(), 0);
        ones.setMask(nodep->width());
        VL_DO_DANGLING(replaceNum(nodep, ones), nodep);
    }
    void replaceConst(AstNodeUniop* nodep) {
        V3Number num(nodep, nodep->width());
        nodep->numberOperate(num, VN_CAST(nodep->lhsp(), Const)->num());
        UINFO(4, "UNICONST -> " << num << endl);
        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
    }
    void replaceConst(AstNodeBiop* nodep) {
        V3Number num(nodep, nodep->width());
        nodep->numberOperate(num, VN_CAST(nodep->lhsp(), Const)->num(),
                             VN_CAST(nodep->rhsp(), Const)->num());
        UINFO(4, "BICONST -> " << num << endl);
        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
    }
    void replaceConst(AstNodeTriop* nodep) {
        V3Number num(nodep, nodep->width());
        nodep->numberOperate(num, VN_CAST(nodep->lhsp(), Const)->num(),
                             VN_CAST(nodep->rhsp(), Const)->num(),
                             VN_CAST(nodep->thsp(), Const)->num());
        UINFO(4, "TRICONST -> " << num << endl);
        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
    }
    void replaceConst(AstNodeQuadop* nodep) {
        V3Number num(nodep, nodep->width());
        nodep->numberOperate(
            num, VN_CAST(nodep->lhsp(), Const)->num(), VN_CAST(nodep->rhsp(), Const)->num(),
            VN_CAST(nodep->thsp(), Const)->num(), VN_CAST(nodep->fhsp(), Const)->num());
        UINFO(4, "QUADCONST -> " << num << endl);
        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
    }

    void replaceConstString(AstNode* oldp, const string& num) {
        // Replace oldp node with a constant set to specified value
        UASSERT(oldp, "Null old");
        AstNode* newp = new AstConst(oldp->fileline(), AstConst::String(), num);
        if (debug() > 5) oldp->dumpTree(cout, "  const_old: ");
        if (debug() > 5) newp->dumpTree(cout, "       _new: ");
        oldp->replaceWith(newp);
        VL_DO_DANGLING(oldp->deleteTree(), oldp);
    }
    //----------------------------------------
    // Replacement functions.
    // These all take a node and replace it with something else

    void replaceWChild(AstNode* nodep, AstNode* childp) {
        // NODE(..., CHILD(...)) -> CHILD(...)
        childp->unlinkFrBackWithNext();
        // If replacing a SEL for example, the data type comes from the parent (is less wide).
        // This may adversely affect the operation of the node being replaced.
        childp->dtypeFrom(nodep);
        nodep->replaceWith(childp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }

    //! Replace a ternary node with its RHS after iterating
    //! Used with short-circuiting, where the RHS has not yet been iterated.
    void replaceWIteratedRhs(AstNodeTriop* nodep) {
        if (AstNode* rhsp = nodep->rhsp()) iterateAndNextNull(rhsp);
        replaceWChild(nodep, nodep->rhsp());  // May have changed
    }

    //! Replace a ternary node with its THS after iterating
    //! Used with short-circuiting, where the THS has not yet been iterated.
    void replaceWIteratedThs(AstNodeTriop* nodep) {
        if (AstNode* thsp = nodep->thsp()) iterateAndNextNull(thsp);
        replaceWChild(nodep, nodep->thsp());  // May have changed
    }
    void replaceWLhs(AstNodeUniop* nodep) {
        // Keep LHS, remove RHS
        replaceWChild(nodep, nodep->lhsp());
    }
    void replaceWLhs(AstNodeBiop* nodep) {
        // Keep LHS, remove RHS
        replaceWChild(nodep, nodep->lhsp());
    }
    void replaceWRhs(AstNodeBiop* nodep) {
        // Keep RHS, remove LHS
        replaceWChild(nodep, nodep->rhsp());
    }
    void replaceAsv(AstNodeBiop* nodep) {
        // BIASV(CONSTa, BIASV(CONSTb, c)) -> BIASV( BIASV_CONSTED(a,b), c)
        // BIASV(SAMEa,  BIASV(SAMEb, c))  -> BIASV( BIASV(SAMEa,SAMEb), c)
        // nodep->dumpTree(cout, "  repAsvConst_old: ");
        AstNode* ap = nodep->lhsp();
        AstNodeBiop* rp = VN_CAST(nodep->rhsp(), NodeBiop);
        AstNode* bp = rp->lhsp();
        AstNode* cp = rp->rhsp();
        ap->unlinkFrBack();
        bp->unlinkFrBack();
        cp->unlinkFrBack();
        rp->unlinkFrBack();
        nodep->lhsp(rp);
        nodep->rhsp(cp);
        rp->lhsp(ap);
        rp->rhsp(bp);
        if (VN_IS(rp->lhsp(), Const) && VN_IS(rp->rhsp(), Const)) replaceConst(rp);
        // nodep->dumpTree(cout, "  repAsvConst_new: ");
    }
    void replaceAsvLUp(AstNodeBiop* nodep) {
        // BIASV(BIASV(CONSTll,lr),r) -> BIASV(CONSTll,BIASV(lr,r))
        AstNodeBiop* lp = VN_CAST(nodep->lhsp()->unlinkFrBack(), NodeBiop);
        AstNode* llp = lp->lhsp()->unlinkFrBack();
        AstNode* lrp = lp->rhsp()->unlinkFrBack();
        AstNode* rp = nodep->rhsp()->unlinkFrBack();
        nodep->lhsp(llp);
        nodep->rhsp(lp);
        lp->lhsp(lrp);
        lp->rhsp(rp);
        // nodep->dumpTree(cout, "  repAsvLUp_new: ");
    }
    void replaceAsvRUp(AstNodeBiop* nodep) {
        // BIASV(l,BIASV(CONSTrl,rr)) -> BIASV(CONSTrl,BIASV(l,rr))
        AstNode* lp = nodep->lhsp()->unlinkFrBack();
        AstNodeBiop* rp = VN_CAST(nodep->rhsp()->unlinkFrBack(), NodeBiop);
        AstNode* rlp = rp->lhsp()->unlinkFrBack();
        AstNode* rrp = rp->rhsp()->unlinkFrBack();
        nodep->lhsp(rlp);
        nodep->rhsp(rp);
        rp->lhsp(lp);
        rp->rhsp(rrp);
        // nodep->dumpTree(cout, "  repAsvRUp_new: ");
    }
    void replaceAndOr(AstNodeBiop* nodep) {
        //  OR  (AND (CONSTll,lr), AND(CONSTrl==ll,rr))    -> AND (CONSTll, OR(lr,rr))
        //  OR  (AND (CONSTll,lr), AND(CONSTrl,    rr=lr)) -> AND (OR(ll,rl), rr)
        // nodep ^lp  ^llp   ^lrp  ^rp  ^rlp       ^rrp
        // (Or/And may also be reversed)
        AstNodeBiop* lp = VN_CAST(nodep->lhsp()->unlinkFrBack(), NodeBiop);
        AstNode* llp = lp->lhsp()->unlinkFrBack();
        AstNode* lrp = lp->rhsp()->unlinkFrBack();
        AstNodeBiop* rp = VN_CAST(nodep->rhsp()->unlinkFrBack(), NodeBiop);
        AstNode* rlp = rp->lhsp()->unlinkFrBack();
        AstNode* rrp = rp->rhsp()->unlinkFrBack();
        nodep->replaceWith(lp);
        if (operandsSame(llp, rlp)) {
            lp->lhsp(llp);
            lp->rhsp(nodep);
            nodep->lhsp(lrp);
            nodep->rhsp(rrp);
            VL_DO_DANGLING(rp->deleteTree(), rp);
            VL_DO_DANGLING(rlp->deleteTree(), rlp);
        } else if (operandsSame(lrp, rrp)) {
            lp->lhsp(nodep);
            lp->rhsp(rrp);
            nodep->lhsp(llp);
            nodep->rhsp(rlp);
            VL_DO_DANGLING(rp->deleteTree(), rp);
            VL_DO_DANGLING(lrp->deleteTree(), lrp);
        } else {
            nodep->v3fatalSrc("replaceAndOr on something operandAndOrSame shouldn't have matched");
        }
        // nodep->dumpTree(cout, "  repAndOr_new: ");
    }
    void replaceShiftSame(AstNodeBiop* nodep) {
        // Or(Shift(ll,CONSTlr),Shift(rl,CONSTrr==lr)) -> Shift(Or(ll,rl),CONSTlr)
        // (Or/And may also be reversed)
        AstNodeBiop* lp = VN_CAST(nodep->lhsp()->unlinkFrBack(), NodeBiop);
        AstNode* llp = lp->lhsp()->unlinkFrBack();
        AstNode* lrp = lp->rhsp()->unlinkFrBack();
        AstNodeBiop* rp = VN_CAST(nodep->rhsp()->unlinkFrBack(), NodeBiop);
        AstNode* rlp = rp->lhsp()->unlinkFrBack();
        AstNode* rrp = rp->rhsp()->unlinkFrBack();
        nodep->replaceWith(lp);
        lp->lhsp(nodep);
        lp->rhsp(lrp);
        nodep->lhsp(llp);
        nodep->rhsp(rlp);
        VL_DO_DANGLING(rp->deleteTree(), rp);
        VL_DO_DANGLING(rrp->deleteTree(), rrp);
        // nodep->dumpTree(cout, "  repShiftSame_new: ");
    }
    void replaceConcatSel(AstConcat* nodep) {
        // {a[1], a[0]} -> a[1:0]
        AstSel* lselp = VN_CAST(nodep->lhsp()->unlinkFrBack(), Sel);
        AstSel* rselp = VN_CAST(nodep->rhsp()->unlinkFrBack(), Sel);
        int lstart = lselp->lsbConst();
        int lwidth = lselp->widthConst();
        int rstart = rselp->lsbConst();
        int rwidth = rselp->widthConst();

        UASSERT_OBJ((rstart + rwidth) == lstart, nodep,
                    "tried to merge two selects which are not adjacent");
        AstSel* newselp = new AstSel(lselp->fromp()->fileline(), rselp->fromp()->cloneTree(false),
                                     rstart, lwidth + rwidth);
        UINFO(5, "merged two adjacent sel " << lselp << " and " << rselp << " to one " << newselp
                                            << endl);

        nodep->replaceWith(newselp);
        VL_DO_DANGLING(lselp->deleteTree(), lselp);
        VL_DO_DANGLING(rselp->deleteTree(), rselp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceConcatMerge(AstConcat* nodep) {
        AstNodeBiop* lp = VN_CAST(nodep->lhsp(), NodeBiop);
        AstNodeBiop* rp = VN_CAST(nodep->rhsp(), NodeBiop);
        AstNode* llp = lp->lhsp()->cloneTree(false);
        AstNode* lrp = lp->rhsp()->cloneTree(false);
        AstNode* rlp = rp->lhsp()->cloneTree(false);
        AstNode* rrp = rp->rhsp()->cloneTree(false);
        if (concatMergeable(lp, rp)) {
            AstConcat* newlp = new AstConcat(rlp->fileline(), llp, rlp);
            AstConcat* newrp = new AstConcat(rrp->fileline(), lrp, rrp);
            // use the lhs to replace the parent concat
            lp->lhsp()->replaceWith(newlp);
            lp->rhsp()->replaceWith(newrp);
            lp->dtypeChgWidthSigned(newlp->width(), newlp->width(), VSigning::UNSIGNED);
            UINFO(5, "merged " << nodep << endl);
            VL_DO_DANGLING(rp->unlinkFrBack()->deleteTree(), rp);
            nodep->replaceWith(lp->unlinkFrBack());
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
            iterate(lp->lhsp());
            iterate(lp->rhsp());
        } else {
            nodep->v3fatalSrc("tried to merge two Concat which are not adjacent");
        }
    }
    void replaceExtend(AstNode* nodep, AstNode* arg0p) {
        // -> EXTEND(nodep)
        // like a AstExtend{$rhsp}, but we need to set the width correctly from base node
        arg0p->unlinkFrBack();
        AstNode* newp = (VN_IS(nodep, ExtendS)
                             ? static_cast<AstNode*>(new AstExtendS(nodep->fileline(), arg0p))
                             : static_cast<AstNode*>(new AstExtend(nodep->fileline(), arg0p)));
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replacePowShift(AstNodeBiop* nodep) {  // Pow or PowS
        UINFO(5, "POW(2,b)->SHIFTL(1,b) " << nodep << endl);
        AstNode* rhsp = nodep->rhsp()->unlinkFrBack();
        AstShiftL* newp
            = new AstShiftL(nodep->fileline(), new AstConst(nodep->fileline(), 1), rhsp);
        newp->dtypeFrom(nodep);
        newp->lhsp()->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceMulShift(AstMul* nodep) {  // Mul, but not MulS as not simple shift
        UINFO(5, "MUL(2^n,b)->SHIFTL(b,n) " << nodep << endl);
        int amount = VN_CAST(nodep->lhsp(), Const)->num().mostSetBitP1() - 1;  // 2^n->n+1
        AstNode* opp = nodep->rhsp()->unlinkFrBack();
        AstShiftL* newp
            = new AstShiftL(nodep->fileline(), opp, new AstConst(nodep->fileline(), amount));
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceDivShift(AstDiv* nodep) {  // Mul, but not MulS as not simple shift
        UINFO(5, "DIV(b,2^n)->SHIFTR(b,n) " << nodep << endl);
        int amount = VN_CAST(nodep->rhsp(), Const)->num().mostSetBitP1() - 1;  // 2^n->n+1
        AstNode* opp = nodep->lhsp()->unlinkFrBack();
        AstShiftR* newp
            = new AstShiftR(nodep->fileline(), opp, new AstConst(nodep->fileline(), amount));
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceModAnd(AstModDiv* nodep) {  // Mod, but not ModS as not simple shift
        UINFO(5, "MOD(b,2^n)->AND(b,2^n-1) " << nodep << endl);
        int amount = VN_CAST(nodep->rhsp(), Const)->num().mostSetBitP1() - 1;  // 2^n->n+1
        V3Number mask(nodep, nodep->width());
        mask.setMask(amount);
        AstNode* opp = nodep->lhsp()->unlinkFrBack();
        AstAnd* newp = new AstAnd(nodep->fileline(), opp, new AstConst(nodep->fileline(), mask));
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceShiftOp(AstNodeBiop* nodep) {
        UINFO(5, "SHIFT(AND(a,b),CONST)->AND(SHIFT(a,CONST),SHIFT(b,CONST)) " << nodep << endl);
        AstNRelinker handle;
        nodep->unlinkFrBack(&handle);
        AstNodeBiop* lhsp = VN_CAST(nodep->lhsp(), NodeBiop);
        lhsp->unlinkFrBack();
        AstNode* shiftp = nodep->rhsp()->unlinkFrBack();
        AstNode* ap = lhsp->lhsp()->unlinkFrBack();
        AstNode* bp = lhsp->rhsp()->unlinkFrBack();
        AstNodeBiop* shift1p = nodep;
        AstNodeBiop* shift2p = nodep->cloneTree(true);
        shift1p->lhsp(ap);
        shift1p->rhsp(shiftp->cloneTree(true));
        shift2p->lhsp(bp);
        shift2p->rhsp(shiftp);
        AstNodeBiop* newp = lhsp;
        newp->lhsp(shift1p);
        newp->rhsp(shift2p);
        handle.relink(newp);
        iterate(newp);  // Further reduce, either node may have more reductions.
    }
    void replaceShiftShift(AstNodeBiop* nodep) {
        UINFO(4, "SHIFT(SHIFT(a,s1),s2)->SHIFT(a,ADD(s1,s2)) " << nodep << endl);
        if (debug() >= 9) nodep->dumpTree(cout, "  repShiftShift_old: ");
        AstNodeBiop* lhsp = VN_CAST(nodep->lhsp(), NodeBiop);
        lhsp->unlinkFrBack();
        AstNode* ap = lhsp->lhsp()->unlinkFrBack();
        AstNode* shift1p = lhsp->rhsp()->unlinkFrBack();
        AstNode* shift2p = nodep->rhsp()->unlinkFrBack();
        // Shift1p and shift2p may have different sizes, both are
        // self-determined so sum with infinite width
        if (nodep->type() == lhsp->type()) {
            int shift1 = VN_CAST(shift1p, Const)->toUInt();
            int shift2 = VN_CAST(shift2p, Const)->toUInt();
            int newshift = shift1 + shift2;
            VL_DO_DANGLING(shift1p->deleteTree(), shift1p);
            VL_DO_DANGLING(shift2p->deleteTree(), shift2p);
            nodep->lhsp(ap);
            nodep->rhsp(new AstConst(nodep->fileline(), newshift));
            iterate(nodep);  // Further reduce, either node may have more reductions.
        } else {
            // We know shift amounts are constant, but might be a mixed left/right shift
            int shift1 = VN_CAST(shift1p, Const)->toUInt();
            if (VN_IS(lhsp, ShiftR)) shift1 = -shift1;
            int shift2 = VN_CAST(shift2p, Const)->toUInt();
            if (VN_IS(nodep, ShiftR)) shift2 = -shift2;
            int newshift = shift1 + shift2;
            VL_DO_DANGLING(shift1p->deleteTree(), shift1p);
            VL_DO_DANGLING(shift2p->deleteTree(), shift2p);
            AstNode* newp;
            V3Number mask1(nodep, nodep->width());
            V3Number ones(nodep, nodep->width());
            ones.setMask(nodep->width());
            if (shift1 < 0) {
                mask1.opShiftR(ones, V3Number(nodep, VL_IDATASIZE, -shift1));
            } else {
                mask1.opShiftL(ones, V3Number(nodep, VL_IDATASIZE, shift1));
            }
            V3Number mask(nodep, nodep->width());
            if (shift2 < 0) {
                mask.opShiftR(mask1, V3Number(nodep, VL_IDATASIZE, -shift2));
            } else {
                mask.opShiftL(mask1, V3Number(nodep, VL_IDATASIZE, shift2));
            }
            if (newshift < 0) {
                newp = new AstShiftR(nodep->fileline(), ap,
                                     new AstConst(nodep->fileline(), -newshift));
            } else {
                newp = new AstShiftL(nodep->fileline(), ap,
                                     new AstConst(nodep->fileline(), newshift));
            }
            newp->dtypeFrom(nodep);
            newp = new AstAnd(nodep->fileline(), newp, new AstConst(nodep->fileline(), mask));
            newp->dtypeFrom(nodep);
            nodep->replaceWith(newp);
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
            // newp->dumpTree(cout, "  repShiftShift_new: ");
            iterate(newp);  // Further reduce, either node may have more reductions.
        }
        VL_DO_DANGLING(lhsp->deleteTree(), lhsp);
    }

    bool replaceAssignMultiSel(AstNodeAssign* nodep) {
        // Multiple assignments to sequential bits can be concated
        // ASSIGN(SEL(a),aq), ASSIGN(SEL(a+1),bq) -> ASSIGN(SEL(a:b),CONCAT(aq,bq)
        // ie. assign var[2]=a, assign var[3]=b -> assign var[3:2]={b,a}

        // Skip if we're not const'ing an entire module (IE doing only one assign, etc)
        if (!m_modp) return false;
        AstSel* sel1p = VN_CAST(nodep->lhsp(), Sel);
        if (!sel1p) return false;
        AstNodeAssign* nextp = VN_CAST(nodep->nextp(), NodeAssign);
        if (!nextp) return false;
        if (nodep->type() != nextp->type()) return false;
        AstSel* sel2p = VN_CAST(nextp->lhsp(), Sel);
        if (!sel2p) return false;
        AstVarRef* varref1p = VN_CAST(sel1p->fromp(), VarRef);
        if (!varref1p) return false;
        AstVarRef* varref2p = VN_CAST(sel2p->fromp(), VarRef);
        if (!varref2p) return false;
        if (!varref1p->sameGateTree(varref2p)) return false;
        AstConst* con1p = VN_CAST(sel1p->lsbp(), Const);
        if (!con1p) return false;
        AstConst* con2p = VN_CAST(sel2p->lsbp(), Const);
        if (!con2p) return false;
        // We need to make sure there's no self-references involved in either
        // assignment.  For speed, we only look 3 deep, then give up.
        if (!varNotReferenced(nodep->rhsp(), varref1p->varp())) return false;
        if (!varNotReferenced(nextp->rhsp(), varref2p->varp())) return false;
        // Swap?
        if ((con1p->toSInt() != con2p->toSInt() + sel2p->width())
            && (con2p->toSInt() != con1p->toSInt() + sel1p->width())) {
            return false;
        }
        bool lsbFirstAssign = (con1p->toUInt() < con2p->toUInt());
        UINFO(4, "replaceAssignMultiSel " << nodep << endl);
        UINFO(4, "                   && " << nextp << endl);
        // nodep->dumpTree(cout, "comb1: ");
        // nextp->dumpTree(cout, "comb2: ");
        AstNode* rhs1p = nodep->rhsp()->unlinkFrBack();
        AstNode* rhs2p = nextp->rhsp()->unlinkFrBack();
        AstNode* newp;
        if (lsbFirstAssign) {
            newp = nodep->cloneType(new AstSel(sel1p->fileline(), varref1p->unlinkFrBack(),
                                               sel1p->lsbConst(), sel1p->width() + sel2p->width()),
                                    new AstConcat(rhs1p->fileline(), rhs2p, rhs1p));
        } else {
            newp = nodep->cloneType(new AstSel(sel1p->fileline(), varref1p->unlinkFrBack(),
                                               sel2p->lsbConst(), sel1p->width() + sel2p->width()),
                                    new AstConcat(rhs1p->fileline(), rhs1p, rhs2p));
        }
        // pnewp->dumpTree(cout, "conew: ");
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        VL_DO_DANGLING(nextp->unlinkFrBack()->deleteTree(), nextp);
        return true;
    }

    bool varNotReferenced(AstNode* nodep, AstVar* varp, int level = 0) {
        // Return true if varp never referenced under node.
        // Return false if referenced, or tree too deep to be worth it, or side effects
        if (!nodep) return true;
        if (level > 2) return false;
        if (!nodep->isPure()) return false;  // For example a $fgetc can't be reordered
        if (VN_IS(nodep, NodeVarRef) && VN_CAST(nodep, NodeVarRef)->varp() == varp) return false;
        return (varNotReferenced(nodep->nextp(), varp, level + 1)
                && varNotReferenced(nodep->op1p(), varp, level + 1)
                && varNotReferenced(nodep->op2p(), varp, level + 1)
                && varNotReferenced(nodep->op3p(), varp, level + 1)
                && varNotReferenced(nodep->op4p(), varp, level + 1));
    }

    bool replaceNodeAssign(AstNodeAssign* nodep) {
        if (VN_IS(nodep->lhsp(), VarRef) && VN_IS(nodep->rhsp(), VarRef)
            && VN_CAST(nodep->lhsp(), VarRef)->sameNoLvalue(VN_CAST(nodep->rhsp(), VarRef))
            && !VN_IS(nodep, AssignDly)) {
            // X = X.  Quite pointless, though X <= X may override another earlier assignment
            if (VN_IS(nodep, AssignW)) {
                nodep->v3error("Wire inputs its own output, creating circular logic (wire x=x)");
                return false;  // Don't delete the assign, or V3Gate will freak out
            } else {
                VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
                return true;
            }
        } else if (m_doV && VN_IS(nodep->lhsp(), Concat)) {
            bool need_temp = false;
            if (m_warn && !VN_IS(nodep, AssignDly)) {  // Is same var on LHS and RHS?
                // Note only do this (need user4) when m_warn, which is
                // done as unique visitor
                AstUser4InUse m_inuser4;
                ConstVarMarkVisitor mark(nodep->lhsp());
                ConstVarFindVisitor find(nodep->rhsp());
                if (find.found()) need_temp = true;
            }
            if (need_temp) {
                // The first time we constify, there may be the same variable on the LHS
                // and RHS.  In that case, we must use temporaries, or {a,b}={b,a} will break.
                UINFO(4, "  ASSITEMP " << nodep << endl);
                // ASSIGN(CONCAT(lc1,lc2),rhs) -> ASSIGN(temp1,SEL(rhs,{size})),
                //                                ASSIGN(temp2,SEL(newrhs,{size}))
                //                                ASSIGN(lc1,temp1),
                //                                ASSIGN(lc2,temp2)
            } else {
                UINFO(4, "  ASSI " << nodep << endl);
                // ASSIGN(CONCAT(lc1,lc2),rhs) -> ASSIGN(lc1,SEL(rhs,{size})),
                //                                ASSIGN(lc2,SEL(newrhs,{size}))
            }
            if (debug() >= 9) nodep->dumpTree(cout, "  Ass_old: ");
            // Unlink the stuff
            AstNode* lc1p = VN_CAST(nodep->lhsp(), Concat)->lhsp()->unlinkFrBack();
            AstNode* lc2p = VN_CAST(nodep->lhsp(), Concat)->rhsp()->unlinkFrBack();
            AstNode* conp = VN_CAST(nodep->lhsp(), Concat)->unlinkFrBack();
            AstNode* rhsp = nodep->rhsp()->unlinkFrBack();
            AstNode* rhs2p = rhsp->cloneTree(false);
            // Calc widths
            int lsb2 = 0;
            int msb2 = lsb2 + lc2p->width() - 1;
            int lsb1 = msb2 + 1;
            int msb1 = lsb1 + lc1p->width() - 1;
            UASSERT_OBJ(msb1 == (conp->width() - 1), nodep, "Width calc mismatch");
            // Form ranges
            AstSel* sel1p = new AstSel(conp->fileline(), rhsp, lsb1, msb1 - lsb1 + 1);
            AstSel* sel2p = new AstSel(conp->fileline(), rhs2p, lsb2, msb2 - lsb2 + 1);
            // Make new assigns of same flavor as old one
            //*** Not cloneTree; just one node.
            AstNode* newp = nullptr;
            if (!need_temp) {
                AstNodeAssign* asn1ap = VN_CAST(nodep->cloneType(lc1p, sel1p), NodeAssign);
                AstNodeAssign* asn2ap = VN_CAST(nodep->cloneType(lc2p, sel2p), NodeAssign);
                asn1ap->dtypeFrom(sel1p);
                asn2ap->dtypeFrom(sel2p);
                newp = AstNode::addNext(newp, asn1ap);
                newp = AstNode::addNext(newp, asn2ap);
            } else {
                UASSERT_OBJ(m_modp, nodep, "Not under module");
                // We could create just one temp variable, but we'll get better optimization
                // if we make one per term.
                string name1 = (string("__Vconcswap") + cvtToStr(m_modp->varNumGetInc()));
                string name2 = (string("__Vconcswap") + cvtToStr(m_modp->varNumGetInc()));
                AstVar* temp1p = new AstVar(sel1p->fileline(), AstVarType::BLOCKTEMP, name1,
                                            VFlagLogicPacked(), msb1 - lsb1 + 1);
                AstVar* temp2p = new AstVar(sel2p->fileline(), AstVarType::BLOCKTEMP, name2,
                                            VFlagLogicPacked(), msb2 - lsb2 + 1);
                m_modp->addStmtp(temp1p);
                m_modp->addStmtp(temp2p);
                AstNodeAssign* asn1ap
                    = VN_CAST(nodep->cloneType(
                                  new AstVarRef(sel1p->fileline(), temp1p, VAccess::WRITE), sel1p),
                              NodeAssign);
                AstNodeAssign* asn2ap
                    = VN_CAST(nodep->cloneType(
                                  new AstVarRef(sel2p->fileline(), temp2p, VAccess::WRITE), sel2p),
                              NodeAssign);
                AstNodeAssign* asn1bp
                    = VN_CAST(nodep->cloneType(
                                  lc1p, new AstVarRef(sel1p->fileline(), temp1p, VAccess::READ)),
                              NodeAssign);
                AstNodeAssign* asn2bp
                    = VN_CAST(nodep->cloneType(
                                  lc2p, new AstVarRef(sel2p->fileline(), temp2p, VAccess::READ)),
                              NodeAssign);
                asn1ap->dtypeFrom(temp1p);
                asn1bp->dtypeFrom(temp1p);
                asn2ap->dtypeFrom(temp2p);
                asn2bp->dtypeFrom(temp2p);
                // This order matters
                newp = AstNode::addNext(newp, asn1ap);
                newp = AstNode::addNext(newp, asn2ap);
                newp = AstNode::addNext(newp, asn1bp);
                newp = AstNode::addNext(newp, asn2bp);
            }
            if (debug() >= 9 && newp) newp->dumpTreeAndNext(cout, "     _new: ");
            nodep->addNextHere(newp);
            // Cleanup
            VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
            VL_DO_DANGLING(conp->deleteTree(), conp);
            // Further reduce, either node may have more reductions.
            return true;
        } else if (m_doV && VN_IS(nodep->rhsp(), StreamR)) {
            // The right-streaming operator on rhs of assignment does not
            // change the order of bits. Eliminate stream but keep its lhsp
            // Unlink the stuff
            AstNode* srcp = VN_CAST(nodep->rhsp(), StreamR)->lhsp()->unlinkFrBack();
            AstNode* sizep = VN_CAST(nodep->rhsp(), StreamR)->rhsp()->unlinkFrBack();
            AstNode* streamp = VN_CAST(nodep->rhsp(), StreamR)->unlinkFrBack();
            nodep->rhsp(srcp);
            // Cleanup
            VL_DO_DANGLING(sizep->deleteTree(), sizep);
            VL_DO_DANGLING(streamp->deleteTree(), streamp);
            // Further reduce, any of the nodes may have more reductions.
            return true;
        } else if (m_doV && VN_IS(nodep->lhsp(), StreamL)) {
            // Push the stream operator to the rhs of the assignment statement
            int dWidth = VN_CAST(nodep->lhsp(), StreamL)->lhsp()->width();
            int sWidth = nodep->rhsp()->width();
            // Unlink the stuff
            AstNode* dstp = VN_CAST(nodep->lhsp(), StreamL)->lhsp()->unlinkFrBack();
            AstNode* streamp = VN_CAST(nodep->lhsp(), StreamL)->unlinkFrBack();
            AstNode* srcp = nodep->rhsp()->unlinkFrBack();
            // Connect the rhs to the stream operator and update its width
            VN_CAST(streamp, StreamL)->lhsp(srcp);
            streamp->dtypeSetLogicUnsized(srcp->width(), srcp->widthMin(), VSigning::UNSIGNED);
            // Shrink the RHS if necessary
            if (sWidth > dWidth) {
                streamp = new AstSel(streamp->fileline(), streamp, sWidth - dWidth, dWidth);
            }
            // Link the nodes back in
            nodep->lhsp(dstp);
            nodep->rhsp(streamp);
            return true;
        } else if (m_doV && VN_IS(nodep->lhsp(), StreamR)) {
            // The right stream operator on lhs of assignment statement does
            // not reorder bits. However, if the rhs is wider than the lhs,
            // then we select bits from the left-most, not the right-most.
            int dWidth = VN_CAST(nodep->lhsp(), StreamR)->lhsp()->width();
            int sWidth = nodep->rhsp()->width();
            // Unlink the stuff
            AstNode* dstp = VN_CAST(nodep->lhsp(), StreamR)->lhsp()->unlinkFrBack();
            AstNode* sizep = VN_CAST(nodep->lhsp(), StreamR)->rhsp()->unlinkFrBack();
            AstNode* streamp = VN_CAST(nodep->lhsp(), StreamR)->unlinkFrBack();
            AstNode* srcp = nodep->rhsp()->unlinkFrBack();
            if (sWidth > dWidth) {
                srcp = new AstSel(streamp->fileline(), srcp, sWidth - dWidth, dWidth);
            }
            nodep->lhsp(dstp);
            nodep->rhsp(srcp);
            // Cleanup
            VL_DO_DANGLING(sizep->deleteTree(), sizep);
            VL_DO_DANGLING(streamp->deleteTree(), streamp);
            // Further reduce, any of the nodes may have more reductions.
            return true;
        } else if (replaceAssignMultiSel(nodep)) {
            return true;
        }
        return false;
    }

    // Boolean replacements
    bool operandBoolShift(const AstNode* nodep) {
        // boolean test of AND(const,SHIFTR(x,const)) -> test of AND(SHIFTL(x,const), x)
        if (!VN_IS(nodep, And)) return false;
        if (!VN_IS(VN_CAST_CONST(nodep, And)->lhsp(), Const)) return false;
        if (!VN_IS(VN_CAST_CONST(nodep, And)->rhsp(), ShiftR)) return false;
        const AstShiftR* shiftp = VN_CAST(VN_CAST_CONST(nodep, And)->rhsp(), ShiftR);
        if (!VN_IS(shiftp->rhsp(), Const)) return false;
        if (static_cast<uint32_t>(nodep->width())
            <= VN_CAST_CONST(shiftp->rhsp(), Const)->toUInt()) {
            return false;
        }
        return true;
    }
    void replaceBoolShift(AstNode* nodep) {
        if (debug() >= 9) nodep->dumpTree(cout, "  bshft_old: ");
        AstConst* andConstp = VN_CAST(VN_CAST(nodep, And)->lhsp(), Const);
        AstNode* fromp = VN_CAST(VN_CAST(nodep, And)->rhsp(), ShiftR)->lhsp()->unlinkFrBack();
        AstConst* shiftConstp
            = VN_CAST(VN_CAST(VN_CAST(nodep, And)->rhsp(), ShiftR)->rhsp(), Const);
        V3Number val(andConstp, andConstp->width());
        val.opShiftL(andConstp->num(), shiftConstp->num());
        AstAnd* newp = new AstAnd(nodep->fileline(), new AstConst(nodep->fileline(), val), fromp);
        // widthMin no longer applicable if different C-expanded width
        newp->dtypeSetLogicSized(nodep->width(), VSigning::UNSIGNED);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        if (debug() >= 9) newp->dumpTree(cout, "       _new: ");
    }

    void replaceWithSimulation(AstNode* nodep) {
        SimulateVisitor simvis;
        // Run it - may be unoptimizable due to large for loop, etc
        simvis.mainParamEmulate(nodep);
        if (!simvis.optimizable()) {
            AstNode* errorp = simvis.whyNotNodep();
            if (!errorp) errorp = nodep;
            nodep->v3error("Expecting expression to be constant, but can't determine constant for "
                           << nodep->prettyTypeName() << '\n'
                           << errorp->warnOther() << "... Location of non-constant "
                           << errorp->prettyTypeName() << ": " << simvis.whyNotMessage());
            VL_DO_DANGLING(replaceZero(nodep), nodep);
        } else {
            // Fetch the result
            AstNode* valuep = simvis.fetchValueNull(nodep);  // valuep is owned by Simulate
            UASSERT_OBJ(valuep, nodep, "No value returned from simulation");
            // Replace it
            AstNode* newp = valuep->cloneTree(false);
            newp->dtypeFrom(nodep);
            newp->fileline(nodep->fileline());
            UINFO(4, "Simulate->" << newp << endl);
            nodep->replaceWith(newp);
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
        }
    }

    //----------------------------------------

    // VISITORS
    virtual void visit(AstNetlist* nodep) override {
        // Iterate modules backwards, in bottom-up order.  That's faster
        iterateChildrenBackwards(nodep);
    }
    virtual void visit(AstNodeModule* nodep) override {
        VL_RESTORER(m_modp);
        {
            m_modp = nodep;
            iterateChildren(nodep);
        }
    }
    virtual void visit(AstCFunc* nodep) override {
        // No ASSIGNW removals under funcs, we've long eliminated INITIALs
        // (We should perhaps rename the assignw's to just assigns)
        VL_RESTORER(m_wremove);
        {
            m_wremove = false;
            iterateChildren(nodep);
        }
    }
    virtual void visit(AstScope* nodep) override {
        // No ASSIGNW removals under scope, we've long eliminated INITIALs
        VL_RESTORER(m_wremove);
        VL_RESTORER(m_scopep);
        {
            m_wremove = false;
            m_scopep = nodep;
            iterateChildren(nodep);
        }
    }

    void swapSides(AstNodeBiCom* nodep) {
        // COMMUTATIVE({a},CONST) -> COMMUTATIVE(CONST,{a})
        // This simplifies later optimizations
        AstNode* lhsp = nodep->lhsp()->unlinkFrBackWithNext();
        AstNode* rhsp = nodep->rhsp()->unlinkFrBackWithNext();
        nodep->lhsp(rhsp);
        nodep->rhsp(lhsp);
        iterate(nodep);  // Again?
    }

    int operandConcatMove(AstConcat* nodep) {
        //    CONCAT under concat  (See moveConcat)
        // Return value: true indicates to do it; 2 means move to LHS
        AstConcat* abConcp = VN_CAST(nodep->lhsp(), Concat);
        AstConcat* bcConcp = VN_CAST(nodep->rhsp(), Concat);
        if (!abConcp && !bcConcp) return 0;
        if (bcConcp) {
            AstNode* ap = nodep->lhsp();
            AstNode* bp = bcConcp->lhsp();
            // If a+b == 32,64,96 etc, then we want to have a+b together on LHS
            if (VL_BITBIT_I(ap->width() + bp->width()) == 0) return 2;  // Transform 2: to abConc
        } else {  // abConcp
            // Unless lhs is already 32 bits due to above, reorder it
            if (VL_BITBIT_I(nodep->lhsp()->width()) != 0) return 1;  // Transform 1: to bcConc
        }
        return 0;  // ok
    }
    void moveConcat(AstConcat* nodep) {
        //    1: CONCAT(CONCAT({a},{b}),{c})  -> CONCAT({a},CONCAT({b}, {c}))
        // or 2: CONCAT({a}, CONCAT({b},{c})) -> CONCAT(CONCAT({a},{b}),{c})
        // Because the lhs of a concat needs a shift but the rhs doesn't,
        // putting additional CONCATs on the RHS leads to fewer assembler operations.
        // However, we'll end up with lots of wide moves if we make huge trees
        // like that, so on 32 bit boundaries, we'll do the opposite form.
        UINFO(4, "Move concat: " << nodep << endl);
        if (operandConcatMove(nodep) > 1) {
            AstNode* ap = nodep->lhsp()->unlinkFrBack();
            AstConcat* bcConcp = VN_CAST(nodep->rhsp(), Concat);
            bcConcp->unlinkFrBack();
            AstNode* bp = bcConcp->lhsp()->unlinkFrBack();
            AstNode* cp = bcConcp->rhsp()->unlinkFrBack();
            AstConcat* abConcp = new AstConcat(bcConcp->fileline(), ap, bp);
            nodep->lhsp(abConcp);
            nodep->rhsp(cp);
            // If bp was a concat, then we have this exact same form again!
            // Recurse rather then calling node->iterate to prevent 2^n recursion!
            if (operandConcatMove(abConcp)) moveConcat(abConcp);
            VL_DO_DANGLING(bcConcp->deleteTree(), bcConcp);
        } else {
            AstConcat* abConcp = VN_CAST(nodep->lhsp(), Concat);
            abConcp->unlinkFrBack();
            AstNode* ap = abConcp->lhsp()->unlinkFrBack();
            AstNode* bp = abConcp->rhsp()->unlinkFrBack();
            AstNode* cp = nodep->rhsp()->unlinkFrBack();
            AstConcat* bcConcp = new AstConcat(abConcp->fileline(), bp, cp);
            nodep->lhsp(ap);
            nodep->rhsp(bcConcp);
            if (operandConcatMove(bcConcp)) moveConcat(bcConcp);
            VL_DO_DANGLING(abConcp->deleteTree(), abConcp);
        }
    }

    // Special cases
    virtual void visit(AstConst*) override {}  // Already constant

    virtual void visit(AstCell* nodep) override {
        if (m_params) {
            iterateAndNextNull(nodep->paramsp());
        } else {
            iterateChildren(nodep);
        }
    }
    virtual void visit(AstPin* nodep) override { iterateChildren(nodep); }

    void replaceLogEq(AstLogEq* nodep) {
        // LOGEQ(a,b) => AstLogAnd{AstLogOr{AstLogNot{a},b},AstLogOr{AstLogNot{b},a}}
        AstNode* lhsp = nodep->lhsp()->unlinkFrBack();
        AstNode* rhsp = nodep->rhsp()->unlinkFrBack();
        // Do exactly as IEEE says, might result in extra terms, so in future may do differently
        AstLogAnd* newp = new AstLogAnd(
            nodep->fileline(),
            new AstLogOr(nodep->fileline(), new AstLogNot(nodep->fileline(), lhsp), rhsp),
            new AstLogOr(nodep->fileline(),
                         new AstLogNot(nodep->fileline(), rhsp->cloneTree(false)),
                         lhsp->cloneTree(false)));
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }

    void replaceSelSel(AstSel* nodep) {
        // SEL(SEL({x},a,b),c,d) => SEL({x},a+c,d)
        AstSel* belowp = VN_CAST(nodep->fromp(), Sel);
        AstNode* fromp = belowp->fromp()->unlinkFrBack();
        AstNode* widthp = nodep->widthp()->unlinkFrBack();
        AstNode* lsb1p = nodep->lsbp()->unlinkFrBack();
        AstNode* lsb2p = belowp->lsbp()->unlinkFrBack();
        // Eliminate lower range
        UINFO(4, "Elim Lower range: " << nodep << endl);
        AstNode* newlsbp;
        if (VN_IS(lsb1p, Const) && VN_IS(lsb2p, Const)) {
            newlsbp = new AstConst(lsb1p->fileline(), VN_CAST(lsb1p, Const)->toUInt()
                                                          + VN_CAST(lsb2p, Const)->toUInt());
            VL_DO_DANGLING(lsb1p->deleteTree(), lsb1p);
            VL_DO_DANGLING(lsb2p->deleteTree(), lsb2p);
        } else {
            // Width is important, we need the width of the fromp's
            // expression, not the potentially smaller lsb1p's width
            newlsbp
                = new AstAdd(lsb1p->fileline(), lsb2p, new AstExtend(lsb1p->fileline(), lsb1p));
            newlsbp->dtypeFrom(lsb2p);  // Unsigned
            VN_CAST(newlsbp, Add)->rhsp()->dtypeFrom(lsb2p);
        }
        AstSel* newp = new AstSel(nodep->fileline(), fromp, newlsbp, widthp);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }

    void replaceSelConcat(AstSel* nodep) {
        // SEL(CONCAT(a,b),c,d) => SEL(a or b, . .)
        AstConcat* conp = VN_CAST(nodep->fromp(), Concat);
        AstNode* conLhsp = conp->lhsp();
        AstNode* conRhsp = conp->rhsp();
        if (static_cast<int>(nodep->lsbConst()) >= conRhsp->width()) {
            conLhsp->unlinkFrBack();
            AstSel* newp = new AstSel(nodep->fileline(), conLhsp,
                                      nodep->lsbConst() - conRhsp->width(), nodep->widthConst());
            nodep->replaceWith(newp);
        } else if (static_cast<int>(nodep->msbConst()) < conRhsp->width()) {
            conRhsp->unlinkFrBack();
            AstSel* newp
                = new AstSel(nodep->fileline(), conRhsp, nodep->lsbConst(), nodep->widthConst());
            nodep->replaceWith(newp);
        } else {
            // Yuk, split between the two
            conRhsp->unlinkFrBack();
            conLhsp->unlinkFrBack();
            AstConcat* newp
                = new AstConcat(nodep->fileline(),
                                new AstSel(nodep->fileline(), conLhsp, 0,
                                           nodep->msbConst() - conRhsp->width() + 1),
                                new AstSel(nodep->fileline(), conRhsp, nodep->lsbConst(),
                                           conRhsp->width() - nodep->lsbConst()));
            nodep->replaceWith(newp);
        }
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    bool operandSelReplicate(AstSel* nodep) {
        // SEL(REPLICATE(from,rep),lsb,width) => SEL(from,0,width) as long
        // as SEL's width <= b's width
        AstReplicate* repp = VN_CAST(nodep->fromp(), Replicate);
        AstNode* fromp = repp->lhsp();
        AstConst* lsbp = VN_CAST(nodep->lsbp(), Const);
        if (!lsbp) return false;
        AstNode* widthp = nodep->widthp();
        if (!VN_IS(widthp, Const)) return false;
        UASSERT_OBJ(fromp->width(), nodep, "Not widthed");
        if ((lsbp->toUInt() / fromp->width())
            != ((lsbp->toUInt() + nodep->width() - 1) / fromp->width())) {
            return false;
        }
        //
        fromp->unlinkFrBack();
        widthp->unlinkFrBack();
        AstSel* newp
            = new AstSel(nodep->fileline(), fromp,
                         new AstConst(lsbp->fileline(), lsbp->toUInt() % fromp->width()), widthp);
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        return true;
    }
    bool operandRepRep(AstReplicate* nodep) {
        // REPLICATE(REPLICATE2(from2,cnt2),cnt1) => REPLICATE(from2,(cnt1+cnt2))
        AstReplicate* rep2p = VN_CAST(nodep->lhsp(), Replicate);
        AstNode* from2p = rep2p->lhsp();
        AstConst* cnt1p = VN_CAST(nodep->rhsp(), Const);
        if (!cnt1p) return false;
        AstConst* cnt2p = VN_CAST(rep2p->rhsp(), Const);
        if (!cnt2p) return false;
        //
        from2p->unlinkFrBack();
        cnt1p->unlinkFrBack();
        cnt2p->unlinkFrBack();
        AstReplicate* newp
            = new AstReplicate(nodep->fileline(), from2p, cnt1p->toUInt() * cnt2p->toUInt());
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        return true;
    }
    bool operandConcatSame(AstConcat* nodep) {
        // CONCAT(fromp,fromp) -> REPLICATE(fromp,1+1)
        // CONCAT(REP(fromp,cnt1),fromp) -> REPLICATE(fromp,cnt1+1)
        // CONCAT(fromp,REP(fromp,cnt1)) -> REPLICATE(fromp,1+cnt1)
        // CONCAT(REP(fromp,cnt1),REP(fromp,cnt2)) -> REPLICATE(fromp,cnt1+cnt2)
        AstNode* from1p = nodep->lhsp();
        uint32_t cnt1 = 1;
        AstNode* from2p = nodep->rhsp();
        uint32_t cnt2 = 1;
        if (VN_IS(from1p, Replicate)) {
            AstConst* cnt1p = VN_CAST(VN_CAST(from1p, Replicate)->rhsp(), Const);
            if (!cnt1p) return false;
            from1p = VN_CAST(from1p, Replicate)->lhsp();
            cnt1 = cnt1p->toUInt();
        }
        if (VN_IS(from2p, Replicate)) {
            AstConst* cnt2p = VN_CAST(VN_CAST(from2p, Replicate)->rhsp(), Const);
            if (!cnt2p) return false;
            from2p = VN_CAST(from2p, Replicate)->lhsp();
            cnt2 = cnt2p->toUInt();
        }
        if (!operandsSame(from1p, from2p)) return false;
        //
        from1p->unlinkFrBack();
        AstReplicate* newp = new AstReplicate(nodep->fileline(), from1p, cnt1 + cnt2);
        newp->dtypeFrom(nodep);
        nodep->replaceWith(newp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
        return true;
    }
    void replaceSelIntoBiop(AstSel* nodep) {
        // SEL(BUFIF1(a,b),1,bit) => BUFIF1(SEL(a,1,bit),SEL(b,1,bit))
        AstNodeBiop* fromp = VN_CAST(nodep->fromp()->unlinkFrBack(), NodeBiop);
        UASSERT_OBJ(fromp, nodep, "Called on non biop");
        AstNode* lsbp = nodep->lsbp()->unlinkFrBack();
        AstNode* widthp = nodep->widthp()->unlinkFrBack();
        //
        AstNode* bilhsp = fromp->lhsp()->unlinkFrBack();
        AstNode* birhsp = fromp->rhsp()->unlinkFrBack();
        //
        fromp->lhsp(
            new AstSel(nodep->fileline(), bilhsp, lsbp->cloneTree(true), widthp->cloneTree(true)));
        fromp->rhsp(new AstSel(nodep->fileline(), birhsp, lsbp, widthp));
        fromp->dtypeFrom(nodep);
        nodep->replaceWith(fromp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }
    void replaceSelIntoUniop(AstSel* nodep) {
        // SEL(NOT(a),1,bit) => NOT(SEL(a,bit))
        AstNodeUniop* fromp = VN_CAST(nodep->fromp()->unlinkFrBack(), NodeUniop);
        UASSERT_OBJ(fromp, nodep, "Called on non biop");
        AstNode* lsbp = nodep->lsbp()->unlinkFrBack();
        AstNode* widthp = nodep->widthp()->unlinkFrBack();
        //
        AstNode* bilhsp = fromp->lhsp()->unlinkFrBack();
        //
        fromp->lhsp(new AstSel(nodep->fileline(), bilhsp, lsbp, widthp));
        fromp->dtypeFrom(nodep);
        nodep->replaceWith(fromp);
        VL_DO_DANGLING(nodep->deleteTree(), nodep);
    }

    virtual void visit(AstAttrOf* nodep) override {
        VL_RESTORER(m_attrp);
        {
            m_attrp = nodep;
            iterateChildren(nodep);
        }
    }

    virtual void visit(AstArraySel* nodep) override {
        iterateAndNextNull(nodep->bitp());
        if (VN_IS(nodep->bitp(), Const)
            && VN_IS(nodep->fromp(), VarRef)
            // Need to make sure it's an array object so don't mis-allow a constant (bug509.)
            && VN_CAST(nodep->fromp(), VarRef)->varp()
            && VN_IS(VN_CAST(nodep->fromp(), VarRef)->varp()->valuep(), InitArray)) {
            m_selp = nodep;  // Ask visit(AstVarRef) to replace varref with const
        }
        iterateAndNextNull(nodep->fromp());
        if (VN_IS(nodep->fromp(), Const)) {  // It did.
            if (!m_selp) {
                nodep->v3error("Illegal assignment of constant to unpacked array");
            } else {
                AstNode* fromp = nodep->fromp()->unlinkFrBack();
                nodep->replaceWith(fromp);
                if (VN_IS(fromp->dtypep()->skipRefp(), NodeArrayDType)) {
                    // Strip off array to find what array references
                    fromp->dtypeFrom(
                        VN_CAST(fromp->dtypep()->skipRefp(), NodeArrayDType)->subDTypep());
                }
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            }
        }
        m_selp = nullptr;
    }
    virtual void visit(AstNodeVarRef* nodep) override {
        iterateChildren(nodep);
        UASSERT_OBJ(nodep->varp(), nodep, "Not linked");
        bool did = false;
        if (m_doV && nodep->varp()->valuep() && !m_attrp) {
            // if (debug()) valuep->dumpTree(cout, "  visitvaref: ");
            iterateAndNextNull(nodep->varp()->valuep());  // May change nodep->varp()->valuep()
            AstNode* valuep = nodep->varp()->valuep();
            if (nodep->access().isReadOnly()
                && ((!m_params  // Can reduce constant wires into equations
                     && m_doNConst
                     && v3Global.opt.oConst()
                     // Default value, not a "known" constant for this usage
                     && !nodep->varp()->isClassMember()
                     && !(nodep->varp()->isFuncLocal() && nodep->varp()->isNonOutput())
                     && !nodep->varp()->noSubst() && !nodep->varp()->isSigPublic())
                    || nodep->varp()->isParam())) {
                if (operandConst(valuep)) {
                    const V3Number& num = VN_CAST(valuep, Const)->num();
                    // UINFO(2,"constVisit "<<cvtToHex(valuep)<<" "<<num<<endl);
                    VL_DO_DANGLING(replaceNum(nodep, num), nodep);
                    did = true;
                } else if (m_selp && VN_IS(valuep, InitArray)) {
                    AstInitArray* initarp = VN_CAST(valuep, InitArray);
                    uint32_t bit = m_selp->bitConst();
                    AstNode* itemp = initarp->getIndexDefaultedValuep(bit);
                    if (VN_IS(itemp, Const)) {
                        const V3Number& num = VN_CAST(itemp, Const)->num();
                        // UINFO(2,"constVisit "<<cvtToHex(valuep)<<" "<<num<<endl);
                        VL_DO_DANGLING(replaceNum(nodep, num), nodep);
                        did = true;
                    }
                } else if (m_params && VN_IS(valuep, InitArray)) {
                    // Allow parameters to pass arrays
                    // Earlier recursion of InitArray made sure each array value is constant
                    // This exception is fairly fragile, i.e. doesn't
                    // support arrays of arrays or other stuff
                    AstNode* newp = valuep->cloneTree(false);
                    nodep->replaceWith(newp);
                    VL_DO_DANGLING(nodep->deleteTree(), nodep);
                    did = true;
                } else if (nodep->varp()->isParam() && VN_IS(valuep, Unbounded)) {
                    AstNode* newp = valuep->cloneTree(false);
                    nodep->replaceWith(newp);
                    VL_DO_DANGLING(nodep->deleteTree(), nodep);
                    did = true;
                }
            }
        }
        if (!did && m_required) {
            nodep->v3error("Expecting expression to be constant, but variable isn't const: "
                           << nodep->varp()->prettyNameQ());
        }
    }
    virtual void visit(AstEnumItemRef* nodep) override {
        iterateChildren(nodep);
        UASSERT_OBJ(nodep->itemp(), nodep, "Not linked");
        bool did = false;
        if (nodep->itemp()->valuep()) {
            // if (debug()) nodep->itemp()->valuep()->dumpTree(cout, "  visitvaref: ");
            if (nodep->itemp()->user4()) {
                nodep->v3error("Recursive enum value: " << nodep->itemp()->prettyNameQ());
            } else {
                nodep->itemp()->user4(true);
                iterateAndNextNull(nodep->itemp()->valuep());
                nodep->itemp()->user4(false);
            }
            if (AstConst* valuep = VN_CAST(nodep->itemp()->valuep(), Const)) {
                const V3Number& num = valuep->num();
                VL_DO_DANGLING(replaceNum(nodep, num), nodep);
                did = true;
            }
        }
        if (!did && m_required) {
            nodep->v3error("Expecting expression to be constant, but enum value isn't const: "
                           << nodep->itemp()->prettyNameQ());
        }
    }

    // virtual void visit(AstCvtPackString* nodep) override {
    // Not constant propagated (for today) because AstNodeMath::isOpaque is set
    // Someday if lower is constant, convert to quoted "string".

    bool onlySenItemInSenTree(AstSenItem* nodep) {
        // Only one if it's not in a list
        return (!nodep->nextp() && nodep->backp()->nextp() != nodep);
    }
    virtual void visit(AstSenItem* nodep) override {
        iterateChildren(nodep);
        if (m_doNConst
            && (VN_IS(nodep->sensp(), Const) || VN_IS(nodep->sensp(), EnumItemRef)
                || (nodep->varrefp() && nodep->varrefp()->varp()->isParam()))) {
            // Constants in sensitivity lists may be removed (we'll simplify later)
            if (nodep->isClocked()) {  // A constant can never get a pos/negedge
                if (onlySenItemInSenTree(nodep)) {
                    nodep->replaceWith(new AstSenItem(nodep->fileline(), AstSenItem::Never()));
                    VL_DO_DANGLING(nodep->deleteTree(), nodep);
                } else {
                    VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
                }
            } else {  // Otherwise it may compute a result that needs to settle out
                nodep->replaceWith(new AstSenItem(nodep->fileline(), AstSenItem::Combo()));
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            }
        } else if (m_doNConst && VN_IS(nodep->sensp(), Not)) {
            // V3Gate may propagate NOTs into clocks... Just deal with it
            AstNode* sensp = nodep->sensp();
            AstNode* lastSensp = sensp;
            bool invert = false;
            while (VN_IS(lastSensp, Not)) {
                lastSensp = VN_CAST(lastSensp, Not)->lhsp();
                invert = !invert;
            }
            UINFO(8, "senItem(NOT...) " << nodep << " " << invert << endl);
            if (invert) nodep->edgeType(nodep->edgeType().invert());
            AstNodeVarRef* senvarp = VN_CAST(lastSensp->unlinkFrBack(), NodeVarRef);
            UASSERT_OBJ(senvarp, sensp, "Non-varref sensitivity variable");
            sensp->replaceWith(senvarp);
            VL_DO_DANGLING(sensp->deleteTree(), sensp);
        } else if (!m_doNConst  // Deal with later when doNConst missing
                   && (VN_IS(nodep->sensp(), EnumItemRef) || VN_IS(nodep->sensp(), Const))) {
        } else if (nodep->isIllegal()) {  // Deal with later
        } else {
            UASSERT_OBJ(!(nodep->hasVar() && !nodep->varrefp()), nodep,
                        "Null sensitivity variable");
        }
    }

    struct SenItemCmp {
        bool operator()(const AstSenItem* lhsp, const AstSenItem* rhsp) const {
            if (lhsp->type() < rhsp->type()) return true;
            if (lhsp->type() > rhsp->type()) return false;
            // Looks visually better if we keep sorted by name
            if (!lhsp->varrefp() && rhsp->varrefp()) return true;
            if (lhsp->varrefp() && !rhsp->varrefp()) return false;
            if (lhsp->varrefp() && rhsp->varrefp()) {
                if (lhsp->varrefp()->name() < rhsp->varrefp()->name()) return true;
                if (lhsp->varrefp()->name() > rhsp->varrefp()->name()) return false;
                // But might be same name with different scopes
                if (lhsp->varrefp()->varScopep() < rhsp->varrefp()->varScopep()) return true;
                if (lhsp->varrefp()->varScopep() > rhsp->varrefp()->varScopep()) return false;
                // Or rarely, different data types
                if (lhsp->varrefp()->dtypep() < rhsp->varrefp()->dtypep()) return true;
                if (lhsp->varrefp()->dtypep() > rhsp->varrefp()->dtypep()) return false;
            }
            // Sort by edge, AFTER variable, as we want multiple edges for same var adjacent.
            // note the SenTree optimizer requires this order (more
            // general first, less general last)
            if (lhsp->edgeType() < rhsp->edgeType()) return true;
            if (lhsp->edgeType() > rhsp->edgeType()) return false;
            return false;
        }
    };

    virtual void visit(AstSenTree* nodep) override {
        iterateChildren(nodep);
        if (m_doExpensive) {
            // cout<<endl; nodep->dumpTree(cout, "ssin: ");
            // Optimize ideas for the future:
            //   SENTREE(... SENGATE(x,a), SENGATE(SENITEM(x),b) ...)  => SENGATE(x,OR(a,b))

            //   SENTREE(... SENITEM(x),   SENGATE(SENITEM(x),*) ...)  => SENITEM(x)
            // Do we need the SENITEM's to be identical?  No because we're
            // ORing between them; we just need to ensure that the result is at
            // least as frequently activating.  So we simply
            // SENGATE(SENITEM(x)) -> SENITEM(x), then let it collapse with the
            // other SENITEM(x).
            {
                AstUser4InUse m_inuse4;
                // Mark x in SENITEM(x)
                for (AstSenItem* senp = VN_CAST(nodep->sensesp(), SenItem); senp;
                     senp = VN_CAST(senp->nextp(), SenItem)) {
                    if (senp->varrefp() && senp->varrefp()->varScopep()) {
                        senp->varrefp()->varScopep()->user4(1);
                    }
                }
            }

            // Sort the sensitivity names so "posedge a or b" and "posedge b or a" end up together.
            // Also, remove duplicate assignments, and fold POS&NEGs into ANYEDGEs
            // Make things a little faster; check first if we need a sort
            for (AstSenItem *nextp, *senp = VN_CAST(nodep->sensesp(), SenItem); senp;
                 senp = nextp) {
                nextp = VN_CAST(senp->nextp(), SenItem);
                // cppcheck-suppress unassignedVariable  // cppcheck bug
                SenItemCmp cmp;
                if (nextp && !cmp(senp, nextp)) {
                    // Something's out of order, sort it
                    senp = nullptr;
                    std::vector<AstSenItem*> vec;
                    for (AstSenItem* senp = VN_CAST(nodep->sensesp(), SenItem); senp;
                         senp = VN_CAST(senp->nextp(), SenItem)) {
                        vec.push_back(senp);
                    }
                    stable_sort(vec.begin(), vec.end(), SenItemCmp());
                    for (const auto& ip : vec) ip->unlinkFrBack();
                    for (const auto& ip : vec) nodep->addSensesp(ip);
                    break;
                }
            }

            // Pass2, remove dup edges
            for (AstSenItem *nextp, *senp = VN_CAST(nodep->sensesp(), SenItem); senp;
                 senp = nextp) {
                nextp = VN_CAST(senp->nextp(), SenItem);
                AstSenItem* const litemp = senp;
                AstSenItem* const ritemp = nextp;
                if (ritemp) {
                    if ((litemp->varrefp() && ritemp->varrefp()
                         && litemp->varrefp()->sameGateTree(ritemp->varrefp()))
                        || (!litemp->varrefp() && !ritemp->varrefp())) {
                        // We've sorted in the order ANY, BOTH, POS, NEG,
                        // so we don't need to try opposite orders
                        if ((litemp->edgeType()
                             == VEdgeType::ET_ANYEDGE)  // ANY  or {BOTH|POS|NEG} -> ANY
                            || (litemp->edgeType()
                                == VEdgeType::ET_BOTHEDGE)  // BOTH or {POS|NEG} -> BOTH
                            || (litemp->edgeType() == VEdgeType::ET_POSEDGE  // POS  or NEG -> BOTH
                                && ritemp->edgeType() == VEdgeType::ET_NEGEDGE)
                            || (litemp->edgeType() == ritemp->edgeType())  // Identical edges
                        ) {
                            // Fix edge of old node
                            if (litemp->edgeType() == VEdgeType::ET_POSEDGE
                                && ritemp->edgeType() == VEdgeType::ET_NEGEDGE)
                                litemp->edgeType(VEdgeType::ET_BOTHEDGE);
                            // Remove redundant node
                            VL_DO_DANGLING(ritemp->unlinkFrBack()->deleteTree(), ritemp);
                            VL_DANGLING(ritemp);
                            // Try to collapse again
                            nextp = litemp;
                        }
                    }
                }
            }
            // nodep->dumpTree(cout, "ssou: ");
        }
    }

    //-----
    // Zero elimination
    virtual void visit(AstNodeAssign* nodep) override {
        iterateChildren(nodep);
        if (m_doNConst && replaceNodeAssign(nodep)) return;
    }
    virtual void visit(AstAssignAlias* nodep) override {
        // Don't perform any optimizations, keep the alias around
    }
    virtual void visit(AstAssignVarScope* nodep) override {
        // Don't perform any optimizations, the node won't be linked yet
    }
    virtual void visit(AstAssignW* nodep) override {
        iterateChildren(nodep);
        if (m_doNConst && replaceNodeAssign(nodep)) return;
        AstNodeVarRef* varrefp = VN_CAST(
            nodep->lhsp(),
            VarRef);  // Not VarXRef, as different refs may set different values to each hierarchy
        if (m_wremove && !m_params && m_doNConst && m_modp && operandConst(nodep->rhsp())
            && !VN_CAST(nodep->rhsp(), Const)->num().isFourState()
            && varrefp  // Don't do messes with BITREFs/ARRAYREFs
            && !varrefp->varp()->valuep()  // Not already constified
            && !varrefp->varScopep()) {  // Not scoped (or each scope may have different initial
                                         // value)
            // ASSIGNW (VARREF, const) -> INITIAL ( ASSIGN (VARREF, const) )
            UINFO(4, "constAssignW " << nodep << endl);
            // Make a initial assignment
            AstNode* exprp = nodep->rhsp()->unlinkFrBack();
            varrefp->unlinkFrBack();
            AstInitial* newinitp = new AstInitial(
                nodep->fileline(), new AstAssign(nodep->fileline(), varrefp, exprp));
            m_modp->addStmtp(newinitp);
            VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
            // Set the initial value right in the variable so we can constant propagate
            AstNode* initvaluep = exprp->cloneTree(false);
            varrefp->varp()->valuep(initvaluep);
        }
    }

    virtual void visit(AstNodeIf* nodep) override {
        iterateChildren(nodep);
        if (m_doNConst) {
            if (const AstConst* constp = VN_CAST(nodep->condp(), Const)) {
                AstNode* keepp = nullptr;
                if (constp->isZero()) {
                    UINFO(4, "IF(0,{any},{x}) => {x}: " << nodep << endl);
                    keepp = nodep->elsesp();
                } else if (!m_doV || constp->isNeqZero()) {  // Might be X in Verilog
                    UINFO(4, "IF(!0,{x},{any}) => {x}: " << nodep << endl);
                    keepp = nodep->ifsp();
                } else {
                    UINFO(4, "IF condition is X, retaining: " << nodep << endl);
                    return;
                }
                if (keepp) {
                    keepp->unlinkFrBackWithNext();
                    nodep->replaceWith(keepp);
                } else {
                    nodep->unlinkFrBack();
                }
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            } else if (!afterComment(nodep->ifsp()) && !afterComment(nodep->elsesp())) {
                // Empty block, remove it
                // Note if we support more C++ then there might be side
                // effects in the condition itself
                VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
            } else if (!afterComment(nodep->ifsp())) {
                UINFO(4, "IF({x}) nullptr {...} => IF(NOT{x}}: " << nodep << endl);
                AstNode* condp = nodep->condp();
                AstNode* elsesp = nodep->elsesp();
                condp->unlinkFrBackWithNext();
                elsesp->unlinkFrBackWithNext();
                if (nodep->ifsp()) {  // Must have been comment
                    nodep->ifsp()->unlinkFrBackWithNext()->deleteTree();
                }
                nodep->condp(new AstLogNot(condp->fileline(),
                                           condp));  // LogNot, as C++ optimization also possible
                nodep->addIfsp(elsesp);
            } else if (((VN_IS(nodep->condp(), Not) && nodep->condp()->width() == 1)
                        || VN_IS(nodep->condp(), LogNot))
                       && nodep->ifsp() && nodep->elsesp()) {
                UINFO(4, "IF(NOT {x})  => IF(x) swapped if/else" << nodep << endl);
                AstNode* condp
                    = VN_CAST(nodep->condp(), NodeUniop)->lhsp()->unlinkFrBackWithNext();
                AstNode* ifsp = nodep->ifsp()->unlinkFrBackWithNext();
                AstNode* elsesp = nodep->elsesp()->unlinkFrBackWithNext();
                AstIf* ifp = new AstIf(nodep->fileline(), condp, elsesp, ifsp);
                ifp->branchPred(nodep->branchPred().invert());
                nodep->replaceWith(ifp);
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            } else if (ifSameAssign(nodep)) {
                UINFO(
                    4,
                    "IF({a}) ASSIGN({b},{c}) else ASSIGN({b},{d}) => ASSIGN({b}, {a}?{c}:{d})\n");
                AstNodeAssign* ifp = VN_CAST(nodep->ifsp(), NodeAssign);
                AstNodeAssign* elsep = VN_CAST(nodep->elsesp(), NodeAssign);
                ifp->unlinkFrBack();
                AstNode* condp = nodep->condp()->unlinkFrBack();
                AstNode* truep = ifp->rhsp()->unlinkFrBack();
                AstNode* falsep = elsep->rhsp()->unlinkFrBack();
                ifp->rhsp(new AstCond(truep->fileline(), condp, truep, falsep));
                nodep->replaceWith(ifp);
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            } else if (false  // Disabled, as vpm assertions are faster
                              // without due to short-circuiting
                       && operandIfIf(nodep)) {
                UINFO(9, "IF({a}) IF({b}) => IF({a} && {b})" << endl);
                AstNodeIf* lowerIfp = VN_CAST(nodep->ifsp(), NodeIf);
                AstNode* condp = nodep->condp()->unlinkFrBack();
                AstNode* lowerIfsp = lowerIfp->ifsp()->unlinkFrBackWithNext();
                AstNode* lowerCondp = lowerIfp->condp()->unlinkFrBackWithNext();
                nodep->condp(new AstLogAnd(lowerIfp->fileline(), condp, lowerCondp));
                lowerIfp->replaceWith(lowerIfsp);
                VL_DO_DANGLING(lowerIfp->deleteTree(), lowerIfp);
            } else if (operandBoolShift(nodep->condp())) {
                replaceBoolShift(nodep->condp());
            }
        }
    }

    virtual void visit(AstDisplay* nodep) override {
        // DISPLAY(SFORMAT(text1)),DISPLAY(SFORMAT(text2)) -> DISPLAY(SFORMAT(text1+text2))
        iterateChildren(nodep);
        if (stmtDisplayDisplay(nodep)) return;
    }
    bool stmtDisplayDisplay(AstDisplay* nodep) {
        // DISPLAY(SFORMAT(text1)),DISPLAY(SFORMAT(text2)) -> DISPLAY(SFORMAT(text1+text2))
        if (!m_modp) return false;  // Don't optimize under single statement
        AstDisplay* prevp = VN_CAST(nodep->backp(), Display);
        if (!prevp) return false;
        if (!((prevp->displayType() == nodep->displayType())
              || (prevp->displayType() == AstDisplayType::DT_WRITE
                  && nodep->displayType() == AstDisplayType::DT_DISPLAY)
              || (prevp->displayType() == AstDisplayType::DT_DISPLAY
                  && nodep->displayType() == AstDisplayType::DT_WRITE)))
            return false;
        if ((prevp->filep() && !nodep->filep()) || (!prevp->filep() && nodep->filep())
            || !prevp->filep()->sameTree(nodep->filep()))
            return false;
        if (!prevp->fmtp() || prevp->fmtp()->nextp() || !nodep->fmtp() || nodep->fmtp()->nextp())
            return false;
        AstSFormatF* pformatp = prevp->fmtp();
        if (!pformatp) return false;
        AstSFormatF* nformatp = nodep->fmtp();
        if (!nformatp) return false;
        // We don't merge scopeNames as can have only one and might be different scopes (late in
        // process) Also rare for real code to print %m multiple times in same message
        if (nformatp->scopeNamep() && pformatp->scopeNamep()) return false;
        // We don't early merge arguments as might need to later print warnings with
        // right line numbers, nor scopeNames as might be different scopes (late in process)
        if (!m_doCpp && pformatp->exprsp()) return false;
        if (!m_doCpp && nformatp->exprsp()) return false;
        // Avoid huge merges
        static constexpr int DISPLAY_MAX_MERGE_LENGTH = 500;
        if (pformatp->text().length() + nformatp->text().length() > DISPLAY_MAX_MERGE_LENGTH)
            return false;
        //
        UINFO(9, "DISPLAY(SF({a})) DISPLAY(SF({b})) -> DISPLAY(SF({a}+{b}))" << endl);
        // Convert DT_DISPLAY to DT_WRITE as may allow later optimizations
        if (prevp->displayType() == AstDisplayType::DT_DISPLAY) {
            prevp->displayType(AstDisplayType::DT_WRITE);
            pformatp->text(pformatp->text() + "\n");
        }
        // We can't replace prev() as the edit tracking iterators will get confused.
        // So instead we edit the prev note itself.
        if (prevp->addNewline()) pformatp->text(pformatp->text() + "\n");
        pformatp->text(pformatp->text() + nformatp->text());
        if (!prevp->addNewline() && nodep->addNewline()) pformatp->text(pformatp->text() + "\n");
        if (nformatp->exprsp()) pformatp->addExprsp(nformatp->exprsp()->unlinkFrBackWithNext());
        if (nformatp->scopeNamep())
            pformatp->scopeNamep(nformatp->scopeNamep()->unlinkFrBackWithNext());
        VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
        return true;
    }
    virtual void visit(AstSFormatF* nodep) override {
        // Substitute constants into displays.  The main point of this is to
        // simplify assertion methodologies which call functions with display's.
        // This eliminates a pile of wide temps, and makes the C a whole lot more readable.
        iterateChildren(nodep);
        bool anyconst = false;
        for (AstNode* argp = nodep->exprsp(); argp; argp = argp->nextp()) {
            if (VN_IS(argp, Const)) {
                anyconst = true;
                break;
            }
        }
        if (m_doNConst && anyconst) {
            // UINFO(9,"  Display in  "<<nodep->text()<<endl);
            string newFormat;
            string fmt;
            bool inPct = false;
            AstNode* argp = nodep->exprsp();
            string text = nodep->text();
            for (const char ch : text) {
                if (!inPct && ch == '%') {
                    inPct = true;
                    fmt = ch;
                } else if (inPct && (isdigit(ch) || ch == '.' || ch == '-')) {
                    fmt += ch;
                } else if (inPct) {
                    inPct = false;
                    fmt += ch;
                    switch (tolower(ch)) {
                    case '%': break;  // %% - just output a %
                    case 'm': break;  // %m - auto insert "name"
                    case 'l': break;  // %l - auto insert "library"
                    case 't':  // FALLTHRU
                    case '^':  // %t/%^ - don't know $timeformat so can't constify
                        if (argp) argp = argp->nextp();
                        break;
                    default:  // Most operators, just move to next argument
                        if (argp) {
                            AstNode* nextp = argp->nextp();
                            if (VN_IS(argp, Const)) {  // Convert it
                                string out = VN_CAST(argp, Const)->num().displayed(nodep, fmt);
                                UINFO(9, "     DispConst: " << fmt << " -> " << out << "  for "
                                                            << argp << endl);
                                // fmt = out w/ replace % with %% as it must be literal.
                                fmt = VString::quotePercent(out);
                                VL_DO_DANGLING(argp->unlinkFrBack()->deleteTree(), argp);
                            }
                            argp = nextp;
                        }
                        break;
                    }  // switch
                    newFormat += fmt;
                } else {
                    newFormat += ch;
                }
            }
            if (newFormat != nodep->text()) {
                nodep->text(newFormat);
                UINFO(9, "  Display out " << nodep << endl);
            }
        }
        if (!nodep->exprsp() && nodep->name().find('%') == string::npos && !nodep->hidden()) {
            // Just a simple constant string - the formatting is pointless
            VL_DO_DANGLING(replaceConstString(nodep, nodep->name()), nodep);
        }
    }

    virtual void visit(AstFuncRef* nodep) override {
        iterateChildren(nodep);
        if (m_params) {  // Only parameters force us to do constant function call propagation
            replaceWithSimulation(nodep);
        }
    }
    virtual void visit(AstArg* nodep) override {
        // replaceWithSimulation on the Arg's parent FuncRef replaces these
        iterateChildren(nodep);
    }
    virtual void visit(AstWhile* nodep) override {
        bool oldHasJumpDelay = m_hasJumpDelay;
        m_hasJumpDelay = false;
        { iterateChildren(nodep); }
        bool thisWhileHasJumpDelay = m_hasJumpDelay;
        m_hasJumpDelay = thisWhileHasJumpDelay || oldHasJumpDelay;
        if (m_doNConst) {
            if (nodep->condp()->isZero()) {
                UINFO(4, "WHILE(0) => nop " << nodep << endl);
                if (nodep->precondsp()) {
                    nodep->replaceWith(nodep->precondsp());
                } else {
                    nodep->unlinkFrBack();
                }
                VL_DO_DANGLING(nodep->deleteTree(), nodep);
            } else if (nodep->condp()->isNeqZero()) {
                if (!thisWhileHasJumpDelay) {
                    nodep->v3warn(INFINITELOOP, "Infinite loop (condition always true)");
                    nodep->fileline()->modifyWarnOff(V3ErrorCode::INFINITELOOP,
                                                     true);  // Complain just once
                }
            } else if (operandBoolShift(nodep->condp())) {
                replaceBoolShift(nodep->condp());
            }
        }
    }
    virtual void visit(AstInitArray* nodep) override { iterateChildren(nodep); }
    virtual void visit(AstInitItem* nodep) override { iterateChildren(nodep); }
    virtual void visit(AstUnbounded* nodep) override { iterateChildren(nodep); }
    // These are converted by V3Param.  Don't constify as we don't want the
    // from() VARREF to disappear, if any.
    // If output of a presel didn't get consted, chances are V3Param didn't visit properly
    virtual void visit(AstNodePreSel*) override {}

    // Ignored, can eliminate early
    virtual void visit(AstSysIgnore* nodep) override {
        iterateChildren(nodep);
        if (m_doNConst) VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
    }

    // Simplify
    virtual void visit(AstBasicDType* nodep) override {
        iterateChildren(nodep);
        nodep->cvtRangeConst();
    }

    //-----
    // Jump elimination

    virtual void visit(AstDelay* nodep) override {
        iterateChildren(nodep);
        m_hasJumpDelay = true;
    }
    virtual void visit(AstJumpGo* nodep) override {
        iterateChildren(nodep);
        // Jump to label where label immediately follows label is not useful
        if (nodep->labelp() == VN_CAST(nodep->nextp(), JumpLabel)) {
            VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
            // Keep the label, might be other jumps pointing to it, gets cleaned later
            return;
        }
        if (m_doExpensive) {
            // Any non-label statements (at this statement level) can never execute
            while (nodep->nextp() && !VN_IS(nodep->nextp(), JumpLabel)) {
                nodep->nextp()->unlinkFrBack()->deleteTree();
            }
            // If last statement in a jump label we have JumpLabel(...., JumpGo)
            // Often caused by "return" in a Verilog function.  The Go is pointless, remove.
            if (!nodep->nextp()) {
                if (AstJumpBlock* aboveBlockp = VN_CAST(nodep->abovep(), JumpBlock)) {
                    if (aboveBlockp == nodep->labelp()->blockp()) {
                        if (aboveBlockp->endStmtsp() == nodep->labelp()) {
                            UINFO(4, "JUMPGO => last remove " << nodep << endl);
                            VL_DO_DANGLING(nodep->unlinkFrBack()->deleteTree(), nodep);
                            return;
                        }
                    }
                }
            }
            nodep->labelp()->blockp()->user4(true);
        }
        m_hasJumpDelay = true;
    }

    virtual void visit(AstJumpBlock* nodep) override {
        // Because JumpLabels disable many optimizations,
        // remove JumpLabels that are not pointed to by any AstJumpGos
        // Note this assumes all AstJumpGos are underneath the given label; V3Broken asserts this
        iterateChildren(nodep);
        // AstJumpGo's below here that point to this node will set user4
        if (m_doExpensive && !nodep->user4()) {
            UINFO(4, "JUMPLABEL => unused " << nodep << endl);
            AstNode* underp = nullptr;
            if (nodep->stmtsp()) underp = nodep->stmtsp()->unlinkFrBackWithNext();
            if (underp) {
                nodep->replaceWith(underp);
            } else {
                nodep->unlinkFrBack();
            }
            nodep->labelp()->unlinkFrBack()->deleteTree();
            VL_DO_DANGLING(nodep->deleteTree(), nodep);
        }
    }

    //-----
    // Below lines are magic expressions processed by astgen
    //  TREE_SKIP_VISIT("AstNODETYPE")    # Rename normal visit to visitGen and don't iterate
    //-----
    // clang-format off
//    TREE_SKIP_VISIT("ArraySel");
#line 2688 "V3Const__gen.cpp"
#line 2685 "../V3Const.cpp"

    //-----
    //  "AstNODETYPE {             # bracket not paren
    //                $accessor_name, ...
    //                             # .castFoo is the test VN_IS(object,Foo)
    //                             # ,, gets replaced with a , rather than &&
    //               }"            # bracket not paren
    //    ,"what to call"
    //
    // Where "what_to_call" is:
    //          "function to call"
    //          "AstREPLACEMENT_TYPE{ $accessor }"
    //          "!              # Print line number when matches, so can see operations
    //          "NEVER"         # Print error message
    //          "DONE"          # Process of matching did the transform already

    // In the future maybe support more complicated match & replace:
    //   ("AstOr  {%a, AstAnd{AstNot{%b}, %c}} if %a.width1 if %a==%b", "AstOr{%a,%c}; %b.delete");
    // Lhs/rhs would be implied; for non math operations you'd need $lhsp etc.

    //    v--- * * This op done on Verilog or C+++ mode, in all non-m_doConst stages
    //    v--- *1* These ops are always first, as we warn before replacing
    //    v--- *C* This op is a (C)++ op, only in m_doCpp mode
    //    v--- *V* This op is a (V)erilog op, only in m_doV mode
    //    v--- *A* This op works on (A)ll constant children, allowed in m_doConst mode
    //    v--- *S* This op specifies a type should use (S)hort-circuiting of its lhs op

//    TREEOP1("AstSel{warnSelect(nodep)}",        "NEVER");
#line 2718 "V3Const__gen.cpp"
    // TREEOP functions, each return true if they matched & transformed
    // Generated by astgen
    bool match_Add_0(AstAdd* nodep) {
	// TREEOP ("AstAdd   {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAdd $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Add_1(AstAdd* nodep) {
	// TREEOP ("AstAdd   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAdd $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_0(AstAnd* nodep) {
	// TREEOP ("AstAnd   {$lhsp.isZero, $rhsp, isTPure($rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero() && isTPure(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp.isZero, $rhsp, isTPure($rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_1(AstAnd* nodep) {
	// TREEOP ("AstAnd   {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp, $rhsp.isZero , replaceZeroChkPure(nodep,$lhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_2(AstAnd* nodep) {
	// TREEOP ("AstAnd   {$lhsp.isAllOnes, $rhsp}",        "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp.isAllOnes, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_3(AstAnd* nodep) {
	// TREEOP ("AstAnd   {$lhsp, $rhsp.isAllOnes}",        "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp, $rhsp.isAllOnes , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_4(AstAnd* nodep) {
	// TREEOPC("AstAnd   {$lhsp.isOne, matchRedundantClean(nodep)}", "DONE")
	if (m_doCpp && nodep->lhsp()->isOne() && matchRedundantClean(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPC( AstAnd $lhsp.isOne, matchRedundantClean(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_5(AstAnd* nodep) {
	// TREEOP ("AstAnd    {operandsSame($lhsp,,$rhsp)}",   "replaceWLhs(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd operandsSame($lhsp,,$rhsp) , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_6(AstAnd* nodep) {
	// TREEOPC("AstAnd {nodep->widthMin() == 1, $lhsp.castConst, $rhsp.castRedXor, matchBitOpTree(nodep)}", "DONE")
	if (m_doCpp && nodep->widthMin() == 1 && VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),RedXor) && matchBitOpTree(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPC( AstAnd nodep->widthMin() == 1, $lhsp.castConst, $rhsp.castRedXor, matchBitOpTree(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_7(AstAnd* nodep) {
	// TREEOP ("AstAnd {$lhsp.castConst,matchAndCond(nodep)}",           "DONE")
	if (m_doNConst && VN_IS(nodep->lhsp(),Const) && matchAndCond(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp.castConst,matchAndCond(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_8(AstAnd* nodep) {
	// TREEOP ("AstAnd {$lhsp.castOr, $rhsp.castOr, operandAndOrSame(nodep)}",     "replaceAndOr(nodep)")
	if (m_doNConst && VN_IS(nodep->lhsp(),Or) && VN_IS(nodep->rhsp(),Or) && operandAndOrSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd $lhsp.castOr, $rhsp.castOr, operandAndOrSame(nodep) , replaceAndOr(nodep) )\n");
	    replaceAndOr(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_9(AstAnd* nodep) {
	// TREEOP ("AstAnd {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)")
	if (m_doNConst && operandShiftSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstAnd operandShiftSame(nodep) , replaceShiftSame(nodep) )\n");
	    replaceShiftSame(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_And_10(AstAnd* nodep) {
	// TREEOPC("AstAnd {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE")
	if (m_doCpp && nodep->widthMin() == 1 && matchBitOpTree(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPC( AstAnd nodep->widthMin() == 1, matchBitOpTree(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ChangeXor_0(AstChangeXor* nodep) {
	// TREEOP ("AstChangeXor{operandsSame($lhsp,,$rhsp)}", "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstChangeXor operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Concat_0(AstConcat* nodep) {
	// TREEOPV("AstConcat{operandConcatMove(nodep)}",      "moveConcat(nodep)")
	if (m_doV && operandConcatMove(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstConcat operandConcatMove(nodep) , moveConcat(nodep) )\n");
	    moveConcat(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Concat_1(AstConcat* nodep) {
	// TREEOPV("AstConcat{$lhsp.isZero, $rhsp}",           "replaceExtend(nodep, nodep->rhsp())")
	if (m_doV && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstConcat $lhsp.isZero, $rhsp , replaceExtend(nodep, nodep->rhsp()) )\n");
	    replaceExtend(nodep, nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Concat_2(AstConcat* nodep) {
	// TREEOPV("AstConcat{$lhsp.castSel, $rhsp.castSel, ifAdjacentSel(VN_CAST($lhsp,,Sel),,VN_CAST($rhsp,,Sel))}",  "replaceConcatSel(nodep)")
	if (m_doV && VN_IS(nodep->lhsp(),Sel) && VN_IS(nodep->rhsp(),Sel) && ifAdjacentSel(VN_CAST(nodep->lhsp(),Sel),VN_CAST(nodep->rhsp(),Sel))) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstConcat $lhsp.castSel, $rhsp.castSel, ifAdjacentSel(VN_CAST($lhsp,,Sel),,VN_CAST($rhsp,,Sel)) , replaceConcatSel(nodep) )\n");
	    replaceConcatSel(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Concat_3(AstConcat* nodep) {
	// TREEOPV("AstConcat{ifConcatMergeableBiop($lhsp), concatMergeable($lhsp,,$rhsp)}", "replaceConcatMerge(nodep)")
	if (m_doV && ifConcatMergeableBiop(nodep->lhsp()) && concatMergeable(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstConcat ifConcatMergeableBiop($lhsp), concatMergeable($lhsp,,$rhsp) , replaceConcatMerge(nodep) )\n");
	    replaceConcatMerge(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Concat_4(AstConcat* nodep) {
	// TREEOPV("AstConcat{operandConcatSame(nodep)}", "DONE")
	if (m_doV && operandConcatSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstConcat operandConcatSame(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Cond_0(AstCond* nodep) {
	// TREEOPS("AstCond {$lhsp.isZero}",           "replaceWIteratedThs(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPS( AstCond $lhsp.isZero , replaceWIteratedThs(nodep) )\n");
	    replaceWIteratedThs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Cond_1(AstCond* nodep) {
	// TREEOPS("AstCond {$lhsp.isNeqZero}",        "replaceWIteratedRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isNeqZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPS( AstCond $lhsp.isNeqZero , replaceWIteratedRhs(nodep) )\n");
	    replaceWIteratedRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Cond_2(AstCond* nodep) {
	// TREEOP ("AstCond{$condp.castNot,       $expr1p, $expr2p}", "AstCond{$condp->op1p(), $expr2p, $expr1p}")
	if (m_doNConst && VN_IS(nodep->condp(),Not)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstCond $condp.castNot, $expr1p, $expr2p , AstCond $condp->op1p(), $expr2p, $expr1p )\n");
	    AstNode* arg1p = nodep->condp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->expr2p()->unlinkFrBack();
AstNode* arg3p = nodep->expr1p()->unlinkFrBack();
AstNode* newp = new AstCond(nodep->fileline(), arg1p, arg2p, arg3p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_CvtPackString_0(AstCvtPackString* nodep) {
	// TREEOPA("AstCvtPackString{$lhsp.castConst}", "replaceConstString(nodep, VN_CAST(nodep->lhsp(), Const)->num().toString())")
	if (VN_IS(nodep->lhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstCvtPackString $lhsp.castConst , replaceConstString(nodep, VN_CAST(nodep->lhsp(), Const)->num().toString()) )\n");
	    replaceConstString(nodep, VN_CAST(nodep->lhsp(), Const)->num().toString());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Div_0(AstDiv* nodep) {
	// TREEOP ("AstDiv   {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDiv $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Div_1(AstDiv* nodep) {
	// TREEOP ("AstDiv   {$lhsp, $rhsp.isOne}",    "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isOne()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDiv $lhsp, $rhsp.isOne , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Div_2(AstDiv* nodep) {
	// TREEOP ("AstDiv   {$lhsp, operandIsPowTwo($rhsp)}", "replaceDivShift(nodep)")
	if (m_doNConst && operandIsPowTwo(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDiv $lhsp, operandIsPowTwo($rhsp) , replaceDivShift(nodep) )\n");
	    replaceDivShift(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Div_3(AstDiv* nodep) {
	// TREEOP ("AstDiv    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDiv operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_DivS_0(AstDivS* nodep) {
	// TREEOP ("AstDivS  {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDivS $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_DivS_1(AstDivS* nodep) {
	// TREEOP ("AstDivS  {$lhsp, $rhsp.isOne}",    "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isOne()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDivS $lhsp, $rhsp.isOne , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_DivS_2(AstDivS* nodep) {
	// TREEOP ("AstDivS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstDivS operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_0(AstEq* nodep) {
	// TREEOPV("AstEq    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_1(AstEq* nodep) {
	// TREEOPV("AstEq    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_2(AstEq* nodep) {
	// TREEOP ("AstEq     {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstEq operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_3(AstEq* nodep) {
	// TREEOPV("AstEq    {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "AstNot{$rhsp}")
	if (m_doV && nodep->rhsp()->width1() && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $rhsp.width1, $lhsp.isZero, $rhsp , AstNot $rhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_4(AstEq* nodep) {
	// TREEOPV("AstEq    {$lhsp.width1, $lhsp, $rhsp.isZero}",     "AstNot{$lhsp}")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $lhsp.width1, $lhsp, $rhsp.isZero , AstNot $lhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_5(AstEq* nodep) {
	// TREEOPV("AstEq    {$rhsp.width1, $lhsp.isAllOnes, $rhsp}",  "replaceWRhs(nodep)")
	if (m_doV && nodep->rhsp()->width1() && nodep->lhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $rhsp.width1, $lhsp.isAllOnes, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Eq_6(AstEq* nodep) {
	// TREEOPV("AstEq    {$lhsp.width1, $lhsp, $rhsp.isAllOnes}",  "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstEq $lhsp.width1, $lhsp, $rhsp.isAllOnes , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_EqCase_0(AstEqCase* nodep) {
	// TREEOP ("AstEqCase {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstEqCase operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_EqD_0(AstEqD* nodep) {
	// TREEOP ("AstEqD    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstEqD operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_EqN_0(AstEqN* nodep) {
	// TREEOP ("AstEqN    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstEqN operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_EqWild_0(AstEqWild* nodep) {
	// TREEOP ("AstEqWild {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstEqWild operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Extend_0(AstExtend* nodep) {
	// TREEOPV("AstExtend {$lhsp.castExtend}",  "replaceExtend(nodep, VN_CAST(nodep->lhsp(), Extend)->lhsp())")
	if (m_doV && VN_IS(nodep->lhsp(),Extend)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstExtend $lhsp.castExtend , replaceExtend(nodep, VN_CAST(nodep->lhsp(), Extend)->lhsp()) )\n");
	    replaceExtend(nodep, VN_CAST(nodep->lhsp(), Extend)->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ExtendS_0(AstExtendS* nodep) {
	// TREEOPV("AstExtendS{$lhsp.castExtendS}", "replaceExtend(nodep, VN_CAST(nodep->lhsp(), ExtendS)->lhsp())")
	if (m_doV && VN_IS(nodep->lhsp(),ExtendS)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstExtendS $lhsp.castExtendS , replaceExtend(nodep, VN_CAST(nodep->lhsp(), ExtendS)->lhsp()) )\n");
	    replaceExtend(nodep, VN_CAST(nodep->lhsp(), ExtendS)->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_0(AstGt* nodep) {
	// TREEOP ("AstGt   {!$lhsp.castConst,$rhsp.castConst}",       "AstLt  {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGt !$lhsp.castConst,$rhsp.castConst , AstLt $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstLt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_1(AstGt* nodep) {
	// TREEOP1("AstGt   {$lhsp.isZero, $rhsp}",            "replaceNumSigned(nodep,0)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstGt $lhsp.isZero, $rhsp , replaceNumSigned(nodep,0) )\n");
	    replaceNumSigned(nodep,0);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_2(AstGt* nodep) {
	// TREEOP1("AstGt   {$lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,0)")
	if (m_doNConst && nodep->rhsp()->isAllOnes() && nodep->lhsp()->width()==nodep->rhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstGt $lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width() , replaceNumLimited(nodep,0) )\n");
	    replaceNumLimited(nodep,0);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_3(AstGt* nodep) {
	// TREEOPV("AstGt    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstGt $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_4(AstGt* nodep) {
	// TREEOPV("AstGt    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstGt $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_5(AstGt* nodep) {
	// TREEOP ("AstGt     {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGt operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gt_6(AstGt* nodep) {
	// TREEOPV("AstGt    {$lhsp.width1, $lhsp, $rhsp.isZero}",     "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstGt $lhsp.width1, $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GtD_0(AstGtD* nodep) {
	// TREEOP ("AstGtD    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGtD operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GtN_0(AstGtN* nodep) {
	// TREEOP ("AstGtN    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGtN operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GtS_0(AstGtS* nodep) {
	// TREEOP ("AstGtS  {!$lhsp.castConst,$rhsp.castConst}",       "AstLtS {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGtS !$lhsp.castConst,$rhsp.castConst , AstLtS $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstLtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GtS_1(AstGtS* nodep) {
	// TREEOP ("AstGtS    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGtS operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_0(AstGte* nodep) {
	// TREEOP ("AstGte  {!$lhsp.castConst,$rhsp.castConst}",       "AstLte {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGte !$lhsp.castConst,$rhsp.castConst , AstLte $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstLte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_1(AstGte* nodep) {
	// TREEOP1("AstGte  {$lhsp, $rhsp.isZero}",            "replaceNumSigned(nodep,1)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstGte $lhsp, $rhsp.isZero , replaceNumSigned(nodep,1) )\n");
	    replaceNumSigned(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_2(AstGte* nodep) {
	// TREEOP1("AstGte  {$lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,1)")
	if (m_doNConst && nodep->lhsp()->isAllOnes() && nodep->lhsp()->width()==nodep->rhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstGte $lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width() , replaceNumLimited(nodep,1) )\n");
	    replaceNumLimited(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_3(AstGte* nodep) {
	// TREEOPV("AstGte   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstGte $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_4(AstGte* nodep) {
	// TREEOPV("AstGte   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstGte $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Gte_5(AstGte* nodep) {
	// TREEOP ("AstGte    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGte operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GteD_0(AstGteD* nodep) {
	// TREEOP ("AstGteD   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGteD operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GteN_0(AstGteN* nodep) {
	// TREEOP ("AstGteN   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGteN operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GteS_0(AstGteS* nodep) {
	// TREEOP ("AstGteS {!$lhsp.castConst,$rhsp.castConst}",       "AstLteS{$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGteS !$lhsp.castConst,$rhsp.castConst , AstLteS $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstLteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_GteS_1(AstGteS* nodep) {
	// TREEOP ("AstGteS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstGteS operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_IsUnbounded_0(AstIsUnbounded* nodep) {
	// TREEOPV("AstIsUnbounded{$lhsp.castUnbounded}", "replaceNum(nodep, 1)")
	if (m_doV && VN_IS(nodep->lhsp(),Unbounded)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstIsUnbounded $lhsp.castUnbounded , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_0(AstLogAnd* nodep) {
	// TREEOPS("AstLogAnd   {$lhsp.isZero}",       "replaceZero(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPS( AstLogAnd $lhsp.isZero , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_1(AstLogAnd* nodep) {
	// TREEOP ("AstLogAnd{$lhsp.isZero, $rhsp}",   "replaceZero(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogAnd $lhsp.isZero, $rhsp , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_2(AstLogAnd* nodep) {
	// TREEOP ("AstLogAnd{$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogAnd $lhsp, $rhsp.isZero , replaceZeroChkPure(nodep,$lhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_3(AstLogAnd* nodep) {
	// TREEOP ("AstLogAnd{$lhsp.isNeqZero, $rhsp}",        "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isNeqZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogAnd $lhsp.isNeqZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_4(AstLogAnd* nodep) {
	// TREEOP ("AstLogAnd{$lhsp, $rhsp.isNeqZero}",        "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isNeqZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogAnd $lhsp, $rhsp.isNeqZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_5(AstLogAnd* nodep) {
	// TREEOP ("AstLogAnd {operandsSame($lhsp,,$rhsp), $lhsp.width1}",     "replaceWLhs(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp()) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogAnd operandsSame($lhsp,,$rhsp), $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogAnd_6(AstLogAnd* nodep) {
	// TREEOPV("AstLogAnd{$lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp)}", "AstAnd{$lhsp,$rhsp}")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->width1() && isTPure(nodep->lhsp()) && isTPure(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLogAnd $lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp) , AstAnd $lhsp,$rhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* arg2p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstAnd(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogEq_0(AstLogEq* nodep) {
	// TREEOPV("AstLogEq{$lhsp, $rhsp}",  "replaceLogEq(nodep)")
	if (m_doV) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLogEq $lhsp, $rhsp , replaceLogEq(nodep) )\n");
	    replaceLogEq(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogIf_0(AstLogIf* nodep) {
	// TREEOPS("AstLogIf{$lhsp.isZero}",  "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPS( AstLogIf $lhsp.isZero , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogIf_1(AstLogIf* nodep) {
	// TREEOPV("AstLogIf{$lhsp, $rhsp}",  "AstLogOr{AstLogNot{$lhsp},$rhsp}")
	if (m_doV) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLogIf $lhsp, $rhsp , AstLogOr AstLogNot $lhsp ,$rhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* arg2p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstLogOr(nodep->fileline(), new AstLogNot(nodep->fileline(), arg1p), arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_0(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castLogNot}",              "replaceWChild(nodep, $lhsp->op1p())")
	if (m_doNConst && VN_IS(nodep->lhsp(),LogNot)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castLogNot , replaceWChild(nodep, $lhsp->op1p()) )\n");
	    replaceWChild(nodep, nodep->lhsp()->op1p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_1(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castEqCase}",              "AstNeqCase{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),EqCase)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castEqCase , AstNeqCase $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeqCase(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_2(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castNeqCase}",             "AstEqCase {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),NeqCase)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castNeqCase , AstEqCase $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEqCase(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_3(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castEqWild}",              "AstNeqWild{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),EqWild)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castEqWild , AstNeqWild $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeqWild(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_4(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castNeqWild}",             "AstEqWild {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),NeqWild)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castNeqWild , AstEqWild $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEqWild(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_5(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castEq}",                  "AstNeq {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Eq)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castEq , AstNeq $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeq(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_6(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castNeq}",                 "AstEq  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Neq)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castNeq , AstEq $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEq(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_7(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castLt}",                  "AstGte {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Lt)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castLt , AstGte $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_8(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castLtS}",                 "AstGteS{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),LtS)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castLtS , AstGteS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_9(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castLte}",                 "AstGt  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Lte)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castLte , AstGt $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_10(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castLteS}",                "AstGtS {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),LteS)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castLteS , AstGtS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_11(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castGt}",                  "AstLte {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Gt)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castGt , AstLte $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_12(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castGtS}",                 "AstLteS{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),GtS)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castGtS , AstLteS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_13(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castGte}",                 "AstLt  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Gte)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castGte , AstLt $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_14(AstLogNot* nodep) {
	// TREEOP ("AstLogNot{$lhsp.castGteS}",                "AstLtS {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),GteS)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogNot $lhsp.castGteS , AstLtS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogNot_15(AstLogNot* nodep) {
	// TREEOPV("AstLogNot{$lhsp.width1, isTPure($lhsp)}",  "AstNot{$lhsp}")
	if (m_doV && nodep->lhsp()->width1() && isTPure(nodep->lhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLogNot $lhsp.width1, isTPure($lhsp) , AstNot $lhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_0(AstLogOr* nodep) {
	// TREEOPS("AstLogOr   {$lhsp.isOne}",         "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->lhsp()->isOne()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPS( AstLogOr $lhsp.isOne , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_1(AstLogOr* nodep) {
	// TREEOP ("AstLogOr {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogOr $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_2(AstLogOr* nodep) {
	// TREEOP ("AstLogOr {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogOr $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_3(AstLogOr* nodep) {
	// TREEOP ("AstLogOr {$lhsp.isNeqZero, $rhsp}",        "replaceNum(nodep,1)")
	if (m_doNConst && nodep->lhsp()->isNeqZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogOr $lhsp.isNeqZero, $rhsp , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_4(AstLogOr* nodep) {
	// TREEOP ("AstLogOr {$lhsp, $rhsp.isNeqZero, isTPure($lhsp)}",        "replaceNum(nodep,1)")
	if (m_doNConst && nodep->rhsp()->isNeqZero() && isTPure(nodep->lhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogOr $lhsp, $rhsp.isNeqZero, isTPure($lhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_5(AstLogOr* nodep) {
	// TREEOP ("AstLogOr  {operandsSame($lhsp,,$rhsp), $lhsp.width1}",     "replaceWLhs(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp()) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLogOr operandsSame($lhsp,,$rhsp), $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LogOr_6(AstLogOr* nodep) {
	// TREEOPV("AstLogOr {$lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp)}", "AstOr{$lhsp,$rhsp}")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->width1() && isTPure(nodep->lhsp()) && isTPure(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLogOr $lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp) , AstOr $lhsp,$rhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* arg2p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstOr(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_0(AstLt* nodep) {
	// TREEOP ("AstLt   {!$lhsp.castConst,$rhsp.castConst}",       "AstGt  {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLt !$lhsp.castConst,$rhsp.castConst , AstGt $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstGt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_1(AstLt* nodep) {
	// TREEOP1("AstLt   {$lhsp, $rhsp.isZero}",            "replaceNumSigned(nodep,0)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstLt $lhsp, $rhsp.isZero , replaceNumSigned(nodep,0) )\n");
	    replaceNumSigned(nodep,0);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_2(AstLt* nodep) {
	// TREEOP1("AstLt   {$lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,0)")
	if (m_doNConst && nodep->lhsp()->isAllOnes() && nodep->lhsp()->width()==nodep->rhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstLt $lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width() , replaceNumLimited(nodep,0) )\n");
	    replaceNumLimited(nodep,0);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_3(AstLt* nodep) {
	// TREEOPV("AstLt    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLt $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_4(AstLt* nodep) {
	// TREEOPV("AstLt    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLt $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_5(AstLt* nodep) {
	// TREEOP ("AstLt     {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLt operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lt_6(AstLt* nodep) {
	// TREEOPV("AstLt    {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "replaceWRhs(nodep)")
	if (m_doV && nodep->rhsp()->width1() && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLt $rhsp.width1, $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LtD_0(AstLtD* nodep) {
	// TREEOP ("AstLtD    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLtD operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LtN_0(AstLtN* nodep) {
	// TREEOP ("AstLtN    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLtN operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LtS_0(AstLtS* nodep) {
	// TREEOP ("AstLtS  {!$lhsp.castConst,$rhsp.castConst}",       "AstGtS {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLtS !$lhsp.castConst,$rhsp.castConst , AstGtS $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstGtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LtS_1(AstLtS* nodep) {
	// TREEOP ("AstLtS    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLtS operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_0(AstLte* nodep) {
	// TREEOP ("AstLte  {!$lhsp.castConst,$rhsp.castConst}",       "AstGte {$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLte !$lhsp.castConst,$rhsp.castConst , AstGte $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstGte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_1(AstLte* nodep) {
	// TREEOP1("AstLte  {$lhsp.isZero, $rhsp}",            "replaceNumSigned(nodep,1)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstLte $lhsp.isZero, $rhsp , replaceNumSigned(nodep,1) )\n");
	    replaceNumSigned(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_2(AstLte* nodep) {
	// TREEOP1("AstLte  {$lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,1)")
	if (m_doNConst && nodep->rhsp()->isAllOnes() && nodep->lhsp()->width()==nodep->rhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstLte $lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width() , replaceNumLimited(nodep,1) )\n");
	    replaceNumLimited(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_3(AstLte* nodep) {
	// TREEOPV("AstLte   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLte $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_4(AstLte* nodep) {
	// TREEOPV("AstLte   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLte $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_5(AstLte* nodep) {
	// TREEOP ("AstLte    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLte operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Lte_6(AstLte* nodep) {
	// TREEOPV("AstLte   {$lhsp->width()==$rhsp->width(), $rhsp.isAllOnes}", "replaceNum(nodep,1)")
	if (m_doV && nodep->lhsp()->width()==nodep->rhsp()->width() && nodep->rhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstLte $lhsp->width()==$rhsp->width(), $rhsp.isAllOnes , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LteD_0(AstLteD* nodep) {
	// TREEOP ("AstLteD   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLteD operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LteN_0(AstLteN* nodep) {
	// TREEOP ("AstLteN   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLteN operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LteS_0(AstLteS* nodep) {
	// TREEOP ("AstLteS {!$lhsp.castConst,$rhsp.castConst}",       "AstGteS{$rhsp,$lhsp}")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLteS !$lhsp.castConst,$rhsp.castConst , AstGteS $rhsp,$lhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstGteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_LteS_1(AstLteS* nodep) {
	// TREEOP ("AstLteS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstLteS operandsSame($lhsp,,$rhsp) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ModDiv_0(AstModDiv* nodep) {
	// TREEOP ("AstModDiv{$lhsp, operandIsPowTwo($rhsp)}", "replaceModAnd(nodep)")
	if (m_doNConst && operandIsPowTwo(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstModDiv $lhsp, operandIsPowTwo($rhsp) , replaceModAnd(nodep) )\n");
	    replaceModAnd(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Mul_0(AstMul* nodep) {
	// TREEOP ("AstMul   {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMul $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Mul_1(AstMul* nodep) {
	// TREEOP ("AstMul   {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMul $lhsp, $rhsp.isZero , replaceZeroChkPure(nodep,$lhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Mul_2(AstMul* nodep) {
	// TREEOP ("AstMul   {$lhsp.isOne, $rhsp}",    "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isOne()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMul $lhsp.isOne, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Mul_3(AstMul* nodep) {
	// TREEOP ("AstMul   {operandIsPowTwo($lhsp), operandsSameSize($lhsp,,$rhsp)}", "replaceMulShift(nodep)")
	if (m_doNConst && operandIsPowTwo(nodep->lhsp()) && operandsSameSize(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMul operandIsPowTwo($lhsp), operandsSameSize($lhsp,,$rhsp) , replaceMulShift(nodep) )\n");
	    replaceMulShift(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_MulS_0(AstMulS* nodep) {
	// TREEOP ("AstMulS  {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMulS $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_MulS_1(AstMulS* nodep) {
	// TREEOP ("AstMulS  {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMulS $lhsp, $rhsp.isZero , replaceZeroChkPure(nodep,$lhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->lhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_MulS_2(AstMulS* nodep) {
	// TREEOP ("AstMulS  {$lhsp.isOne, $rhsp}",    "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isOne()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstMulS $lhsp.isOne, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_0(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstShrink(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $rhsp.castExtend,operandBiExtendConstShrink(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_1(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)")
	if (m_doV && VN_IS(nodep->rhsp(),Extend) && operandBiExtendConstOver(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $rhsp.castExtend,operandBiExtendConstOver(nodep) , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_2(AstNeq* nodep) {
	// TREEOP ("AstNeq    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNeq operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_3(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "replaceWRhs(nodep)")
	if (m_doV && nodep->rhsp()->width1() && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $rhsp.width1, $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_4(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$lhsp.width1, $lhsp, $rhsp.isZero}",     "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $lhsp.width1, $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_5(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$rhsp.width1, $lhsp.isAllOnes, $rhsp}",  "AstNot{$rhsp}")
	if (m_doV && nodep->rhsp()->width1() && nodep->lhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $rhsp.width1, $lhsp.isAllOnes, $rhsp , AstNot $rhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Neq_6(AstNeq* nodep) {
	// TREEOPV("AstNeq   {$lhsp.width1, $lhsp, $rhsp.isAllOnes}",  "AstNot{$lhsp}")
	if (m_doV && nodep->lhsp()->width1() && nodep->rhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNeq $lhsp.width1, $lhsp, $rhsp.isAllOnes , AstNot $lhsp )\n");
	    AstNode* arg1p = nodep->lhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NeqCase_0(AstNeqCase* nodep) {
	// TREEOP ("AstNeqCase{operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNeqCase operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NeqD_0(AstNeqD* nodep) {
	// TREEOP ("AstNeqD   {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNeqD operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NeqN_0(AstNeqN* nodep) {
	// TREEOP ("AstNeqN   {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNeqN operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NeqWild_0(AstNeqWild* nodep) {
	// TREEOP ("AstNeqWild{operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNeqWild operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiCom_0(AstNodeBiCom* nodep) {
	// TREEOP ("AstNodeBiCom{!$lhsp.castConst, $rhsp.castConst}",  "swapSides(nodep)")
	if (m_doNConst && !VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeBiCom !$lhsp.castConst, $rhsp.castConst , swapSides(nodep) )\n");
	    swapSides(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiComAsv_0(AstNodeBiComAsv* nodep) {
	// TREEOP ("AstNodeBiComAsv{operandAsvConst(nodep)}",  "replaceAsv(nodep)")
	if (m_doNConst && operandAsvConst(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeBiComAsv operandAsvConst(nodep) , replaceAsv(nodep) )\n");
	    replaceAsv(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiComAsv_1(AstNodeBiComAsv* nodep) {
	// TREEOP ("AstNodeBiComAsv{operandAsvSame(nodep)}",   "replaceAsv(nodep)")
	if (m_doNConst && operandAsvSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeBiComAsv operandAsvSame(nodep) , replaceAsv(nodep) )\n");
	    replaceAsv(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiComAsv_2(AstNodeBiComAsv* nodep) {
	// TREEOP ("AstNodeBiComAsv{operandAsvLUp(nodep)}",    "replaceAsvLUp(nodep)")
	if (m_doNConst && operandAsvLUp(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeBiComAsv operandAsvLUp(nodep) , replaceAsvLUp(nodep) )\n");
	    replaceAsvLUp(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiComAsv_3(AstNodeBiComAsv* nodep) {
	// TREEOP ("AstNodeBiComAsv{operandAsvRUp(nodep)}",    "replaceAsvRUp(nodep)")
	if (m_doNConst && operandAsvRUp(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeBiComAsv operandAsvRUp(nodep) , replaceAsvRUp(nodep) )\n");
	    replaceAsvRUp(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeBiop_0(AstNodeBiop* nodep) {
	// TREEOPA("AstNodeBiop {$lhsp.castConst, $rhsp.castConst, nodep->isPredictOptimizable()}",  "replaceConst(nodep)")
	if (VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const) && nodep->isPredictOptimizable()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstNodeBiop $lhsp.castConst, $rhsp.castConst, nodep->isPredictOptimizable() , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_0(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.isZero,       $expr1p, $expr2p}", "replaceWChild(nodep,$expr2p)")
	if (m_doNConst && nodep->condp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.isZero, $expr1p, $expr2p , replaceWChild(nodep,$expr2p) )\n");
	    replaceWChild(nodep,nodep->expr2p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_1(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.isNeqZero,    $expr1p, $expr2p}", "replaceWChild(nodep,$expr1p)")
	if (m_doNConst && nodep->condp()->isNeqZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.isNeqZero, $expr1p, $expr2p , replaceWChild(nodep,$expr1p) )\n");
	    replaceWChild(nodep,nodep->expr1p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_2(AstNodeCond* nodep) {
	// TREEOPA("AstNodeCond{$condp.isZero,       $expr1p.castConst, $expr2p.castConst}", "replaceWChild(nodep,$expr2p)")
	if (nodep->condp()->isZero() && VN_IS(nodep->expr1p(),Const) && VN_IS(nodep->expr2p(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstNodeCond $condp.isZero, $expr1p.castConst, $expr2p.castConst , replaceWChild(nodep,$expr2p) )\n");
	    replaceWChild(nodep,nodep->expr2p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_3(AstNodeCond* nodep) {
	// TREEOPA("AstNodeCond{$condp.isNeqZero,    $expr1p.castConst, $expr2p.castConst}", "replaceWChild(nodep,$expr1p)")
	if (nodep->condp()->isNeqZero() && VN_IS(nodep->expr1p(),Const) && VN_IS(nodep->expr2p(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstNodeCond $condp.isNeqZero, $expr1p.castConst, $expr2p.castConst , replaceWChild(nodep,$expr1p) )\n");
	    replaceWChild(nodep,nodep->expr1p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_4(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp, operandsSame($expr1p,,$expr2p)}","replaceWChild(nodep,$expr1p)")
	if (m_doNConst && operandsSame(nodep->expr1p(),nodep->expr2p())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp, operandsSame($expr1p,,$expr2p) , replaceWChild(nodep,$expr1p) )\n");
	    replaceWChild(nodep,nodep->expr1p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_5(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p.isAllOnes, $expr2p}", "AstLogOr {$condp, $expr2p}")
	if (m_doNConst && nodep->condp()->width1() && nodep->expr1p()->width1() && nodep->expr1p()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.width1, $expr1p.width1, $expr1p.isAllOnes, $expr2p , AstLogOr $condp, $expr2p )\n");
	    AstNode* arg1p = nodep->condp()->unlinkFrBack();
AstNode* arg2p = nodep->expr2p()->unlinkFrBack();
AstNode* newp = new AstLogOr(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_6(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p,    $expr2p.isZero}", "AstLogAnd{$condp, $expr1p}")
	if (m_doNConst && nodep->condp()->width1() && nodep->expr1p()->width1() && nodep->expr2p()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.width1, $expr1p.width1, $expr1p, $expr2p.isZero , AstLogAnd $condp, $expr1p )\n");
	    AstNode* arg1p = nodep->condp()->unlinkFrBack();
AstNode* arg2p = nodep->expr1p()->unlinkFrBack();
AstNode* newp = new AstLogAnd(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_7(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p, $expr2p.isAllOnes}", "AstLogOr {AstNot{$condp}, $expr1p}")
	if (m_doNConst && nodep->condp()->width1() && nodep->expr1p()->width1() && nodep->expr2p()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.width1, $expr1p.width1, $expr1p, $expr2p.isAllOnes , AstLogOr AstNot $condp , $expr1p )\n");
	    AstNode* arg1p = nodep->condp()->unlinkFrBack();
AstNode* arg2p = nodep->expr1p()->unlinkFrBack();
AstNode* newp = new AstLogOr(nodep->fileline(), new AstNot(nodep->fileline(), arg1p), arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_8(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p.isZero,    $expr2p}", "AstLogAnd{AstNot{$condp}, $expr2p}")
	if (m_doNConst && nodep->condp()->width1() && nodep->expr1p()->width1() && nodep->expr1p()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond $condp.width1, $expr1p.width1, $expr1p.isZero, $expr2p , AstLogAnd AstNot $condp , $expr2p )\n");
	    AstNode* arg1p = nodep->condp()->unlinkFrBack();
AstNode* arg2p = nodep->expr2p()->unlinkFrBack();
AstNode* newp = new AstLogAnd(nodep->fileline(), new AstNot(nodep->fileline(), arg1p), arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeCond_9(AstNodeCond* nodep) {
	// TREEOP ("AstNodeCond{!$condp.width1, operandBoolShift(nodep->condp())}", "replaceBoolShift(nodep->condp())")
	if (m_doNConst && !nodep->condp()->width1() && operandBoolShift(nodep->condp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNodeCond !$condp.width1, operandBoolShift(nodep->condp()) , replaceBoolShift(nodep->condp()) )\n");
	    replaceBoolShift(nodep->condp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeQuadop_0(AstNodeQuadop* nodep) {
	// TREEOPA("AstNodeQuadop{$lhsp.castConst, $rhsp.castConst, $thsp.castConst, $fhsp.castConst}",  "replaceConst(nodep)")
	if (VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const) && VN_IS(nodep->thsp(),Const) && VN_IS(nodep->fhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstNodeQuadop $lhsp.castConst, $rhsp.castConst, $thsp.castConst, $fhsp.castConst , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_NodeUniop_0(AstNodeUniop* nodep) {
	// TREEOPA("AstNodeUniop{$lhsp.castConst, !nodep->isOpaque(), nodep->isPredictOptimizable()}",  "replaceConst(nodep)")
	if (VN_IS(nodep->lhsp(),Const) && !nodep->isOpaque() && nodep->isPredictOptimizable()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstNodeUniop $lhsp.castConst, !nodep->isOpaque(), nodep->isPredictOptimizable() , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_0(AstNot* nodep) {
	// TREEOP ("AstNot   {$lhsp.castNot,  $lhsp->width()==VN_CAST($lhsp,,Not)->lhsp()->width()}", "replaceWChild(nodep, $lhsp->op1p())")
	if (m_doNConst && VN_IS(nodep->lhsp(),Not) && nodep->lhsp()->width()==VN_CAST(nodep->lhsp(),Not)->lhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstNot $lhsp.castNot, $lhsp->width()==VN_CAST($lhsp,,Not)->lhsp()->width() , replaceWChild(nodep, $lhsp->op1p()) )\n");
	    replaceWChild(nodep, nodep->lhsp()->op1p());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_1(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castEqCase, $lhsp.width1}","AstNeqCase{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),EqCase) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castEqCase, $lhsp.width1 , AstNeqCase $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeqCase(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_2(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castNeqCase, $lhsp.width1}","AstEqCase {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),NeqCase) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castNeqCase, $lhsp.width1 , AstEqCase $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEqCase(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_3(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castEqWild, $lhsp.width1}","AstNeqWild{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),EqWild) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castEqWild, $lhsp.width1 , AstNeqWild $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeqWild(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_4(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castNeqWild, $lhsp.width1}","AstEqWild {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),NeqWild) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castNeqWild, $lhsp.width1 , AstEqWild $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEqWild(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_5(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castEq, $lhsp.width1}",    "AstNeq {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Eq) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castEq, $lhsp.width1 , AstNeq $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstNeq(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_6(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castNeq, $lhsp.width1}",   "AstEq  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Neq) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castNeq, $lhsp.width1 , AstEq $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstEq(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_7(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castLt, $lhsp.width1}",    "AstGte {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Lt) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castLt, $lhsp.width1 , AstGte $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_8(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castLtS, $lhsp.width1}",   "AstGteS{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),LtS) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castLtS, $lhsp.width1 , AstGteS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_9(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castLte, $lhsp.width1}",   "AstGt  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Lte) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castLte, $lhsp.width1 , AstGt $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_10(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castLteS, $lhsp.width1}",  "AstGtS {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),LteS) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castLteS, $lhsp.width1 , AstGtS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstGtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_11(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castGt, $lhsp.width1}",    "AstLte {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Gt) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castGt, $lhsp.width1 , AstLte $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLte(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_12(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castGtS, $lhsp.width1}",   "AstLteS{$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),GtS) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castGtS, $lhsp.width1 , AstLteS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLteS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_13(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castGte, $lhsp.width1}",   "AstLt  {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),Gte) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castGte, $lhsp.width1 , AstLt $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLt(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Not_14(AstNot* nodep) {
	// TREEOPV("AstNot   {$lhsp.castGteS, $lhsp.width1}",  "AstLtS {$lhsp->op1p(),$lhsp->op2p()}")
	if (m_doV && VN_IS(nodep->lhsp(),GteS) && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstNot $lhsp.castGteS, $lhsp.width1 , AstLtS $lhsp->op1p(),$lhsp->op2p() )\n");
	    AstNode* arg1p = nodep->lhsp()->op1p()->unlinkFrBack();
AstNode* arg2p = nodep->lhsp()->op2p()->unlinkFrBack();
AstNode* newp = new AstLtS(nodep->fileline(), arg1p, arg2p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_OneHot_0(AstOneHot* nodep) {
	// TREEOPV("AstOneHot{$lhsp.width1}",          "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstOneHot $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_OneHot0_0(AstOneHot0* nodep) {
	// TREEOPV("AstOneHot0{$lhsp.width1}",         "replaceNum(nodep,1)")
	if (m_doV && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstOneHot0 $lhsp.width1 , replaceNum(nodep,1) )\n");
	    replaceNum(nodep,1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_0(AstOr* nodep) {
	// TREEOP ("AstOr    {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_1(AstOr* nodep) {
	// TREEOP ("AstOr    {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_2(AstOr* nodep) {
	// TREEOP ("AstOr    {$lhsp.isAllOnes, $rhsp, isTPure($rhsp)}",        "replaceWLhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isAllOnes() && isTPure(nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr $lhsp.isAllOnes, $rhsp, isTPure($rhsp) , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_3(AstOr* nodep) {
	// TREEOP ("AstOr    {$lhsp, $rhsp.isAllOnes, isTPure($lhsp)}",        "replaceWRhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isAllOnes() && isTPure(nodep->lhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr $lhsp, $rhsp.isAllOnes, isTPure($lhsp) , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_4(AstOr* nodep) {
	// TREEOP ("AstOr     {operandsSame($lhsp,,$rhsp)}",   "replaceWLhs(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr operandsSame($lhsp,,$rhsp) , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_5(AstOr* nodep) {
	// TREEOP ("AstOr  {$lhsp.castAnd,$rhsp.castAnd,operandAndOrSame(nodep)}",     "replaceAndOr(nodep)")
	if (m_doNConst && VN_IS(nodep->lhsp(),And) && VN_IS(nodep->rhsp(),And) && operandAndOrSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr $lhsp.castAnd,$rhsp.castAnd,operandAndOrSame(nodep) , replaceAndOr(nodep) )\n");
	    replaceAndOr(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_6(AstOr* nodep) {
	// TREEOP ("AstOr  {matchOrAndNot(nodep)}",            "DONE")
	if (m_doNConst && matchOrAndNot(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr matchOrAndNot(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_7(AstOr* nodep) {
	// TREEOP ("AstOr  {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)")
	if (m_doNConst && operandShiftSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstOr operandShiftSame(nodep) , replaceShiftSame(nodep) )\n");
	    replaceShiftSame(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Or_8(AstOr* nodep) {
	// TREEOPC("AstOr  {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE")
	if (m_doCpp && nodep->widthMin() == 1 && matchBitOpTree(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPC( AstOr nodep->widthMin() == 1, matchBitOpTree(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Pow_0(AstPow* nodep) {
	// TREEOP ("AstPow   {$rhsp.isZero}",          "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPow $rhsp.isZero , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Pow_1(AstPow* nodep) {
	// TREEOP ("AstPow   {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero() && !nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPow $lhsp.isZero, !$rhsp.isZero , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Pow_2(AstPow* nodep) {
	// TREEOP ("AstPow   {operandIsTwo($lhsp), $rhsp}",    "replacePowShift(nodep)")
	if (m_doNConst && operandIsTwo(nodep->lhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPow operandIsTwo($lhsp), $rhsp , replacePowShift(nodep) )\n");
	    replacePowShift(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowSS_0(AstPowSS* nodep) {
	// TREEOP ("AstPowSS {$rhsp.isZero}",          "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowSS $rhsp.isZero , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowSU_0(AstPowSU* nodep) {
	// TREEOP ("AstPowSU {$rhsp.isZero}",          "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowSU $rhsp.isZero , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowSU_1(AstPowSU* nodep) {
	// TREEOP ("AstPowSU {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero() && !nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowSU $lhsp.isZero, !$rhsp.isZero , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowSU_2(AstPowSU* nodep) {
	// TREEOP ("AstPowSU {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero() && !nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowSU $lhsp.isZero, !$rhsp.isZero , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowUS_0(AstPowUS* nodep) {
	// TREEOP ("AstPowUS {$rhsp.isZero}",          "replaceNum(nodep, 1)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowUS $rhsp.isZero , replaceNum(nodep, 1) )\n");
	    replaceNum(nodep, 1);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PowUS_1(AstPowUS* nodep) {
	// TREEOP ("AstPowUS {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero() && !nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstPowUS $lhsp.isZero, !$rhsp.isZero , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_PutcN_0(AstPutcN* nodep) {
	// TREEOPA("AstPutcN{$lhsp.castConst, $rhsp.castConst, $thsp.castConst}",  "replaceConst(nodep)")
	if (VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const) && VN_IS(nodep->thsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstPutcN $lhsp.castConst, $rhsp.castConst, $thsp.castConst , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedAnd_0(AstRedAnd* nodep) {
	// TREEOPV("AstRedAnd{$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedAnd $lhsp, $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedAnd_1(AstRedAnd* nodep) {
	// TREEOPV("AstRedAnd{$lhsp.castConcat}",      "AstAnd{AstRedAnd{$lhsp->castConcat()->lhsp()}, AstRedAnd{$lhsp->castConcat()->rhsp()}}")
	if (m_doV && VN_IS(nodep->lhsp(),Concat)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedAnd $lhsp.castConcat , AstAnd AstRedAnd $lhsp->castConcat()->lhsp() , AstRedAnd $lhsp->castConcat()->rhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Concat)->lhsp()->unlinkFrBack();
AstNode* arg2p = VN_CAST(nodep->lhsp(),Concat)->rhsp()->unlinkFrBack();
AstNode* newp = new AstAnd(nodep->fileline(), new AstRedAnd(nodep->fileline(), arg1p), new AstRedAnd(nodep->fileline(), arg2p));
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedAnd_2(AstRedAnd* nodep) {
	// TREEOPV("AstRedAnd{$lhsp.castExtend, $lhsp->width() > VN_CAST($lhsp,,Extend)->lhsp()->width()}", "replaceZero(nodep)")
	if (m_doV && VN_IS(nodep->lhsp(),Extend) && nodep->lhsp()->width() > VN_CAST(nodep->lhsp(),Extend)->lhsp()->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedAnd $lhsp.castExtend, $lhsp->width() > VN_CAST($lhsp,,Extend)->lhsp()->width() , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedOr_0(AstRedOr* nodep) {
	// TREEOPV("AstRedOr {$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedOr $lhsp, $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedOr_1(AstRedOr* nodep) {
	// TREEOPV("AstRedOr {$lhsp.castConcat}",      "AstOr {AstRedOr {$lhsp->castConcat()->lhsp()}, AstRedOr {$lhsp->castConcat()->rhsp()}}")
	if (m_doV && VN_IS(nodep->lhsp(),Concat)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedOr $lhsp.castConcat , AstOr AstRedOr $lhsp->castConcat()->lhsp() , AstRedOr $lhsp->castConcat()->rhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Concat)->lhsp()->unlinkFrBack();
AstNode* arg2p = VN_CAST(nodep->lhsp(),Concat)->rhsp()->unlinkFrBack();
AstNode* newp = new AstOr(nodep->fileline(), new AstRedOr(nodep->fileline(), arg1p), new AstRedOr(nodep->fileline(), arg2p));
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedOr_2(AstRedOr* nodep) {
	// TREEOPV("AstRedOr {$lhsp.castExtend}",      "AstRedOr {$lhsp->castExtend()->lhsp()}")
	if (m_doV && VN_IS(nodep->lhsp(),Extend)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedOr $lhsp.castExtend , AstRedOr $lhsp->castExtend()->lhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Extend)->lhsp()->unlinkFrBack();
AstNode* newp = new AstRedOr(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedXor_0(AstRedXor* nodep) {
	// TREEOPV("AstRedXor{$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)")
	if (m_doV && nodep->lhsp()->width1()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedXor $lhsp, $lhsp.width1 , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedXor_1(AstRedXor* nodep) {
	// TREEOPV("AstRedXor{$lhsp.castConcat}",      "AstXor{AstRedXor{$lhsp->castConcat()->lhsp()}, AstRedXor{$lhsp->castConcat()->rhsp()}}")
	if (m_doV && VN_IS(nodep->lhsp(),Concat)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedXor $lhsp.castConcat , AstXor AstRedXor $lhsp->castConcat()->lhsp() , AstRedXor $lhsp->castConcat()->rhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Concat)->lhsp()->unlinkFrBack();
AstNode* arg2p = VN_CAST(nodep->lhsp(),Concat)->rhsp()->unlinkFrBack();
AstNode* newp = new AstXor(nodep->fileline(), new AstRedXor(nodep->fileline(), arg1p), new AstRedXor(nodep->fileline(), arg2p));
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedXor_2(AstRedXor* nodep) {
	// TREEOPV("AstRedXor{$lhsp.castExtend}",      "AstRedXor{$lhsp->castExtend()->lhsp()}")
	if (m_doV && VN_IS(nodep->lhsp(),Extend)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstRedXor $lhsp.castExtend , AstRedXor $lhsp->castExtend()->lhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Extend)->lhsp()->unlinkFrBack();
AstNode* newp = new AstRedXor(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_RedXor_3(AstRedXor* nodep) {
	// TREEOP ("AstRedXor{$lhsp.castXor, VN_IS(VN_CAST($lhsp,,Xor)->lhsp(),,Const)}", "AstXor{AstRedXor{$lhsp->castXor()->lhsp()}, AstRedXor{$lhsp->castXor()->rhsp()}}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Xor) && VN_IS(VN_CAST(nodep->lhsp(),Xor)->lhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstRedXor $lhsp.castXor, VN_IS(VN_CAST($lhsp,,Xor)->lhsp(),,Const) , AstXor AstRedXor $lhsp->castXor()->lhsp() , AstRedXor $lhsp->castXor()->rhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Xor)->lhsp()->unlinkFrBack();
AstNode* arg2p = VN_CAST(nodep->lhsp(),Xor)->rhsp()->unlinkFrBack();
AstNode* newp = new AstXor(nodep->fileline(), new AstRedXor(nodep->fileline(), arg1p), new AstRedXor(nodep->fileline(), arg2p));
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Replicate_0(AstReplicate* nodep) {
	// TREEOPV("AstReplicate{$lhsp, $rhsp.isOne, $lhsp->width()==nodep->width()}", "replaceWLhs(nodep)")
	if (m_doV && nodep->rhsp()->isOne() && nodep->lhsp()->width()==nodep->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstReplicate $lhsp, $rhsp.isOne, $lhsp->width()==nodep->width() , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Replicate_1(AstReplicate* nodep) {
	// TREEOPV("AstReplicate{$lhsp.castReplicate, operandRepRep(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->lhsp(),Replicate) && operandRepRep(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstReplicate $lhsp.castReplicate, operandRepRep(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ReplicateN_0(AstReplicateN* nodep) {
	// TREEOPV("AstReplicateN{$lhsp, $rhsp.isOne, $lhsp->width()==nodep->width()}", "replaceWLhs(nodep)")
	if (m_doV && nodep->rhsp()->isOne() && nodep->lhsp()->width()==nodep->width()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstReplicateN $lhsp, $rhsp.isOne, $lhsp->width()==nodep->width() , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_0(AstSel* nodep) {
	// TREEOP1("AstSel{warnSelect(nodep)}",        "NEVER")
	if (m_doNConst && warnSelect(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP1( AstSel warnSelect(nodep) , NEVER )\n");
	    nodep->v3fatalSrc("Executing transform that was NEVERed");
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_1(AstSel* nodep) {
	// TREEOPV("AstSel{operandSelExtend(nodep)}",  "DONE")
	if (m_doV && operandSelExtend(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel operandSelExtend(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_2(AstSel* nodep) {
	// TREEOPV("AstSel{operandSelFull(nodep)}",    "replaceWChild(nodep, nodep->fromp())")
	if (m_doV && operandSelFull(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel operandSelFull(nodep) , replaceWChild(nodep, nodep->fromp()) )\n");
	    replaceWChild(nodep, nodep->fromp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_3(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castSel}",           "replaceSelSel(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),Sel)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castSel , replaceSelSel(nodep) )\n");
	    replaceSelSel(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_4(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castAdd, operandSelBiLower(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->fromp(),Add) && operandSelBiLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castAdd, operandSelBiLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_5(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castAnd, operandSelBiLower(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->fromp(),And) && operandSelBiLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castAnd, operandSelBiLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_6(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castOr,  operandSelBiLower(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->fromp(),Or) && operandSelBiLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castOr, operandSelBiLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_7(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castSub, operandSelBiLower(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->fromp(),Sub) && operandSelBiLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castSub, operandSelBiLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_8(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castXor, operandSelBiLower(nodep)}", "DONE")
	if (m_doV && VN_IS(nodep->fromp(),Xor) && operandSelBiLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castXor, operandSelBiLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_9(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castShiftR, operandSelShiftLower(nodep)}",   "DONE")
	if (m_doV && VN_IS(nodep->fromp(),ShiftR) && operandSelShiftLower(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castShiftR, operandSelShiftLower(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_10(AstSel* nodep) {
	// TREEOPA("AstSel{$fromp.castConst, $lsbp.castConst, $widthp.castConst, }",   "replaceConst(nodep)")
	if (VN_IS(nodep->fromp(),Const) && VN_IS(nodep->lsbp(),Const) && VN_IS(nodep->widthp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstSel $fromp.castConst, $lsbp.castConst, $widthp.castConst, , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_11(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castConcat, $lsbp.castConst, $widthp.castConst, }",  "replaceSelConcat(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),Concat) && VN_IS(nodep->lsbp(),Const) && VN_IS(nodep->widthp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castConcat, $lsbp.castConst, $widthp.castConst, , replaceSelConcat(nodep) )\n");
	    replaceSelConcat(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_12(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castReplicate, $lsbp.castConst, $widthp.castConst, operandSelReplicate(nodep) }",    "DONE")
	if (m_doV && VN_IS(nodep->fromp(),Replicate) && VN_IS(nodep->lsbp(),Const) && VN_IS(nodep->widthp(),Const) && operandSelReplicate(nodep) ) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castReplicate, $lsbp.castConst, $widthp.castConst, operandSelReplicate(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_13(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castBufIf1}",                "replaceSelIntoBiop(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),BufIf1)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castBufIf1 , replaceSelIntoBiop(nodep) )\n");
	    replaceSelIntoBiop(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_14(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castNot}",                   "replaceSelIntoUniop(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),Not)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castNot , replaceSelIntoUniop(nodep) )\n");
	    replaceSelIntoUniop(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_15(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castAnd,$lhsp.castConst}",   "replaceSelIntoUniop(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),And) && VN_IS(nodep->lhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castAnd,$lhsp.castConst , replaceSelIntoUniop(nodep) )\n");
	    replaceSelIntoUniop(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_16(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castOr,$lhsp.castConst}",    "replaceSelIntoUniop(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),Or) && VN_IS(nodep->lhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castOr,$lhsp.castConst , replaceSelIntoUniop(nodep) )\n");
	    replaceSelIntoUniop(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sel_17(AstSel* nodep) {
	// TREEOPV("AstSel{$fromp.castXor,$lhsp.castConst}",   "replaceSelIntoUniop(nodep)")
	if (m_doV && VN_IS(nodep->fromp(),Xor) && VN_IS(nodep->lhsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPV( AstSel $fromp.castXor,$lhsp.castConst , replaceSelIntoUniop(nodep) )\n");
	    replaceSelIntoUniop(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftL_0(AstShiftL* nodep) {
	// TREEOP ("AstShiftL{$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftL $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftL_1(AstShiftL* nodep) {
	// TREEOP ("AstShiftL{$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftL $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftL_2(AstShiftL* nodep) {
	// TREEOP ("AstShiftL{operandHugeShiftL(nodep)}",      "replaceZero(nodep)")
	if (m_doNConst && operandHugeShiftL(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftL operandHugeShiftL(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftL_3(AstShiftL* nodep) {
	// TREEOP ("AstShiftL{operandShiftOp(nodep)}",         "replaceShiftOp(nodep)")
	if (m_doNConst && operandShiftOp(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftL operandShiftOp(nodep) , replaceShiftOp(nodep) )\n");
	    replaceShiftOp(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftL_4(AstShiftL* nodep) {
	// TREEOP ("AstShiftL{operandShiftShift(nodep)}",      "replaceShiftShift(nodep)")
	if (m_doNConst && operandShiftShift(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftL operandShiftShift(nodep) , replaceShiftShift(nodep) )\n");
	    replaceShiftShift(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftR_0(AstShiftR* nodep) {
	// TREEOP ("AstShiftR{$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftR $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftR_1(AstShiftR* nodep) {
	// TREEOP ("AstShiftR{$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftR $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftR_2(AstShiftR* nodep) {
	// TREEOP ("AstShiftR{operandHugeShiftR(nodep)}",      "replaceZero(nodep)")
	if (m_doNConst && operandHugeShiftR(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftR operandHugeShiftR(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftR_3(AstShiftR* nodep) {
	// TREEOP ("AstShiftR{operandShiftOp(nodep)}",         "replaceShiftOp(nodep)")
	if (m_doNConst && operandShiftOp(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftR operandShiftOp(nodep) , replaceShiftOp(nodep) )\n");
	    replaceShiftOp(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftR_4(AstShiftR* nodep) {
	// TREEOP ("AstShiftR{operandShiftShift(nodep)}",      "replaceShiftShift(nodep)")
	if (m_doNConst && operandShiftShift(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftR operandShiftShift(nodep) , replaceShiftShift(nodep) )\n");
	    replaceShiftShift(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftRS_0(AstShiftRS* nodep) {
	// TREEOP ("AstShiftRS{$lhsp.isZero, $rhsp}",  "replaceZeroChkPure(nodep,$rhsp)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftRS $lhsp.isZero, $rhsp , replaceZeroChkPure(nodep,$rhsp) )\n");
	    replaceZeroChkPure(nodep,nodep->rhsp());
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_ShiftRS_1(AstShiftRS* nodep) {
	// TREEOP ("AstShiftRS{$lhsp, $rhsp.isZero}",  "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstShiftRS $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sub_0(AstSub* nodep) {
	// TREEOP ("AstSub   {$lhsp.isZero, $rhsp}",   "AstNegate{$rhsp}")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstSub $lhsp.isZero, $rhsp , AstNegate $rhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstNegate(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sub_1(AstSub* nodep) {
	// TREEOP ("AstSub   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstSub $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sub_2(AstSub* nodep) {
	// TREEOP ("AstSub   {$lhsp.castAdd, operandSubAdd(nodep)}", "AstAdd{AstSub{$lhsp->castAdd()->lhsp(),$rhsp}, $lhsp->castAdd()->rhsp()}")
	if (m_doNConst && VN_IS(nodep->lhsp(),Add) && operandSubAdd(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstSub $lhsp.castAdd, operandSubAdd(nodep) , AstAdd AstSub $lhsp->castAdd()->lhsp(),$rhsp , $lhsp->castAdd()->rhsp() )\n");
	    AstNode* arg1p = VN_CAST(nodep->lhsp(),Add)->lhsp()->unlinkFrBack();
AstNode* arg2p = nodep->rhsp()->unlinkFrBack();
AstNode* arg3p = VN_CAST(nodep->lhsp(),Add)->rhsp()->unlinkFrBack();
AstNode* newp = new AstAdd(nodep->fileline(), new AstSub(nodep->fileline(), arg1p, arg2p), arg3p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Sub_3(AstSub* nodep) {
	// TREEOP ("AstSub    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstSub operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_SubstrN_0(AstSubstrN* nodep) {
	// TREEOPA("AstSubstrN{$lhsp.castConst, $rhsp.castConst, $thsp.castConst}",  "replaceConst(nodep)")
	if (VN_IS(nodep->lhsp(),Const) && VN_IS(nodep->rhsp(),Const) && VN_IS(nodep->thsp(),Const)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPA( AstSubstrN $lhsp.castConst, $rhsp.castConst, $thsp.castConst , replaceConst(nodep) )\n");
	    replaceConst(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_WordSel_0(AstWordSel* nodep) {
	// TREEOP ("AstWordSel{operandWordOOB(nodep)}",        "replaceZero(nodep)")
	if (m_doNConst && operandWordOOB(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstWordSel operandWordOOB(nodep) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_0(AstXor* nodep) {
	// TREEOP ("AstXor   {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)")
	if (m_doNConst && nodep->lhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstXor $lhsp.isZero, $rhsp , replaceWRhs(nodep) )\n");
	    replaceWRhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_1(AstXor* nodep) {
	// TREEOP ("AstXor   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)")
	if (m_doNConst && nodep->rhsp()->isZero()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstXor $lhsp, $rhsp.isZero , replaceWLhs(nodep) )\n");
	    replaceWLhs(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_2(AstXor* nodep) {
	// TREEOP ("AstXor   {$lhsp.isAllOnes, $rhsp}",        "AstNot{$rhsp}")
	if (m_doNConst && nodep->lhsp()->isAllOnes()) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstXor $lhsp.isAllOnes, $rhsp , AstNot $rhsp )\n");
	    AstNode* arg1p = nodep->rhsp()->unlinkFrBack();
AstNode* newp = new AstNot(nodep->fileline(), arg1p);
nodep->replaceWith(newp);VL_DO_DANGLING(nodep->deleteTree(), nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_3(AstXor* nodep) {
	// TREEOP ("AstXor    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)")
	if (m_doNConst && operandsSame(nodep->lhsp(),nodep->rhsp())) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstXor operandsSame($lhsp,,$rhsp) , replaceZero(nodep) )\n");
	    replaceZero(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_4(AstXor* nodep) {
	// TREEOP ("AstXor {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)")
	if (m_doNConst && operandShiftSame(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOP ( AstXor operandShiftSame(nodep) , replaceShiftSame(nodep) )\n");
	    replaceShiftSame(nodep);
	    return true;
	}
	return false;
    }
    // Generated by astgen
    bool match_Xor_5(AstXor* nodep) {
	// TREEOPC("AstXor {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE")
	if (m_doCpp && nodep->widthMin() == 1 && matchBitOpTree(nodep)) {
	    UINFO(7,cvtToHex(nodep)<<" TREEOPC( AstXor nodep->widthMin() == 1, matchBitOpTree(nodep) , DONE )\n");
	    
	    return true;
	}
	return false;
    }
    // TREEOP visitors, call each base type's match
    // Bottom class up, as more simple transforms are generally better
    // Generated by astgen
    virtual void visit(AstAcosD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAcoshD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAdd* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_Add_0(nodep)) return;
        if (match_Add_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAddD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAnd* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_And_0(nodep)) return;
        if (match_And_1(nodep)) return;
        if (match_And_2(nodep)) return;
        if (match_And_3(nodep)) return;
        if (match_And_4(nodep)) return;
        if (match_And_5(nodep)) return;
        if (match_And_6(nodep)) return;
        if (match_And_7(nodep)) return;
        if (match_And_8(nodep)) return;
        if (match_And_9(nodep)) return;
        if (match_And_10(nodep)) return;
    }
    // Generated by astgen
    virtual void visitGen(AstArraySel* nodep) {
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAsinD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAsinhD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAssocSel* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAtan2D* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAtanD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAtanhD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstAtoN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstBitsToRealD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstBufIf1* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCCast* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCLog2* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCastDynamic* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCeilD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstChangeXor* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_ChangeXor_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCompareNN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstConcat* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_Concat_0(nodep)) return;
        if (match_Concat_1(nodep)) return;
        if (match_Concat_2(nodep)) return;
        if (match_Concat_3(nodep)) return;
        if (match_Concat_4(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstConcatN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen with short-circuiting
    virtual void visit(AstCond* nodep) override {
      iterateAndNextNull(nodep->lhsp());
        if (match_Cond_0(nodep)) return;
        if (match_Cond_1(nodep)) return;
      iterateAndNextNull(nodep->rhsp());
      AstNodeTriop *tnp = VN_CAST(nodep, NodeTriop);
      if (tnp && tnp->thsp()) iterateAndNextNull(tnp->thsp());
        if (match_NodeCond_0(nodep)) return;
        if (match_NodeCond_1(nodep)) return;
        if (match_NodeCond_2(nodep)) return;
        if (match_NodeCond_3(nodep)) return;
        if (match_NodeCond_4(nodep)) return;
        if (match_NodeCond_5(nodep)) return;
        if (match_NodeCond_6(nodep)) return;
        if (match_NodeCond_7(nodep)) return;
        if (match_NodeCond_8(nodep)) return;
        if (match_NodeCond_9(nodep)) return;
        if (match_Cond_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCondBound* nodep) override {
        iterateChildren(nodep);
        if (match_NodeCond_0(nodep)) return;
        if (match_NodeCond_1(nodep)) return;
        if (match_NodeCond_2(nodep)) return;
        if (match_NodeCond_3(nodep)) return;
        if (match_NodeCond_4(nodep)) return;
        if (match_NodeCond_5(nodep)) return;
        if (match_NodeCond_6(nodep)) return;
        if (match_NodeCond_7(nodep)) return;
        if (match_NodeCond_8(nodep)) return;
        if (match_NodeCond_9(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCosD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCoshD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCountBits* nodep) override {
        iterateChildren(nodep);
        if (match_NodeQuadop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCountOnes* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstCvtPackString* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_CvtPackString_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstDiv* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_Div_0(nodep)) return;
        if (match_Div_1(nodep)) return;
        if (match_Div_2(nodep)) return;
        if (match_Div_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstDivD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstDivS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_DivS_0(nodep)) return;
        if (match_DivS_1(nodep)) return;
        if (match_DivS_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstEq* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_Eq_0(nodep)) return;
        if (match_Eq_1(nodep)) return;
        if (match_Eq_2(nodep)) return;
        if (match_Eq_3(nodep)) return;
        if (match_Eq_4(nodep)) return;
        if (match_Eq_5(nodep)) return;
        if (match_Eq_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstEqCase* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_EqCase_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstEqD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_EqD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstEqN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_EqN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstEqWild* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_EqWild_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstExpD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstExtend* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_Extend_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstExtendS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_ExtendS_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstFEof* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstFGetC* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstFGetS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstFUngetC* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstFloorD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGetcN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGetcRefN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGt* nodep) override {
        iterateChildren(nodep);
        if (match_Gt_2(nodep)) return;
        if (match_Gt_1(nodep)) return;
        if (match_NodeBiop_0(nodep)) return;
        if (match_Gt_0(nodep)) return;
        if (match_Gt_3(nodep)) return;
        if (match_Gt_4(nodep)) return;
        if (match_Gt_5(nodep)) return;
        if (match_Gt_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGtD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GtD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGtN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GtN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGtS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GtS_0(nodep)) return;
        if (match_GtS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGte* nodep) override {
        iterateChildren(nodep);
        if (match_Gte_2(nodep)) return;
        if (match_Gte_1(nodep)) return;
        if (match_NodeBiop_0(nodep)) return;
        if (match_Gte_0(nodep)) return;
        if (match_Gte_3(nodep)) return;
        if (match_Gte_4(nodep)) return;
        if (match_Gte_5(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGteD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GteD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGteN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GteN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstGteS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_GteS_0(nodep)) return;
        if (match_GteS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstHypotD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstISToRD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstIToRD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstIsUnbounded* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_IsUnbounded_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstIsUnknown* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLenN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLog10D* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen with short-circuiting
    virtual void visit(AstLogAnd* nodep) override {
      iterateAndNextNull(nodep->lhsp());
        if (match_LogAnd_0(nodep)) return;
      iterateAndNextNull(nodep->rhsp());
      AstNodeTriop *tnp = VN_CAST(nodep, NodeTriop);
      if (tnp && tnp->thsp()) iterateAndNextNull(tnp->thsp());
        if (match_NodeBiop_0(nodep)) return;
        if (match_LogAnd_1(nodep)) return;
        if (match_LogAnd_2(nodep)) return;
        if (match_LogAnd_3(nodep)) return;
        if (match_LogAnd_4(nodep)) return;
        if (match_LogAnd_5(nodep)) return;
        if (match_LogAnd_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLogD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLogEq* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_LogEq_0(nodep)) return;
    }
    // Generated by astgen with short-circuiting
    virtual void visit(AstLogIf* nodep) override {
      iterateAndNextNull(nodep->lhsp());
        if (match_LogIf_0(nodep)) return;
      iterateAndNextNull(nodep->rhsp());
      AstNodeTriop *tnp = VN_CAST(nodep, NodeTriop);
      if (tnp && tnp->thsp()) iterateAndNextNull(tnp->thsp());
        if (match_NodeBiop_0(nodep)) return;
        if (match_LogIf_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLogNot* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_LogNot_0(nodep)) return;
        if (match_LogNot_1(nodep)) return;
        if (match_LogNot_2(nodep)) return;
        if (match_LogNot_3(nodep)) return;
        if (match_LogNot_4(nodep)) return;
        if (match_LogNot_5(nodep)) return;
        if (match_LogNot_6(nodep)) return;
        if (match_LogNot_7(nodep)) return;
        if (match_LogNot_8(nodep)) return;
        if (match_LogNot_9(nodep)) return;
        if (match_LogNot_10(nodep)) return;
        if (match_LogNot_11(nodep)) return;
        if (match_LogNot_12(nodep)) return;
        if (match_LogNot_13(nodep)) return;
        if (match_LogNot_14(nodep)) return;
        if (match_LogNot_15(nodep)) return;
    }
    // Generated by astgen with short-circuiting
    virtual void visit(AstLogOr* nodep) override {
      iterateAndNextNull(nodep->lhsp());
        if (match_LogOr_0(nodep)) return;
      iterateAndNextNull(nodep->rhsp());
      AstNodeTriop *tnp = VN_CAST(nodep, NodeTriop);
      if (tnp && tnp->thsp()) iterateAndNextNull(tnp->thsp());
        if (match_NodeBiop_0(nodep)) return;
        if (match_LogOr_1(nodep)) return;
        if (match_LogOr_2(nodep)) return;
        if (match_LogOr_3(nodep)) return;
        if (match_LogOr_4(nodep)) return;
        if (match_LogOr_5(nodep)) return;
        if (match_LogOr_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLt* nodep) override {
        iterateChildren(nodep);
        if (match_Lt_2(nodep)) return;
        if (match_Lt_1(nodep)) return;
        if (match_NodeBiop_0(nodep)) return;
        if (match_Lt_0(nodep)) return;
        if (match_Lt_3(nodep)) return;
        if (match_Lt_4(nodep)) return;
        if (match_Lt_5(nodep)) return;
        if (match_Lt_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLtD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LtD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLtN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LtN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLtS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LtS_0(nodep)) return;
        if (match_LtS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLte* nodep) override {
        iterateChildren(nodep);
        if (match_Lte_2(nodep)) return;
        if (match_Lte_1(nodep)) return;
        if (match_NodeBiop_0(nodep)) return;
        if (match_Lte_0(nodep)) return;
        if (match_Lte_3(nodep)) return;
        if (match_Lte_4(nodep)) return;
        if (match_Lte_5(nodep)) return;
        if (match_Lte_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLteD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LteD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLteN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LteN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstLteS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_LteS_0(nodep)) return;
        if (match_LteS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstModDiv* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_ModDiv_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstModDivS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstMul* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_Mul_0(nodep)) return;
        if (match_Mul_1(nodep)) return;
        if (match_Mul_2(nodep)) return;
        if (match_Mul_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstMulD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstMulS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_MulS_0(nodep)) return;
        if (match_MulS_1(nodep)) return;
        if (match_MulS_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNegate* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNegateD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNeq* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_Neq_0(nodep)) return;
        if (match_Neq_1(nodep)) return;
        if (match_Neq_2(nodep)) return;
        if (match_Neq_3(nodep)) return;
        if (match_Neq_4(nodep)) return;
        if (match_Neq_5(nodep)) return;
        if (match_Neq_6(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNeqCase* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NeqCase_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNeqD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NeqD_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNeqN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NeqN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNeqWild* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NeqWild_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeBiCom* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeBiComAsv* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeBiop* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeCond* nodep) override {
        iterateChildren(nodep);
        if (match_NodeCond_0(nodep)) return;
        if (match_NodeCond_1(nodep)) return;
        if (match_NodeCond_2(nodep)) return;
        if (match_NodeCond_3(nodep)) return;
        if (match_NodeCond_4(nodep)) return;
        if (match_NodeCond_5(nodep)) return;
        if (match_NodeCond_6(nodep)) return;
        if (match_NodeCond_7(nodep)) return;
        if (match_NodeCond_8(nodep)) return;
        if (match_NodeCond_9(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeQuadop* nodep) override {
        iterateChildren(nodep);
        if (match_NodeQuadop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeSel* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeStream* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeSystemBiop* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeSystemUniop* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNodeUniop* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNot* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_Not_0(nodep)) return;
        if (match_Not_1(nodep)) return;
        if (match_Not_2(nodep)) return;
        if (match_Not_3(nodep)) return;
        if (match_Not_4(nodep)) return;
        if (match_Not_5(nodep)) return;
        if (match_Not_6(nodep)) return;
        if (match_Not_7(nodep)) return;
        if (match_Not_8(nodep)) return;
        if (match_Not_9(nodep)) return;
        if (match_Not_10(nodep)) return;
        if (match_Not_11(nodep)) return;
        if (match_Not_12(nodep)) return;
        if (match_Not_13(nodep)) return;
        if (match_Not_14(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstNullCheck* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstOneHot* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_OneHot_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstOneHot0* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_OneHot0_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstOr* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_Or_0(nodep)) return;
        if (match_Or_1(nodep)) return;
        if (match_Or_2(nodep)) return;
        if (match_Or_3(nodep)) return;
        if (match_Or_4(nodep)) return;
        if (match_Or_5(nodep)) return;
        if (match_Or_6(nodep)) return;
        if (match_Or_7(nodep)) return;
        if (match_Or_8(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPow* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_Pow_0(nodep)) return;
        if (match_Pow_1(nodep)) return;
        if (match_Pow_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPowD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPowSS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_PowSS_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPowSU* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_PowSU_0(nodep)) return;
        if (match_PowSU_1(nodep)) return;
        if (match_PowSU_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPowUS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_PowUS_0(nodep)) return;
        if (match_PowUS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstPutcN* nodep) override {
        iterateChildren(nodep);
        if (match_PutcN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRToIRoundS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRToIS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRealToBits* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRedAnd* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_RedAnd_0(nodep)) return;
        if (match_RedAnd_1(nodep)) return;
        if (match_RedAnd_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRedOr* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_RedOr_0(nodep)) return;
        if (match_RedOr_1(nodep)) return;
        if (match_RedOr_2(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstRedXor* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
        if (match_RedXor_0(nodep)) return;
        if (match_RedXor_1(nodep)) return;
        if (match_RedXor_2(nodep)) return;
        if (match_RedXor_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstReplicate* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_Replicate_0(nodep)) return;
        if (match_Replicate_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstReplicateN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_ReplicateN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSel* nodep) override {
        iterateChildren(nodep);
        if (match_Sel_0(nodep)) return;
        if (match_Sel_1(nodep)) return;
        if (match_Sel_2(nodep)) return;
        if (match_Sel_3(nodep)) return;
        if (match_Sel_4(nodep)) return;
        if (match_Sel_5(nodep)) return;
        if (match_Sel_6(nodep)) return;
        if (match_Sel_7(nodep)) return;
        if (match_Sel_8(nodep)) return;
        if (match_Sel_9(nodep)) return;
        if (match_Sel_10(nodep)) return;
        if (match_Sel_11(nodep)) return;
        if (match_Sel_12(nodep)) return;
        if (match_Sel_13(nodep)) return;
        if (match_Sel_14(nodep)) return;
        if (match_Sel_15(nodep)) return;
        if (match_Sel_16(nodep)) return;
        if (match_Sel_17(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstShiftL* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_ShiftL_0(nodep)) return;
        if (match_ShiftL_1(nodep)) return;
        if (match_ShiftL_2(nodep)) return;
        if (match_ShiftL_3(nodep)) return;
        if (match_ShiftL_4(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstShiftR* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_ShiftR_0(nodep)) return;
        if (match_ShiftR_1(nodep)) return;
        if (match_ShiftR_2(nodep)) return;
        if (match_ShiftR_3(nodep)) return;
        if (match_ShiftR_4(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstShiftRS* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_ShiftRS_0(nodep)) return;
        if (match_ShiftRS_1(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSigned* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSinD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSinhD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSqrtD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstStreamL* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstStreamR* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSub* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_Sub_0(nodep)) return;
        if (match_Sub_1(nodep)) return;
        if (match_Sub_2(nodep)) return;
        if (match_Sub_3(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSubD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstSubstrN* nodep) override {
        iterateChildren(nodep);
        if (match_SubstrN_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstTanD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstTanhD* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstTimeImport* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstToLowerN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstToUpperN* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstURandomRange* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstUnsigned* nodep) override {
        iterateChildren(nodep);
        if (match_NodeUniop_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstWordSel* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_WordSel_0(nodep)) return;
    }
    // Generated by astgen
    virtual void visit(AstXor* nodep) override {
        iterateChildren(nodep);
        if (match_NodeBiop_0(nodep)) return;
        if (match_NodeBiCom_0(nodep)) return;
        if (match_NodeBiComAsv_0(nodep)) return;
        if (match_NodeBiComAsv_1(nodep)) return;
        if (match_NodeBiComAsv_2(nodep)) return;
        if (match_NodeBiComAsv_3(nodep)) return;
        if (match_Xor_0(nodep)) return;
        if (match_Xor_1(nodep)) return;
        if (match_Xor_2(nodep)) return;
        if (match_Xor_3(nodep)) return;
        if (match_Xor_4(nodep)) return;
        if (match_Xor_5(nodep)) return;
    }
#line 2713 "../V3Const.cpp"
    // Generic constants on both side.  Do this first to avoid other replacements
//    TREEOPA("AstNodeBiop {$lhsp.castConst, $rhsp.castConst, nodep->isPredictOptimizable()}",  "replaceConst(nodep)");
#line 6397 "V3Const__gen.cpp"
#line 2715 "../V3Const.cpp"
//    TREEOPA("AstNodeUniop{$lhsp.castConst, !nodep->isOpaque(), nodep->isPredictOptimizable()}",  "replaceConst(nodep)");
#line 6400 "V3Const__gen.cpp"
#line 2716 "../V3Const.cpp"
//    TREEOPA("AstNodeQuadop{$lhsp.castConst, $rhsp.castConst, $thsp.castConst, $fhsp.castConst}",  "replaceConst(nodep)");
#line 6403 "V3Const__gen.cpp"
#line 2717 "../V3Const.cpp"
    // Zero on one side or the other
//    TREEOP ("AstAdd   {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)");
#line 6407 "V3Const__gen.cpp"
#line 2719 "../V3Const.cpp"
//    TREEOP ("AstAnd   {$lhsp.isZero, $rhsp, isTPure($rhsp)}",   "replaceZero(nodep)");  // Can't use replaceZeroChkPure as we make this pattern in ChkPure
#line 6410 "V3Const__gen.cpp"
#line 2720 "../V3Const.cpp"
    // This visit function here must allow for short-circuiting.
//    TREEOPS("AstLogAnd   {$lhsp.isZero}",       "replaceZero(nodep)");
#line 6414 "V3Const__gen.cpp"
#line 2722 "../V3Const.cpp"
//    TREEOP ("AstLogAnd{$lhsp.isZero, $rhsp}",   "replaceZero(nodep)");
#line 6417 "V3Const__gen.cpp"
#line 2723 "../V3Const.cpp"
    // This visit function here must allow for short-circuiting.
//    TREEOPS("AstLogOr   {$lhsp.isOne}",         "replaceNum(nodep, 1)");
#line 6421 "V3Const__gen.cpp"
#line 2725 "../V3Const.cpp"
//    TREEOP ("AstLogOr {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)");
#line 6424 "V3Const__gen.cpp"
#line 2726 "../V3Const.cpp"
//    TREEOP ("AstDiv   {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6427 "V3Const__gen.cpp"
#line 2727 "../V3Const.cpp"
//    TREEOP ("AstDivS  {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6430 "V3Const__gen.cpp"
#line 2728 "../V3Const.cpp"
//    TREEOP ("AstMul   {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6433 "V3Const__gen.cpp"
#line 2729 "../V3Const.cpp"
//    TREEOP ("AstMulS  {$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6436 "V3Const__gen.cpp"
#line 2730 "../V3Const.cpp"
//    TREEOP ("AstPow   {$rhsp.isZero}",          "replaceNum(nodep, 1)");  // Overrides lhs zero rule
#line 6439 "V3Const__gen.cpp"
#line 2731 "../V3Const.cpp"
//    TREEOP ("AstPowSS {$rhsp.isZero}",          "replaceNum(nodep, 1)");  // Overrides lhs zero rule
#line 6442 "V3Const__gen.cpp"
#line 2732 "../V3Const.cpp"
//    TREEOP ("AstPowSU {$rhsp.isZero}",          "replaceNum(nodep, 1)");  // Overrides lhs zero rule
#line 6445 "V3Const__gen.cpp"
#line 2733 "../V3Const.cpp"
//    TREEOP ("AstPowUS {$rhsp.isZero}",          "replaceNum(nodep, 1)");  // Overrides lhs zero rule
#line 6448 "V3Const__gen.cpp"
#line 2734 "../V3Const.cpp"
//    TREEOP ("AstPow   {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6451 "V3Const__gen.cpp"
#line 2735 "../V3Const.cpp"
//    TREEOP ("AstPowSU {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6454 "V3Const__gen.cpp"
#line 2736 "../V3Const.cpp"
//    TREEOP ("AstPowUS {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6457 "V3Const__gen.cpp"
#line 2737 "../V3Const.cpp"
//    TREEOP ("AstPowSU {$lhsp.isZero, !$rhsp.isZero}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6460 "V3Const__gen.cpp"
#line 2738 "../V3Const.cpp"
//    TREEOP ("AstOr    {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)");
#line 6463 "V3Const__gen.cpp"
#line 2739 "../V3Const.cpp"
//    TREEOP ("AstShiftL{$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6466 "V3Const__gen.cpp"
#line 2740 "../V3Const.cpp"
//    TREEOP ("AstShiftR{$lhsp.isZero, $rhsp}",   "replaceZeroChkPure(nodep,$rhsp)");
#line 6469 "V3Const__gen.cpp"
#line 2741 "../V3Const.cpp"
//    TREEOP ("AstShiftRS{$lhsp.isZero, $rhsp}",  "replaceZeroChkPure(nodep,$rhsp)");
#line 6472 "V3Const__gen.cpp"
#line 2742 "../V3Const.cpp"
//    TREEOP ("AstXor   {$lhsp.isZero, $rhsp}",   "replaceWRhs(nodep)");
#line 6475 "V3Const__gen.cpp"
#line 2743 "../V3Const.cpp"
//    TREEOP ("AstSub   {$lhsp.isZero, $rhsp}",   "AstNegate{$rhsp}");
#line 6478 "V3Const__gen.cpp"
#line 2744 "../V3Const.cpp"
//    TREEOP ("AstAdd   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6481 "V3Const__gen.cpp"
#line 2745 "../V3Const.cpp"
//    TREEOP ("AstAnd   {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)");
#line 6484 "V3Const__gen.cpp"
#line 2746 "../V3Const.cpp"
//    TREEOP ("AstLogAnd{$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)");
#line 6487 "V3Const__gen.cpp"
#line 2747 "../V3Const.cpp"
//    TREEOP ("AstLogOr {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6490 "V3Const__gen.cpp"
#line 2748 "../V3Const.cpp"
//    TREEOP ("AstMul   {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)");
#line 6493 "V3Const__gen.cpp"
#line 2749 "../V3Const.cpp"
//    TREEOP ("AstMulS  {$lhsp, $rhsp.isZero}",   "replaceZeroChkPure(nodep,$lhsp)");
#line 6496 "V3Const__gen.cpp"
#line 2750 "../V3Const.cpp"
//    TREEOP ("AstOr    {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6499 "V3Const__gen.cpp"
#line 2751 "../V3Const.cpp"
//    TREEOP ("AstShiftL{$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6502 "V3Const__gen.cpp"
#line 2752 "../V3Const.cpp"
//    TREEOP ("AstShiftR{$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6505 "V3Const__gen.cpp"
#line 2753 "../V3Const.cpp"
//    TREEOP ("AstShiftRS{$lhsp, $rhsp.isZero}",  "replaceWLhs(nodep)");
#line 6508 "V3Const__gen.cpp"
#line 2754 "../V3Const.cpp"
//    TREEOP ("AstSub   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6511 "V3Const__gen.cpp"
#line 2755 "../V3Const.cpp"
//    TREEOP ("AstXor   {$lhsp, $rhsp.isZero}",   "replaceWLhs(nodep)");
#line 6514 "V3Const__gen.cpp"
#line 2756 "../V3Const.cpp"
    // Non-zero on one side or the other
//    TREEOP ("AstAnd   {$lhsp.isAllOnes, $rhsp}",        "replaceWRhs(nodep)");
#line 6518 "V3Const__gen.cpp"
#line 2758 "../V3Const.cpp"
//    TREEOP ("AstLogAnd{$lhsp.isNeqZero, $rhsp}",        "replaceWRhs(nodep)");
#line 6521 "V3Const__gen.cpp"
#line 2759 "../V3Const.cpp"
//    TREEOP ("AstOr    {$lhsp.isAllOnes, $rhsp, isTPure($rhsp)}",        "replaceWLhs(nodep)");  //->allOnes
#line 6524 "V3Const__gen.cpp"
#line 2760 "../V3Const.cpp"
//    TREEOP ("AstLogOr {$lhsp.isNeqZero, $rhsp}",        "replaceNum(nodep,1)");
#line 6527 "V3Const__gen.cpp"
#line 2761 "../V3Const.cpp"
//    TREEOP ("AstAnd   {$lhsp, $rhsp.isAllOnes}",        "replaceWLhs(nodep)");
#line 6530 "V3Const__gen.cpp"
#line 2762 "../V3Const.cpp"
//    TREEOP ("AstLogAnd{$lhsp, $rhsp.isNeqZero}",        "replaceWLhs(nodep)");
#line 6533 "V3Const__gen.cpp"
#line 2763 "../V3Const.cpp"
//    TREEOP ("AstOr    {$lhsp, $rhsp.isAllOnes, isTPure($lhsp)}",        "replaceWRhs(nodep)");  //->allOnes
#line 6536 "V3Const__gen.cpp"
#line 2764 "../V3Const.cpp"
//    TREEOP ("AstLogOr {$lhsp, $rhsp.isNeqZero, isTPure($lhsp)}",        "replaceNum(nodep,1)");
#line 6539 "V3Const__gen.cpp"
#line 2765 "../V3Const.cpp"
//    TREEOP ("AstXor   {$lhsp.isAllOnes, $rhsp}",        "AstNot{$rhsp}");
#line 6542 "V3Const__gen.cpp"
#line 2766 "../V3Const.cpp"
//    TREEOP ("AstMul   {$lhsp.isOne, $rhsp}",    "replaceWRhs(nodep)");
#line 6545 "V3Const__gen.cpp"
#line 2767 "../V3Const.cpp"
//    TREEOP ("AstMulS  {$lhsp.isOne, $rhsp}",    "replaceWRhs(nodep)");
#line 6548 "V3Const__gen.cpp"
#line 2768 "../V3Const.cpp"
//    TREEOP ("AstDiv   {$lhsp, $rhsp.isOne}",    "replaceWLhs(nodep)");
#line 6551 "V3Const__gen.cpp"
#line 2769 "../V3Const.cpp"
//    TREEOP ("AstDivS  {$lhsp, $rhsp.isOne}",    "replaceWLhs(nodep)");
#line 6554 "V3Const__gen.cpp"
#line 2770 "../V3Const.cpp"
//    TREEOP ("AstMul   {operandIsPowTwo($lhsp), operandsSameSize($lhsp,,$rhsp)}", "replaceMulShift(nodep)");  // a*2^n -> a<<n
#line 6557 "V3Const__gen.cpp"
#line 2771 "../V3Const.cpp"
//    TREEOP ("AstDiv   {$lhsp, operandIsPowTwo($rhsp)}", "replaceDivShift(nodep)");  // a/2^n -> a>>n
#line 6560 "V3Const__gen.cpp"
#line 2772 "../V3Const.cpp"
//    TREEOP ("AstModDiv{$lhsp, operandIsPowTwo($rhsp)}", "replaceModAnd(nodep)");  // a % 2^n -> a&(2^n-1)
#line 6563 "V3Const__gen.cpp"
#line 2773 "../V3Const.cpp"
//    TREEOP ("AstPow   {operandIsTwo($lhsp), $rhsp}",    "replacePowShift(nodep)");  // 2**a == 1<<a
#line 6566 "V3Const__gen.cpp"
#line 2774 "../V3Const.cpp"
//    TREEOP ("AstSub   {$lhsp.castAdd, operandSubAdd(nodep)}", "AstAdd{AstSub{$lhsp->castAdd()->lhsp(),$rhsp}, $lhsp->castAdd()->rhsp()}");  // ((a+x)-y) -> (a+(x-y))
#line 6569 "V3Const__gen.cpp"
#line 2775 "../V3Const.cpp"
//    TREEOPC("AstAnd   {$lhsp.isOne, matchRedundantClean(nodep)}", "DONE")  // 1 & (a == b) -> (IData)(a == b)
#line 6572 "V3Const__gen.cpp"
#line 2776 "../V3Const.cpp"
    // Trinary ops
    // Note V3Case::Sel requires Cond to always be conditionally executed in C to prevent core dump!
//    TREEOP ("AstNodeCond{$condp.isZero,       $expr1p, $expr2p}", "replaceWChild(nodep,$expr2p)");
#line 6577 "V3Const__gen.cpp"
#line 2779 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp.isNeqZero,    $expr1p, $expr2p}", "replaceWChild(nodep,$expr1p)");
#line 6580 "V3Const__gen.cpp"
#line 2780 "../V3Const.cpp"
//    TREEOPA("AstNodeCond{$condp.isZero,       $expr1p.castConst, $expr2p.castConst}", "replaceWChild(nodep,$expr2p)");
#line 6583 "V3Const__gen.cpp"
#line 2781 "../V3Const.cpp"
//    TREEOPA("AstNodeCond{$condp.isNeqZero,    $expr1p.castConst, $expr2p.castConst}", "replaceWChild(nodep,$expr1p)");
#line 6586 "V3Const__gen.cpp"
#line 2782 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp, operandsSame($expr1p,,$expr2p)}","replaceWChild(nodep,$expr1p)");
#line 6589 "V3Const__gen.cpp"
#line 2783 "../V3Const.cpp"
    // This visit function here must allow for short-circuiting.
//    TREEOPS("AstCond {$lhsp.isZero}",           "replaceWIteratedThs(nodep)");
#line 6593 "V3Const__gen.cpp"
#line 2785 "../V3Const.cpp"
//    TREEOPS("AstCond {$lhsp.isNeqZero}",        "replaceWIteratedRhs(nodep)");
#line 6596 "V3Const__gen.cpp"
#line 2786 "../V3Const.cpp"
//    TREEOP ("AstCond{$condp.castNot,       $expr1p, $expr2p}", "AstCond{$condp->op1p(), $expr2p, $expr1p}");
#line 6599 "V3Const__gen.cpp"
#line 2787 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p.isAllOnes, $expr2p}", "AstLogOr {$condp, $expr2p}");  // a?1:b == a||b
#line 6602 "V3Const__gen.cpp"
#line 2788 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p,    $expr2p.isZero}", "AstLogAnd{$condp, $expr1p}");  // a?b:0 == a&&b
#line 6605 "V3Const__gen.cpp"
#line 2789 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p, $expr2p.isAllOnes}", "AstLogOr {AstNot{$condp}, $expr1p}");  // a?b:1 == ~a||b
#line 6608 "V3Const__gen.cpp"
#line 2790 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{$condp.width1, $expr1p.width1,   $expr1p.isZero,    $expr2p}", "AstLogAnd{AstNot{$condp}, $expr2p}");  // a?0:b == ~a&&b
#line 6611 "V3Const__gen.cpp"
#line 2791 "../V3Const.cpp"
//    TREEOP ("AstNodeCond{!$condp.width1, operandBoolShift(nodep->condp())}", "replaceBoolShift(nodep->condp())");
#line 6614 "V3Const__gen.cpp"
#line 2792 "../V3Const.cpp"
    // Prefer constants on left, since that often needs a shift, it lets
    // constant red remove the shift
//    TREEOP ("AstNodeBiCom{!$lhsp.castConst, $rhsp.castConst}",  "swapSides(nodep)");
#line 6619 "V3Const__gen.cpp"
#line 2795 "../V3Const.cpp"
//    TREEOP ("AstNodeBiComAsv{operandAsvConst(nodep)}",  "replaceAsv(nodep)");
#line 6622 "V3Const__gen.cpp"
#line 2796 "../V3Const.cpp"
//    TREEOP ("AstNodeBiComAsv{operandAsvSame(nodep)}",   "replaceAsv(nodep)");
#line 6625 "V3Const__gen.cpp"
#line 2797 "../V3Const.cpp"
//    TREEOP ("AstNodeBiComAsv{operandAsvLUp(nodep)}",    "replaceAsvLUp(nodep)");
#line 6628 "V3Const__gen.cpp"
#line 2798 "../V3Const.cpp"
//    TREEOP ("AstNodeBiComAsv{operandAsvRUp(nodep)}",    "replaceAsvRUp(nodep)");
#line 6631 "V3Const__gen.cpp"
#line 2799 "../V3Const.cpp"
//    TREEOP ("AstLt   {!$lhsp.castConst,$rhsp.castConst}",       "AstGt  {$rhsp,$lhsp}");
#line 6634 "V3Const__gen.cpp"
#line 2800 "../V3Const.cpp"
//    TREEOP ("AstLtS  {!$lhsp.castConst,$rhsp.castConst}",       "AstGtS {$rhsp,$lhsp}");
#line 6637 "V3Const__gen.cpp"
#line 2801 "../V3Const.cpp"
//    TREEOP ("AstLte  {!$lhsp.castConst,$rhsp.castConst}",       "AstGte {$rhsp,$lhsp}");
#line 6640 "V3Const__gen.cpp"
#line 2802 "../V3Const.cpp"
//    TREEOP ("AstLteS {!$lhsp.castConst,$rhsp.castConst}",       "AstGteS{$rhsp,$lhsp}");
#line 6643 "V3Const__gen.cpp"
#line 2803 "../V3Const.cpp"
//    TREEOP ("AstGt   {!$lhsp.castConst,$rhsp.castConst}",       "AstLt  {$rhsp,$lhsp}");
#line 6646 "V3Const__gen.cpp"
#line 2804 "../V3Const.cpp"
//    TREEOP ("AstGtS  {!$lhsp.castConst,$rhsp.castConst}",       "AstLtS {$rhsp,$lhsp}");
#line 6649 "V3Const__gen.cpp"
#line 2805 "../V3Const.cpp"
//    TREEOP ("AstGte  {!$lhsp.castConst,$rhsp.castConst}",       "AstLte {$rhsp,$lhsp}");
#line 6652 "V3Const__gen.cpp"
#line 2806 "../V3Const.cpp"
//    TREEOP ("AstGteS {!$lhsp.castConst,$rhsp.castConst}",       "AstLteS{$rhsp,$lhsp}");
#line 6655 "V3Const__gen.cpp"
#line 2807 "../V3Const.cpp"
    //    v--- *1* as These ops are always first, as we warn before replacing
//    TREEOP1("AstLt   {$lhsp, $rhsp.isZero}",            "replaceNumSigned(nodep,0)");
#line 6659 "V3Const__gen.cpp"
#line 2809 "../V3Const.cpp"
//    TREEOP1("AstGte  {$lhsp, $rhsp.isZero}",            "replaceNumSigned(nodep,1)");
#line 6662 "V3Const__gen.cpp"
#line 2810 "../V3Const.cpp"
//    TREEOP1("AstGt   {$lhsp.isZero, $rhsp}",            "replaceNumSigned(nodep,0)");
#line 6665 "V3Const__gen.cpp"
#line 2811 "../V3Const.cpp"
//    TREEOP1("AstLte  {$lhsp.isZero, $rhsp}",            "replaceNumSigned(nodep,1)");
#line 6668 "V3Const__gen.cpp"
#line 2812 "../V3Const.cpp"
//    TREEOP1("AstGt   {$lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,0)");
#line 6671 "V3Const__gen.cpp"
#line 2813 "../V3Const.cpp"
//    TREEOP1("AstLte  {$lhsp, $rhsp.isAllOnes, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,1)");
#line 6674 "V3Const__gen.cpp"
#line 2814 "../V3Const.cpp"
//    TREEOP1("AstLt   {$lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,0)");
#line 6677 "V3Const__gen.cpp"
#line 2815 "../V3Const.cpp"
//    TREEOP1("AstGte  {$lhsp.isAllOnes, $rhsp, $lhsp->width()==$rhsp->width()}",  "replaceNumLimited(nodep,1)");
#line 6680 "V3Const__gen.cpp"
#line 2816 "../V3Const.cpp"
    // Two level bubble pushing
//    TREEOP ("AstNot   {$lhsp.castNot,  $lhsp->width()==VN_CAST($lhsp,,Not)->lhsp()->width()}", "replaceWChild(nodep, $lhsp->op1p())");  // NOT(NOT(x))->x
#line 6684 "V3Const__gen.cpp"
#line 2818 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castLogNot}",              "replaceWChild(nodep, $lhsp->op1p())");  // LOGNOT(LOGNOT(x))->x
#line 6687 "V3Const__gen.cpp"
#line 2819 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castEqCase, $lhsp.width1}","AstNeqCase{$lhsp->op1p(),$lhsp->op2p()}");
#line 6690 "V3Const__gen.cpp"
#line 2820 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castEqCase}",              "AstNeqCase{$lhsp->op1p(),$lhsp->op2p()}");
#line 6693 "V3Const__gen.cpp"
#line 2821 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castNeqCase, $lhsp.width1}","AstEqCase {$lhsp->op1p(),$lhsp->op2p()}");
#line 6696 "V3Const__gen.cpp"
#line 2822 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castNeqCase}",             "AstEqCase {$lhsp->op1p(),$lhsp->op2p()}");
#line 6699 "V3Const__gen.cpp"
#line 2823 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castEqWild, $lhsp.width1}","AstNeqWild{$lhsp->op1p(),$lhsp->op2p()}");
#line 6702 "V3Const__gen.cpp"
#line 2824 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castEqWild}",              "AstNeqWild{$lhsp->op1p(),$lhsp->op2p()}");
#line 6705 "V3Const__gen.cpp"
#line 2825 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castNeqWild, $lhsp.width1}","AstEqWild {$lhsp->op1p(),$lhsp->op2p()}");
#line 6708 "V3Const__gen.cpp"
#line 2826 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castNeqWild}",             "AstEqWild {$lhsp->op1p(),$lhsp->op2p()}");
#line 6711 "V3Const__gen.cpp"
#line 2827 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castEq, $lhsp.width1}",    "AstNeq {$lhsp->op1p(),$lhsp->op2p()}");
#line 6714 "V3Const__gen.cpp"
#line 2828 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castEq}",                  "AstNeq {$lhsp->op1p(),$lhsp->op2p()}");
#line 6717 "V3Const__gen.cpp"
#line 2829 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castNeq, $lhsp.width1}",   "AstEq  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6720 "V3Const__gen.cpp"
#line 2830 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castNeq}",                 "AstEq  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6723 "V3Const__gen.cpp"
#line 2831 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castLt, $lhsp.width1}",    "AstGte {$lhsp->op1p(),$lhsp->op2p()}");
#line 6726 "V3Const__gen.cpp"
#line 2832 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castLt}",                  "AstGte {$lhsp->op1p(),$lhsp->op2p()}");
#line 6729 "V3Const__gen.cpp"
#line 2833 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castLtS, $lhsp.width1}",   "AstGteS{$lhsp->op1p(),$lhsp->op2p()}");
#line 6732 "V3Const__gen.cpp"
#line 2834 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castLtS}",                 "AstGteS{$lhsp->op1p(),$lhsp->op2p()}");
#line 6735 "V3Const__gen.cpp"
#line 2835 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castLte, $lhsp.width1}",   "AstGt  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6738 "V3Const__gen.cpp"
#line 2836 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castLte}",                 "AstGt  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6741 "V3Const__gen.cpp"
#line 2837 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castLteS, $lhsp.width1}",  "AstGtS {$lhsp->op1p(),$lhsp->op2p()}");
#line 6744 "V3Const__gen.cpp"
#line 2838 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castLteS}",                "AstGtS {$lhsp->op1p(),$lhsp->op2p()}");
#line 6747 "V3Const__gen.cpp"
#line 2839 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castGt, $lhsp.width1}",    "AstLte {$lhsp->op1p(),$lhsp->op2p()}");
#line 6750 "V3Const__gen.cpp"
#line 2840 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castGt}",                  "AstLte {$lhsp->op1p(),$lhsp->op2p()}");
#line 6753 "V3Const__gen.cpp"
#line 2841 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castGtS, $lhsp.width1}",   "AstLteS{$lhsp->op1p(),$lhsp->op2p()}");
#line 6756 "V3Const__gen.cpp"
#line 2842 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castGtS}",                 "AstLteS{$lhsp->op1p(),$lhsp->op2p()}");
#line 6759 "V3Const__gen.cpp"
#line 2843 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castGte, $lhsp.width1}",   "AstLt  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6762 "V3Const__gen.cpp"
#line 2844 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castGte}",                 "AstLt  {$lhsp->op1p(),$lhsp->op2p()}");
#line 6765 "V3Const__gen.cpp"
#line 2845 "../V3Const.cpp"
//    TREEOPV("AstNot   {$lhsp.castGteS, $lhsp.width1}",  "AstLtS {$lhsp->op1p(),$lhsp->op2p()}");
#line 6768 "V3Const__gen.cpp"
#line 2846 "../V3Const.cpp"
//    TREEOP ("AstLogNot{$lhsp.castGteS}",                "AstLtS {$lhsp->op1p(),$lhsp->op2p()}");
#line 6771 "V3Const__gen.cpp"
#line 2847 "../V3Const.cpp"
    // Not common, but avoids compiler warnings about over shifting
//    TREEOP ("AstShiftL{operandHugeShiftL(nodep)}",      "replaceZero(nodep)");
#line 6775 "V3Const__gen.cpp"
#line 2849 "../V3Const.cpp"
//    TREEOP ("AstShiftR{operandHugeShiftR(nodep)}",      "replaceZero(nodep)");
#line 6778 "V3Const__gen.cpp"
#line 2850 "../V3Const.cpp"
//    TREEOP ("AstShiftL{operandShiftOp(nodep)}",         "replaceShiftOp(nodep)");
#line 6781 "V3Const__gen.cpp"
#line 2851 "../V3Const.cpp"
//    TREEOP ("AstShiftR{operandShiftOp(nodep)}",         "replaceShiftOp(nodep)");
#line 6784 "V3Const__gen.cpp"
#line 2852 "../V3Const.cpp"
//    TREEOP ("AstShiftL{operandShiftShift(nodep)}",      "replaceShiftShift(nodep)");
#line 6787 "V3Const__gen.cpp"
#line 2853 "../V3Const.cpp"
//    TREEOP ("AstShiftR{operandShiftShift(nodep)}",      "replaceShiftShift(nodep)");
#line 6790 "V3Const__gen.cpp"
#line 2854 "../V3Const.cpp"
//    TREEOP ("AstWordSel{operandWordOOB(nodep)}",        "replaceZero(nodep)");
#line 6793 "V3Const__gen.cpp"
#line 2855 "../V3Const.cpp"
    // Compress out EXTENDs to appease loop unroller
//    TREEOPV("AstEq    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6797 "V3Const__gen.cpp"
#line 2857 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6800 "V3Const__gen.cpp"
#line 2858 "../V3Const.cpp"
//    TREEOPV("AstGt    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6803 "V3Const__gen.cpp"
#line 2859 "../V3Const.cpp"
//    TREEOPV("AstGte   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6806 "V3Const__gen.cpp"
#line 2860 "../V3Const.cpp"
//    TREEOPV("AstLt    {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6809 "V3Const__gen.cpp"
#line 2861 "../V3Const.cpp"
//    TREEOPV("AstLte   {$rhsp.castExtend,operandBiExtendConstShrink(nodep)}",    "DONE");
#line 6812 "V3Const__gen.cpp"
#line 2862 "../V3Const.cpp"
//    TREEOPV("AstEq    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)");
#line 6815 "V3Const__gen.cpp"
#line 2863 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)");
#line 6818 "V3Const__gen.cpp"
#line 2864 "../V3Const.cpp"
//    TREEOPV("AstGt    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)");
#line 6821 "V3Const__gen.cpp"
#line 2865 "../V3Const.cpp"
//    TREEOPV("AstGte   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceNum(nodep,1)");
#line 6824 "V3Const__gen.cpp"
#line 2866 "../V3Const.cpp"
//    TREEOPV("AstLt    {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)");
#line 6827 "V3Const__gen.cpp"
#line 2867 "../V3Const.cpp"
//    TREEOPV("AstLte   {$rhsp.castExtend,operandBiExtendConstOver(nodep)}",      "replaceZero(nodep)");
#line 6830 "V3Const__gen.cpp"
#line 2868 "../V3Const.cpp"
    // Identical operands on both sides
    // AstLogAnd/AstLogOr already converted to AstAnd/AstOr for these rules
    // AstAdd->ShiftL(#,1) but uncommon
//    TREEOP ("AstAnd    {operandsSame($lhsp,,$rhsp)}",   "replaceWLhs(nodep)");
#line 6836 "V3Const__gen.cpp"
#line 2872 "../V3Const.cpp"
//    TREEOP ("AstChangeXor{operandsSame($lhsp,,$rhsp)}", "replaceZero(nodep)");
#line 6839 "V3Const__gen.cpp"
#line 2873 "../V3Const.cpp"
//    TREEOP ("AstDiv    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6842 "V3Const__gen.cpp"
#line 2874 "../V3Const.cpp"
//    TREEOP ("AstDivS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6845 "V3Const__gen.cpp"
#line 2875 "../V3Const.cpp"
//    TREEOP ("AstOr     {operandsSame($lhsp,,$rhsp)}",   "replaceWLhs(nodep)");
#line 6848 "V3Const__gen.cpp"
#line 2876 "../V3Const.cpp"
//    TREEOP ("AstSub    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6851 "V3Const__gen.cpp"
#line 2877 "../V3Const.cpp"
//    TREEOP ("AstXor    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6854 "V3Const__gen.cpp"
#line 2878 "../V3Const.cpp"
//    TREEOP ("AstEq     {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");  // We let X==X -> 1, although in a true 4-state sim it's X.
#line 6857 "V3Const__gen.cpp"
#line 2879 "../V3Const.cpp"
//    TREEOP ("AstEqD    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");  // We let X==X -> 1, although in a true 4-state sim it's X.
#line 6860 "V3Const__gen.cpp"
#line 2880 "../V3Const.cpp"
//    TREEOP ("AstEqN    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");  // We let X==X -> 1, although in a true 4-state sim it's X.
#line 6863 "V3Const__gen.cpp"
#line 2881 "../V3Const.cpp"
//    TREEOP ("AstEqCase {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6866 "V3Const__gen.cpp"
#line 2882 "../V3Const.cpp"
//    TREEOP ("AstEqWild {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6869 "V3Const__gen.cpp"
#line 2883 "../V3Const.cpp"
//    TREEOP ("AstGt     {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6872 "V3Const__gen.cpp"
#line 2884 "../V3Const.cpp"
//    TREEOP ("AstGtD    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6875 "V3Const__gen.cpp"
#line 2885 "../V3Const.cpp"
//    TREEOP ("AstGtN    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6878 "V3Const__gen.cpp"
#line 2886 "../V3Const.cpp"
//    TREEOP ("AstGtS    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6881 "V3Const__gen.cpp"
#line 2887 "../V3Const.cpp"
//    TREEOP ("AstGte    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6884 "V3Const__gen.cpp"
#line 2888 "../V3Const.cpp"
//    TREEOP ("AstGteD   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6887 "V3Const__gen.cpp"
#line 2889 "../V3Const.cpp"
//    TREEOP ("AstGteN   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6890 "V3Const__gen.cpp"
#line 2890 "../V3Const.cpp"
//    TREEOP ("AstGteS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6893 "V3Const__gen.cpp"
#line 2891 "../V3Const.cpp"
//    TREEOP ("AstLt     {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6896 "V3Const__gen.cpp"
#line 2892 "../V3Const.cpp"
//    TREEOP ("AstLtD    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6899 "V3Const__gen.cpp"
#line 2893 "../V3Const.cpp"
//    TREEOP ("AstLtN    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6902 "V3Const__gen.cpp"
#line 2894 "../V3Const.cpp"
//    TREEOP ("AstLtS    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6905 "V3Const__gen.cpp"
#line 2895 "../V3Const.cpp"
//    TREEOP ("AstLte    {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6908 "V3Const__gen.cpp"
#line 2896 "../V3Const.cpp"
//    TREEOP ("AstLteD   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6911 "V3Const__gen.cpp"
#line 2897 "../V3Const.cpp"
//    TREEOP ("AstLteN   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6914 "V3Const__gen.cpp"
#line 2898 "../V3Const.cpp"
//    TREEOP ("AstLteS   {operandsSame($lhsp,,$rhsp)}",   "replaceNum(nodep,1)");
#line 6917 "V3Const__gen.cpp"
#line 2899 "../V3Const.cpp"
//    TREEOP ("AstNeq    {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6920 "V3Const__gen.cpp"
#line 2900 "../V3Const.cpp"
//    TREEOP ("AstNeqD   {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6923 "V3Const__gen.cpp"
#line 2901 "../V3Const.cpp"
//    TREEOP ("AstNeqN   {operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6926 "V3Const__gen.cpp"
#line 2902 "../V3Const.cpp"
//    TREEOP ("AstNeqCase{operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6929 "V3Const__gen.cpp"
#line 2903 "../V3Const.cpp"
//    TREEOP ("AstNeqWild{operandsSame($lhsp,,$rhsp)}",   "replaceZero(nodep)");
#line 6932 "V3Const__gen.cpp"
#line 2904 "../V3Const.cpp"
//    TREEOP ("AstLogAnd {operandsSame($lhsp,,$rhsp), $lhsp.width1}",     "replaceWLhs(nodep)");
#line 6935 "V3Const__gen.cpp"
#line 2905 "../V3Const.cpp"
//    TREEOP ("AstLogOr  {operandsSame($lhsp,,$rhsp), $lhsp.width1}",     "replaceWLhs(nodep)");
#line 6938 "V3Const__gen.cpp"
#line 2906 "../V3Const.cpp"
    ///=== Verilog operators
    // Comparison against 1'b0/1'b1; must be careful about widths.
    // These use Not, so must be Verilog only
//    TREEOPV("AstEq    {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "AstNot{$rhsp}");
#line 6944 "V3Const__gen.cpp"
#line 2910 "../V3Const.cpp"
//    TREEOPV("AstEq    {$lhsp.width1, $lhsp, $rhsp.isZero}",     "AstNot{$lhsp}");
#line 6947 "V3Const__gen.cpp"
#line 2911 "../V3Const.cpp"
//    TREEOPV("AstEq    {$rhsp.width1, $lhsp.isAllOnes, $rhsp}",  "replaceWRhs(nodep)");
#line 6950 "V3Const__gen.cpp"
#line 2912 "../V3Const.cpp"
//    TREEOPV("AstEq    {$lhsp.width1, $lhsp, $rhsp.isAllOnes}",  "replaceWLhs(nodep)");
#line 6953 "V3Const__gen.cpp"
#line 2913 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "replaceWRhs(nodep)");
#line 6956 "V3Const__gen.cpp"
#line 2914 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$lhsp.width1, $lhsp, $rhsp.isZero}",     "replaceWLhs(nodep)");
#line 6959 "V3Const__gen.cpp"
#line 2915 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$rhsp.width1, $lhsp.isAllOnes, $rhsp}",  "AstNot{$rhsp}");
#line 6962 "V3Const__gen.cpp"
#line 2916 "../V3Const.cpp"
//    TREEOPV("AstNeq   {$lhsp.width1, $lhsp, $rhsp.isAllOnes}",  "AstNot{$lhsp}");
#line 6965 "V3Const__gen.cpp"
#line 2917 "../V3Const.cpp"
//    TREEOPV("AstLt    {$rhsp.width1, $lhsp.isZero,    $rhsp}",  "replaceWRhs(nodep)");  // Because not signed #s
#line 6968 "V3Const__gen.cpp"
#line 2918 "../V3Const.cpp"
//    TREEOPV("AstGt    {$lhsp.width1, $lhsp, $rhsp.isZero}",     "replaceWLhs(nodep)");  // Because not signed #s
#line 6971 "V3Const__gen.cpp"
#line 2919 "../V3Const.cpp"
    // Useful for CONDs added around ARRAYSEL's in V3Case step
//    TREEOPV("AstLte   {$lhsp->width()==$rhsp->width(), $rhsp.isAllOnes}", "replaceNum(nodep,1)");
#line 6975 "V3Const__gen.cpp"
#line 2921 "../V3Const.cpp"
    // Simplify reduction operators
    // This also gets &{...,0,....} => const 0  (Common for unused_ok signals)
//    TREEOPV("AstRedAnd{$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)");
#line 6980 "V3Const__gen.cpp"
#line 2924 "../V3Const.cpp"
//    TREEOPV("AstRedOr {$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)");
#line 6983 "V3Const__gen.cpp"
#line 2925 "../V3Const.cpp"
//    TREEOPV("AstRedXor{$lhsp, $lhsp.width1}",   "replaceWLhs(nodep)");
#line 6986 "V3Const__gen.cpp"
#line 2926 "../V3Const.cpp"
//    TREEOPV("AstRedAnd{$lhsp.castConcat}",      "AstAnd{AstRedAnd{$lhsp->castConcat()->lhsp()}, AstRedAnd{$lhsp->castConcat()->rhsp()}}");  // &{a,b} => {&a}&{&b}
#line 6989 "V3Const__gen.cpp"
#line 2927 "../V3Const.cpp"
//    TREEOPV("AstRedOr {$lhsp.castConcat}",      "AstOr {AstRedOr {$lhsp->castConcat()->lhsp()}, AstRedOr {$lhsp->castConcat()->rhsp()}}");  // |{a,b} => {|a}|{|b}
#line 6992 "V3Const__gen.cpp"
#line 2928 "../V3Const.cpp"
//    TREEOPV("AstRedXor{$lhsp.castConcat}",      "AstXor{AstRedXor{$lhsp->castConcat()->lhsp()}, AstRedXor{$lhsp->castConcat()->rhsp()}}");  // ^{a,b} => {^a}^{^b}
#line 6995 "V3Const__gen.cpp"
#line 2929 "../V3Const.cpp"
//    TREEOPV("AstRedAnd{$lhsp.castExtend, $lhsp->width() > VN_CAST($lhsp,,Extend)->lhsp()->width()}", "replaceZero(nodep)");  // &{0,...} => 0  Prevents compiler limited range error
#line 6998 "V3Const__gen.cpp"
#line 2930 "../V3Const.cpp"
//    TREEOPV("AstRedOr {$lhsp.castExtend}",      "AstRedOr {$lhsp->castExtend()->lhsp()}");
#line 7001 "V3Const__gen.cpp"
#line 2931 "../V3Const.cpp"
//    TREEOPV("AstRedXor{$lhsp.castExtend}",      "AstRedXor{$lhsp->castExtend()->lhsp()}");
#line 7004 "V3Const__gen.cpp"
#line 2932 "../V3Const.cpp"
//    TREEOP ("AstRedXor{$lhsp.castXor, VN_IS(VN_CAST($lhsp,,Xor)->lhsp(),,Const)}", "AstXor{AstRedXor{$lhsp->castXor()->lhsp()}, AstRedXor{$lhsp->castXor()->rhsp()}}");  // ^(const ^ a) => (^const)^(^a)
#line 7007 "V3Const__gen.cpp"
#line 2933 "../V3Const.cpp"
//    TREEOPC("AstAnd {nodep->widthMin() == 1, $lhsp.castConst, $rhsp.castRedXor, matchBitOpTree(nodep)}", "DONE");
#line 7010 "V3Const__gen.cpp"
#line 2934 "../V3Const.cpp"
//    TREEOPV("AstOneHot{$lhsp.width1}",          "replaceWLhs(nodep)");
#line 7013 "V3Const__gen.cpp"
#line 2935 "../V3Const.cpp"
//    TREEOPV("AstOneHot0{$lhsp.width1}",         "replaceNum(nodep,1)");
#line 7016 "V3Const__gen.cpp"
#line 2936 "../V3Const.cpp"
    // Binary AND/OR is faster than logical and/or (usually)
//    TREEOPV("AstLogAnd{$lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp)}", "AstAnd{$lhsp,$rhsp}");
#line 7020 "V3Const__gen.cpp"
#line 2938 "../V3Const.cpp"
//    TREEOPV("AstLogOr {$lhsp.width1, $rhsp.width1, isTPure($lhsp), isTPure($rhsp)}", "AstOr{$lhsp,$rhsp}");
#line 7023 "V3Const__gen.cpp"
#line 2939 "../V3Const.cpp"
//    TREEOPV("AstLogNot{$lhsp.width1, isTPure($lhsp)}",  "AstNot{$lhsp}");
#line 7026 "V3Const__gen.cpp"
#line 2940 "../V3Const.cpp"
    // CONCAT(CONCAT({a},{b}),{c}) -> CONCAT({a},CONCAT({b},{c}))
    // CONCAT({const},CONCAT({const},{c})) -> CONCAT((constifiedCONC{const|const},{c}))
//    TREEOPV("AstConcat{operandConcatMove(nodep)}",      "moveConcat(nodep)");
#line 7031 "V3Const__gen.cpp"
#line 2943 "../V3Const.cpp"
//    TREEOPV("AstConcat{$lhsp.isZero, $rhsp}",           "replaceExtend(nodep, nodep->rhsp())");
#line 7034 "V3Const__gen.cpp"
#line 2944 "../V3Const.cpp"
    // CONCAT(a[1],a[0]) -> a[1:0]
//    TREEOPV("AstConcat{$lhsp.castSel, $rhsp.castSel, ifAdjacentSel(VN_CAST($lhsp,,Sel),,VN_CAST($rhsp,,Sel))}",  "replaceConcatSel(nodep)");
#line 7038 "V3Const__gen.cpp"
#line 2946 "../V3Const.cpp"
//    TREEOPV("AstConcat{ifConcatMergeableBiop($lhsp), concatMergeable($lhsp,,$rhsp)}", "replaceConcatMerge(nodep)");
#line 7041 "V3Const__gen.cpp"
#line 2947 "../V3Const.cpp"
    // Common two-level operations that can be simplified
//    TREEOP ("AstAnd {$lhsp.castConst,matchAndCond(nodep)}",           "DONE");
#line 7045 "V3Const__gen.cpp"
#line 2949 "../V3Const.cpp"
//    TREEOP ("AstAnd {$lhsp.castOr, $rhsp.castOr, operandAndOrSame(nodep)}",     "replaceAndOr(nodep)");
#line 7048 "V3Const__gen.cpp"
#line 2950 "../V3Const.cpp"
//    TREEOP ("AstOr  {$lhsp.castAnd,$rhsp.castAnd,operandAndOrSame(nodep)}",     "replaceAndOr(nodep)");
#line 7051 "V3Const__gen.cpp"
#line 2951 "../V3Const.cpp"
//    TREEOP ("AstOr  {matchOrAndNot(nodep)}",            "DONE");
#line 7054 "V3Const__gen.cpp"
#line 2952 "../V3Const.cpp"
//    TREEOP ("AstAnd {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)");
#line 7057 "V3Const__gen.cpp"
#line 2953 "../V3Const.cpp"
//    TREEOP ("AstOr  {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)");
#line 7060 "V3Const__gen.cpp"
#line 2954 "../V3Const.cpp"
//    TREEOP ("AstXor {operandShiftSame(nodep)}",         "replaceShiftSame(nodep)");
#line 7063 "V3Const__gen.cpp"
#line 2955 "../V3Const.cpp"
//    TREEOPC("AstAnd {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE");
#line 7066 "V3Const__gen.cpp"
#line 2956 "../V3Const.cpp"
//    TREEOPC("AstOr  {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE");
#line 7069 "V3Const__gen.cpp"
#line 2957 "../V3Const.cpp"
//    TREEOPC("AstXor {nodep->widthMin() == 1, matchBitOpTree(nodep)}", "DONE");
#line 7072 "V3Const__gen.cpp"
#line 2958 "../V3Const.cpp"
    // Note can't simplify a extend{extends}, extends{extend}, as the sign
    // bits end up in the wrong places
//    TREEOPV("AstExtend {$lhsp.castExtend}",  "replaceExtend(nodep, VN_CAST(nodep->lhsp(), Extend)->lhsp())");
#line 7077 "V3Const__gen.cpp"
#line 2961 "../V3Const.cpp"
//    TREEOPV("AstExtendS{$lhsp.castExtendS}", "replaceExtend(nodep, VN_CAST(nodep->lhsp(), ExtendS)->lhsp())");
#line 7080 "V3Const__gen.cpp"
#line 2962 "../V3Const.cpp"
//    TREEOPV("AstReplicate{$lhsp, $rhsp.isOne, $lhsp->width()==nodep->width()}", "replaceWLhs(nodep)");  // {1{lhs}}->lhs
#line 7083 "V3Const__gen.cpp"
#line 2963 "../V3Const.cpp"
//    TREEOPV("AstReplicateN{$lhsp, $rhsp.isOne, $lhsp->width()==nodep->width()}", "replaceWLhs(nodep)");  // {1{lhs}}->lhs
#line 7086 "V3Const__gen.cpp"
#line 2964 "../V3Const.cpp"
//    TREEOPV("AstReplicate{$lhsp.castReplicate, operandRepRep(nodep)}", "DONE");  // {2{3{lhs}}}->{6{lhs}}
#line 7089 "V3Const__gen.cpp"
#line 2965 "../V3Const.cpp"
//    TREEOPV("AstConcat{operandConcatSame(nodep)}", "DONE");  // {a,a}->{2{a}}, {a,2{a}}->{3{a}, etc
#line 7092 "V3Const__gen.cpp"
#line 2966 "../V3Const.cpp"
    // Next rule because AUTOINST puts the width of bits in
    // to pins, even when the widths are exactly the same across the hierarchy.
//    TREEOPV("AstSel{operandSelExtend(nodep)}",  "DONE");
#line 7097 "V3Const__gen.cpp"
#line 2969 "../V3Const.cpp"
//    TREEOPV("AstSel{operandSelFull(nodep)}",    "replaceWChild(nodep, nodep->fromp())");
#line 7100 "V3Const__gen.cpp"
#line 2970 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castSel}",           "replaceSelSel(nodep)");
#line 7103 "V3Const__gen.cpp"
#line 2971 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castAdd, operandSelBiLower(nodep)}", "DONE");
#line 7106 "V3Const__gen.cpp"
#line 2972 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castAnd, operandSelBiLower(nodep)}", "DONE");
#line 7109 "V3Const__gen.cpp"
#line 2973 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castOr,  operandSelBiLower(nodep)}", "DONE");
#line 7112 "V3Const__gen.cpp"
#line 2974 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castSub, operandSelBiLower(nodep)}", "DONE");
#line 7115 "V3Const__gen.cpp"
#line 2975 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castXor, operandSelBiLower(nodep)}", "DONE");
#line 7118 "V3Const__gen.cpp"
#line 2976 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castShiftR, operandSelShiftLower(nodep)}",   "DONE");
#line 7121 "V3Const__gen.cpp"
#line 2977 "../V3Const.cpp"
//    TREEOPA("AstSel{$fromp.castConst, $lsbp.castConst, $widthp.castConst, }",   "replaceConst(nodep)");
#line 7124 "V3Const__gen.cpp"
#line 2978 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castConcat, $lsbp.castConst, $widthp.castConst, }",  "replaceSelConcat(nodep)");
#line 7127 "V3Const__gen.cpp"
#line 2979 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castReplicate, $lsbp.castConst, $widthp.castConst, operandSelReplicate(nodep) }",    "DONE");
#line 7130 "V3Const__gen.cpp"
#line 2980 "../V3Const.cpp"
    // V3Tristate requires selects below BufIf1.
    // Also do additional operators that are bit-independent, but only definite
    // win if bit select is a constant (otherwise we may need to compute bit index several times)
//    TREEOPV("AstSel{$fromp.castBufIf1}",                "replaceSelIntoBiop(nodep)");
#line 7136 "V3Const__gen.cpp"
#line 2984 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castNot}",                   "replaceSelIntoUniop(nodep)");
#line 7139 "V3Const__gen.cpp"
#line 2985 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castAnd,$lhsp.castConst}",   "replaceSelIntoUniop(nodep)");
#line 7142 "V3Const__gen.cpp"
#line 2986 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castOr,$lhsp.castConst}",    "replaceSelIntoUniop(nodep)");
#line 7145 "V3Const__gen.cpp"
#line 2987 "../V3Const.cpp"
//    TREEOPV("AstSel{$fromp.castXor,$lhsp.castConst}",   "replaceSelIntoUniop(nodep)");
#line 7148 "V3Const__gen.cpp"
#line 2988 "../V3Const.cpp"
    // This visit function here must allow for short-circuiting.
//    TREEOPS("AstLogIf{$lhsp.isZero}",  "replaceNum(nodep, 1)");
#line 7152 "V3Const__gen.cpp"
#line 2990 "../V3Const.cpp"
//    TREEOPV("AstLogIf{$lhsp, $rhsp}",  "AstLogOr{AstLogNot{$lhsp},$rhsp}");
#line 7155 "V3Const__gen.cpp"
#line 2991 "../V3Const.cpp"
//    TREEOPV("AstLogEq{$lhsp, $rhsp}",  "replaceLogEq(nodep)");
#line 7158 "V3Const__gen.cpp"
#line 2992 "../V3Const.cpp"
    // Strings
//    TREEOPA("AstPutcN{$lhsp.castConst, $rhsp.castConst, $thsp.castConst}",  "replaceConst(nodep)");
#line 7162 "V3Const__gen.cpp"
#line 2994 "../V3Const.cpp"
//    TREEOPA("AstSubstrN{$lhsp.castConst, $rhsp.castConst, $thsp.castConst}",  "replaceConst(nodep)");
#line 7165 "V3Const__gen.cpp"
#line 2995 "../V3Const.cpp"
//    TREEOPA("AstCvtPackString{$lhsp.castConst}", "replaceConstString(nodep, VN_CAST(nodep->lhsp(), Const)->num().toString())");
#line 7168 "V3Const__gen.cpp"
#line 2996 "../V3Const.cpp"
    // Custom
    // Implied by AstIsUnbounded::numberOperate: V("AstIsUnbounded{$lhsp.castConst}", "replaceNum(nodep, 0)");
//    TREEOPV("AstIsUnbounded{$lhsp.castUnbounded}", "replaceNum(nodep, 1)");
#line 7173 "V3Const__gen.cpp"
#line 2999 "../V3Const.cpp"
    // clang-format on

    // Possible futures:
    // (a?(b?y:x):y) -> (a&&!b)?x:y
    // (a?(b?x:y):y) -> (a&&b)?x:y
    // (a?x:(b?x:y)) -> (a||b)?x:y
    // (a?x:(b?y:x)) -> (a||!b)?x:y

    // Note we can't convert EqCase/NeqCase to Eq/Neq here because that would break 3'b1x1==3'b101

    //-----
    virtual void visit(AstNode* nodep) override {
        // Default: Just iterate
        if (m_required) {
            if (VN_IS(nodep, NodeDType) || VN_IS(nodep, Range)) {
                // Ignore dtypes for parameter type pins
            } else {
                nodep->v3error("Expecting expression to be constant, but can't convert a "
                               << nodep->prettyTypeName() << " to constant.");
            }
        } else {
            // Calculate the width of this operation
            if (m_params && !nodep->width()) nodep = V3Width::widthParamsEdit(nodep);
            iterateChildren(nodep);
        }
    }

public:
    // Processing Mode Enum
    enum ProcMode : uint8_t {
        PROC_PARAMS,
        PROC_GENERATE,
        PROC_LIVE,
        PROC_V_WARN,
        PROC_V_NOWARN,
        PROC_V_EXPENSIVE,
        PROC_CPP
    };

    // CONSTRUCTORS
    explicit ConstVisitor(ProcMode pmode) {
        // clang-format off
        switch (pmode) {
        case PROC_PARAMS:       m_doV = true;  m_doNConst = true; m_params = true;
                                m_required = true; break;
        case PROC_GENERATE:     m_doV = true;  m_doNConst = true; m_params = true;
                                m_required = true; m_doGenerate = true; break;
        case PROC_LIVE:         break;
        case PROC_V_WARN:       m_doV = true;  m_doNConst = true; m_warn = true; break;
        case PROC_V_NOWARN:     m_doV = true;  m_doNConst = true; break;
        case PROC_V_EXPENSIVE:  m_doV = true;  m_doNConst = true; m_doExpensive = true; break;
        case PROC_CPP:          m_doV = false; m_doNConst = true; m_doCpp = true; break;
        default:                v3fatalSrc("Bad case"); break;
        }
        // clang-format on
    }
    virtual ~ConstVisitor() override {
        if (m_doCpp) {
            V3Stats::addStat("Optimizations, Const bit op reduction", m_statBitOpReduction);
        }
    }

    AstNode* mainAcceptEdit(AstNode* nodep) {
        // Operate starting at a random place
        return iterateSubtreeReturnEdits(nodep);
    }
};

//######################################################################
// Const class functions

//! Force this cell node's parameter list to become a constant
//! @return  Pointer to the edited node.
AstNode* V3Const::constifyParamsEdit(AstNode* nodep) {
    // if (debug() > 0) nodep->dumpTree(cout, "  forceConPRE : ");
    // Resize even if the node already has a width, because buried in the tree
    // we may have a node we just created with signing, etc, that isn't sized yet.

    // Make sure we've sized everything first
    nodep = V3Width::widthParamsEdit(nodep);
    ConstVisitor visitor(ConstVisitor::PROC_PARAMS);
    if (AstVar* varp = VN_CAST(nodep, Var)) {
        // If a var wants to be constified, it's really a param, and
        // we want the value to be constant.  We aren't passed just the
        // init value because we need widthing above to handle the var's type.
        if (varp->valuep()) visitor.mainAcceptEdit(varp->valuep());
    } else {
        nodep = visitor.mainAcceptEdit(nodep);
    }
    // Because we do edits, nodep links may get trashed and core dump this.
    // if (debug() > 0) nodep->dumpTree(cout, "  forceConDONE: ");
    return nodep;
}

//! Force this cell node's parameter list to become a constant inside generate.
//! If we are inside a generated "if", "case" or "for", we don't want to
//! trigger warnings when we deal with the width. It is possible that these
//! are spurious, existing within sub-expressions that will not actually be
//! generated. Since such occurrences, must be constant, in order to be
//! something a generate block can depend on, we can wait until later to do the
//! width check.
//! @return  Pointer to the edited node.
AstNode* V3Const::constifyGenerateParamsEdit(AstNode* nodep) {
    // if (debug() > 0) nodep->dumpTree(cout, "  forceConPRE : ");
    // Resize even if the node already has a width, because buried in the tree
    // we may have a node we just created with signing, etc, that isn't sized
    // yet.

    // Make sure we've sized everything first
    nodep = V3Width::widthGenerateParamsEdit(nodep);
    ConstVisitor visitor(ConstVisitor::PROC_GENERATE);
    if (AstVar* varp = VN_CAST(nodep, Var)) {
        // If a var wants to be constified, it's really a param, and
        // we want the value to be constant.  We aren't passed just the
        // init value because we need widthing above to handle the var's type.
        if (varp->valuep()) visitor.mainAcceptEdit(varp->valuep());
    } else {
        nodep = visitor.mainAcceptEdit(nodep);
    }
    // Because we do edits, nodep links may get trashed and core dump this.
    // if (debug() > 0) nodep->dumpTree(cout, "  forceConDONE: ");
    return nodep;
}

void V3Const::constifyAllLint(AstNetlist* nodep) {
    // Only call from Verilator.cpp, as it uses user#'s
    UINFO(2, __FUNCTION__ << ": " << endl);
    {
        ConstVisitor visitor(ConstVisitor::PROC_V_WARN);
        (void)visitor.mainAcceptEdit(nodep);
    }  // Destruct before checking
    V3Global::dumpCheckGlobalTree("const", 0, v3Global.opt.dumpTreeLevel(__FILE__) >= 3);
}

void V3Const::constifyCpp(AstNetlist* nodep) {
    UINFO(2, __FUNCTION__ << ": " << endl);
    {
        ConstVisitor visitor(ConstVisitor::PROC_CPP);
        (void)visitor.mainAcceptEdit(nodep);
    }  // Destruct before checking
    V3Global::dumpCheckGlobalTree("const_cpp", 0, v3Global.opt.dumpTreeLevel(__FILE__) >= 3);
}

AstNode* V3Const::constifyEdit(AstNode* nodep) {
    ConstVisitor visitor(ConstVisitor::PROC_V_NOWARN);
    nodep = visitor.mainAcceptEdit(nodep);
    return nodep;
}

void V3Const::constifyAllLive(AstNetlist* nodep) {
    // Only call from Verilator.cpp, as it uses user#'s
    // This only pushes constants up, doesn't make any other edits
    // IE doesn't prune dead statements, as we need to do some usability checks after this
    UINFO(2, __FUNCTION__ << ": " << endl);
    {
        ConstVisitor visitor(ConstVisitor::PROC_LIVE);
        (void)visitor.mainAcceptEdit(nodep);
    }  // Destruct before checking
    V3Global::dumpCheckGlobalTree("const", 0, v3Global.opt.dumpTreeLevel(__FILE__) >= 3);
}

void V3Const::constifyAll(AstNetlist* nodep) {
    // Only call from Verilator.cpp, as it uses user#'s
    UINFO(2, __FUNCTION__ << ": " << endl);
    {
        ConstVisitor visitor(ConstVisitor::PROC_V_EXPENSIVE);
        (void)visitor.mainAcceptEdit(nodep);
    }  // Destruct before checking
    V3Global::dumpCheckGlobalTree("const", 0, v3Global.opt.dumpTreeLevel(__FILE__) >= 3);
}

AstNode* V3Const::constifyExpensiveEdit(AstNode* nodep) {
    ConstVisitor visitor(ConstVisitor::PROC_V_EXPENSIVE);
    nodep = visitor.mainAcceptEdit(nodep);
    return nodep;
}
